#include <scheme/EulerKineticSolver.hpp>

#include <language/utils/EvaluateAtPoints.hpp>
#include <mesh/Connectivity.hpp>
#include <mesh/Mesh.hpp>
#include <mesh/MeshVariant.hpp>
#include <scheme/DiscreteFunctionDescriptorP0Vector.hpp>
#include <scheme/DiscreteFunctionP0Vector.hpp>
#include <scheme/DiscreteFunctionUtils.hpp>
#include <scheme/IDiscreteFunctionDescriptor.hpp>

#include <tuple>

template <MeshConcept MeshType>
std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticExactSolution(const std::shared_ptr<const MeshType> p_mesh, const double t)
{
  DiscreteFunctionP0<double> rho_ex(p_mesh);
  DiscreteFunctionP0<double> u_ex(p_mesh);

  MeshData<MeshType>& mesh_data                                    = MeshDataManager::instance().getMeshData(*p_mesh);
  const CellValue<const TinyVector<MeshType::Dimension>>& x_vector = mesh_data.xj();

  const double pi = acos(-1);
  // EvaluateAtPoints<double(double)>::evaluateTo(density_function_id, x, rho0);

  for (CellId cell_id = 0; cell_id < p_mesh->numberOfCells(); ++cell_id) {
    double xj      = x_vector[cell_id][0];
    double a1      = -1.5;
    double b1      = 1.5;
    double c1      = 0.5 * (a1 + b1);
    double x1      = c1;
    double rho0_x1 = 1 + 0.5 * sin(2 * pi * x1);
    double rho0_a1 = 1 + 0.5 * sin(2 * pi * a1);
    double rho0_b1 = 1 + 0.5 * sin(2 * pi * b1);
    double rho0_c1 = 1 + 0.5 * sin(2 * pi * c1);

    double fx1 = xj + sqrt(3) * rho0_x1 * t - x1;
    double fa1 = xj + sqrt(3) * rho0_a1 * t - a1;
    double fb1 = xj + sqrt(3) * rho0_b1 * t - b1;
    double fc1 = xj + sqrt(3) * rho0_c1 * t - c1;

    while (std::abs(fx1) > 1E-15) {
      if (fa1 * fc1 < 0) {
        b1 = c1;
        c1 = 0.5 * (a1 + b1);
      } else if (fc1 * fb1 < 0) {
        a1 = c1;
        c1 = 0.5 * (a1 + b1);
      }
      x1 = c1;

      rho0_x1 = 1 + 0.5 * sin(2 * pi * x1);
      rho0_a1 = 1 + 0.5 * sin(2 * pi * a1);
      rho0_b1 = 1 + 0.5 * sin(2 * pi * b1);
      rho0_c1 = 1 + 0.5 * sin(2 * pi * c1);

      fx1 = xj + sqrt(3) * rho0_x1 * t - x1;
      fa1 = xj + sqrt(3) * rho0_a1 * t - a1;
      fb1 = xj + sqrt(3) * rho0_b1 * t - b1;
      fc1 = xj + sqrt(3) * rho0_c1 * t - c1;
    }

    double a2      = -1.5;
    double b2      = 1.5;
    double c2      = 0.5 * (a2 + b2);
    double x2      = c2;
    double rho0_x2 = 1 + 0.5 * sin(2 * pi * x2);
    double rho0_a2 = 1 + 0.5 * sin(2 * pi * a2);
    double rho0_b2 = 1 + 0.5 * sin(2 * pi * b2);
    double rho0_c2 = 1 + 0.5 * sin(2 * pi * c2);

    double fx2 = xj - sqrt(3) * rho0_x2 * t - x2;
    double fa2 = xj - sqrt(3) * rho0_a2 * t - a2;
    double fb2 = xj - sqrt(3) * rho0_b2 * t - b2;
    double fc2 = xj - sqrt(3) * rho0_c2 * t - c2;

    while (std::abs(fx2) > 1E-15) {
      if (fa2 * fc2 < 0) {
        b2 = c2;
        c2 = 0.5 * (a2 + b2);
      } else if (fc2 * fb2 < 0) {
        a2 = c2;
        c2 = 0.5 * (a2 + b2);
      }
      x2 = c2;

      rho0_x2 = 1 + 0.5 * sin(2 * pi * x2);
      rho0_a2 = 1 + 0.5 * sin(2 * pi * a2);
      rho0_b2 = 1 + 0.5 * sin(2 * pi * b2);
      rho0_c2 = 1 + 0.5 * sin(2 * pi * c2);

      fx2 = xj - sqrt(3) * rho0_x2 * t - x2;
      fa2 = xj - sqrt(3) * rho0_a2 * t - a2;
      fb2 = xj - sqrt(3) * rho0_b2 * t - b2;
      fc2 = xj - sqrt(3) * rho0_c2 * t - c2;
    }

    rho0_x1 = 1 + 0.5 * sin(2 * pi * x1);
    rho0_x2 = 1 + 0.5 * sin(2 * pi * x2);

    rho_ex[cell_id] = 0.5 * (rho0_x1 + rho0_x2);
    u_ex[cell_id]   = sqrt(3) * (rho_ex[cell_id] - rho0_x1);
  }

  return std::make_tuple(std::make_shared<DiscreteFunctionVariant>(rho_ex),
                         std::make_shared<DiscreteFunctionVariant>(u_ex));
}

std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticExactSolution(const double& t,
                             const std::shared_ptr<const DiscreteFunctionVariant>& rho_ex,
                             const std::shared_ptr<const DiscreteFunctionVariant>& u_ex)
{
  const std::shared_ptr<const MeshVariant> mesh_v = getCommonMesh({u_ex, rho_ex});
  if (mesh_v.use_count() == 0) {
    throw NormalError("rho_ex and u_ex have to be defined on the same mesh");
  }

  return std::visit(
    [&](auto&& p_mesh)
      -> std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>> {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      if constexpr (is_polygonal_mesh_v<MeshType> and (MeshType::Dimension == 1)) {
        return getEulerKineticExactSolution(p_mesh, t);
      } else {
        throw NotImplementedError("Invalid mesh type for Scalar Kinetic solver");
      }
    },
    mesh_v->variant());
}

template <MeshConcept MeshType>
std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticWaves(const std::shared_ptr<const MeshType> p_mesh,
                     const std::vector<double> lambda_vector,
                     DiscreteFunctionP0<const double> rho,
                     DiscreteFunctionP0<const double> rho_u,
                     DiscreteFunctionP0<const double> rho_E,
                     const double gamma)
{
  if (lambda_vector.size() < 2) {
    throw NormalError("lambda vector must be of dimension greater than 2");
  }

  DiscreteFunctionP0Vector<double> Fn_rho(p_mesh, lambda_vector.size());
  DiscreteFunctionP0Vector<double> Fn_rho_u(p_mesh, lambda_vector.size());
  DiscreteFunctionP0Vector<double> Fn_rho_E(p_mesh, lambda_vector.size());

  CellArray<double> vec_A{p_mesh->connectivity(), 3};
  CellValue<double> p{p_mesh->connectivity()};

  for (CellId cell_id = 0; cell_id < p_mesh->numberOfCells(); ++cell_id) {
    double rho_u_2    = (rho_u[cell_id] * rho_u[cell_id]) / rho[cell_id];
    p[cell_id]        = (gamma - 1) * (rho_E[cell_id] - 0.5 * rho_u_2);
    vec_A[cell_id][0] = rho_u[cell_id];
    vec_A[cell_id][1] = rho_u_2 + p[cell_id];
    vec_A[cell_id][2] = rho_u[cell_id] / rho[cell_id] * (rho_E[cell_id] * p[cell_id]);
  }

  const size_t nb_waves           = lambda_vector.size();
  double inv_L2norm_lambda_vector = 0;
  for (size_t i = 0; i < nb_waves; ++i) {
    inv_L2norm_lambda_vector += std::pow(lambda_vector[i], 2);
  }
  inv_L2norm_lambda_vector = 1. / inv_L2norm_lambda_vector;

  for (CellId cell_id = 0; cell_id < p_mesh->numberOfCells(); ++cell_id) {
    for (size_t i = 0; i < nb_waves; ++i) {
      Fn_rho[cell_id][i] =
        (1. / nb_waves) * rho[cell_id] + inv_L2norm_lambda_vector * vec_A[cell_id][0] * lambda_vector[i];
      Fn_rho_u[cell_id][i] =
        (1. / nb_waves) * rho_u[cell_id] + inv_L2norm_lambda_vector * vec_A[cell_id][1] * lambda_vector[i];
      Fn_rho_E[cell_id][i] =
        (1. / nb_waves) * rho_E[cell_id] + inv_L2norm_lambda_vector * vec_A[cell_id][2] * lambda_vector[i];
    }
  }

  return std::make_tuple(std::make_shared<DiscreteFunctionVariant>(Fn_rho),
                         std::make_shared<DiscreteFunctionVariant>(Fn_rho_u),
                         std::make_shared<DiscreteFunctionVariant>(Fn_rho_E));
}

std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticWaves(const std::vector<double>& lambda_vector,
                     const std::shared_ptr<const DiscreteFunctionVariant>& rho,
                     const std::shared_ptr<const DiscreteFunctionVariant>& rho_u,
                     const std::shared_ptr<const DiscreteFunctionVariant>& rho_E,
                     const double& gamma)
{
  const std::shared_ptr<const MeshVariant> mesh_v = getCommonMesh({rho, rho_u, rho_E});
  if (mesh_v.use_count() == 0) {
    throw NormalError("rho rho_u and rho_E have to be defined on the same mesh");
  }

  return std::visit(
    [&](auto&& p_mesh)
      -> std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                    std::shared_ptr<const DiscreteFunctionVariant>> {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      if constexpr (is_polygonal_mesh_v<MeshType> and (MeshType::Dimension == 1)) {
        auto rho_h   = rho->get<DiscreteFunctionP0<const double>>();
        auto rho_u_h = rho_u->get<DiscreteFunctionP0<const double>>();
        auto rho_E_h = rho_E->get<DiscreteFunctionP0<const double>>();

        return getEulerKineticWaves(p_mesh, lambda_vector, rho_h, rho_u_h, rho_E_h, gamma);
      } else {
        throw NotImplementedError("Invalid mesh type for Scalar Kinetic solver");
      }
    },
    mesh_v->variant());
}

template <MeshConcept MeshType>
class EulerKineticSolver
{
 private:
  constexpr static size_t Dimension = MeshType::Dimension;
  using Rd                          = TinyVector<Dimension>;

  using DiscreteEulerFunction = DiscreteFunctionP0<double>;

  const double m_dt;
  const std::vector<double>& m_lambda_vector;
  const DiscreteFunctionP0Vector<const double> m_Fn_rho;
  const DiscreteFunctionP0Vector<const double> m_Fn_rho_u;
  const DiscreteFunctionP0Vector<const double> m_Fn_rho_E;
  const double m_gamma;
  const std::shared_ptr<const MeshType> p_mesh;
  const MeshType& m_mesh;
  const CellValue<const double> m_dx_table          = MeshDataManager::instance().getMeshData(m_mesh).Vj();
  const CellValue<const TinyVector<Dimension>> m_xj = MeshDataManager::instance().getMeshData(m_mesh).xj();
  const NodeValue<const TinyVector<Dimension>> m_xr = m_mesh.xr();

 public:
  CellArray<const double>
  getA(const CellValue<const double>& rho,
       const CellValue<const double>& rho_u,
       const CellValue<const double>& rho_E) const
  {
    CellArray<double> vec_A{m_mesh.connectivity(), 3};
    CellValue<double> p{m_mesh.connectivity()};
    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      double rho_u_2    = (rho_u[cell_id] * rho_u[cell_id]) / rho[cell_id];
      p[cell_id]        = (m_gamma - 1) * (rho_E[cell_id] - 0.5 * rho_u_2);
      vec_A[cell_id][0] = rho_u[cell_id];
      vec_A[cell_id][1] = rho_u_2 + p[cell_id];
      vec_A[cell_id][2] = rho_u[cell_id] / rho[cell_id] * (rho_E[cell_id] + p[cell_id]);
    }

    return vec_A;
  }

  CellArray<const double>
  compute_M(const CellValue<const double>& u, const CellValue<const double> Au) const
  {
    const size_t nb_waves = m_lambda_vector.size();
    CellArray<double> M{m_mesh.connectivity(), nb_waves};

    double inv_L2norm_lambda_vector = 0;
    for (size_t i = 0; i < nb_waves; ++i) {
      inv_L2norm_lambda_vector += pow(m_lambda_vector[i], 2);
    }
    inv_L2norm_lambda_vector = 1. / inv_L2norm_lambda_vector;

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < nb_waves; ++i) {
        M[cell_id][i] = (1. / nb_waves) * u[cell_id] + inv_L2norm_lambda_vector * Au[cell_id] * m_lambda_vector[i];
      }
    }
    return M;
  }

  DiscreteFunctionP0Vector<double>
  compute_delta1Fn(const DiscreteFunctionP0Vector<const double>& Fn)
  {
    const size_t nb_waves = m_lambda_vector.size();

    DiscreteFunctionP0Vector<double> deltaFn(p_mesh, nb_waves);
    const size_t last_cell     = m_mesh.numberOfCells() - 1;
    const CellId last_cell_id  = last_cell;
    const CellId first_cell_id = 0;

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      const CellId prev_cell_id = cell_id - 1;
      const CellId next_cell_id = cell_id + 1;
      const double dx           = m_dx_table[cell_id];
      for (size_t i = 0; i < nb_waves; ++i) {
        if (m_lambda_vector[i] < 0) {
          if (cell_id == last_cell_id) {
            double last_dx      = 0.5 * (dx + m_dx_table[first_cell_id]);
            deltaFn[cell_id][i] = 1. / (last_dx) * (Fn[first_cell_id][i] - Fn[cell_id][i]);

          } else {
            double xj   = m_xj[cell_id][0];
            double xjp1 = m_xj[next_cell_id][0];

            deltaFn[cell_id][i] = 1. / (xjp1 - xj) * (Fn[cell_id + 1][i] - Fn[cell_id][i]);
          }

        } else if (m_lambda_vector[i] > 0) {
          if (cell_id == first_cell_id) {
            double first_dx     = 0.5 * (dx + m_dx_table[last_cell_id]);
            deltaFn[cell_id][i] = 1. / (first_dx) * (Fn[cell_id][i] - Fn[last_cell_id][i]);
          } else {
            double xj   = m_xj[cell_id][0];
            double xjm1 = m_xj[prev_cell_id][0];

            deltaFn[cell_id][i] = 1. / (xj - xjm1) * (Fn[cell_id][i] - Fn[cell_id - 1][i]);
          }
        } else {
          deltaFn[cell_id][i] = 0;
        }
      }
    }

    return deltaFn;
  }

  DiscreteFunctionP0Vector<double>
  compute_delta2Fn(const DiscreteFunctionP0Vector<const double>& Fn)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    DiscreteFunctionP0Vector<double> deltaFn(p_mesh, nb_waves);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < nb_waves; ++i) {
        const NodeId first_node_id = 0;
        const NodeId last_node_id  = m_mesh.numberOfNodes() - 1;

        const double xmin = m_xr[first_node_id][0];
        const double xmax = m_xr[last_node_id][0];

        const size_t N = m_mesh.numberOfCells();

        const auto& cell_to_node   = cell_to_node_matrix[cell_id];
        const NodeId left_node_id  = cell_to_node[0];
        const NodeId right_node_id = cell_to_node[1];

        const CellId prev2_cell_id = (cell_id + N - 2) % N;
        const CellId prev_cell_id  = (cell_id + N - 1) % N;
        const CellId next_cell_id  = (cell_id + 1) % N;
        const CellId next2_cell_id = (cell_id + 2) % N;

        const double xjm2 = (-1 * (cell_id < 2) * (xmax - xmin)) + m_xj[prev2_cell_id][0];
        const double xjm1 = (-1 * (cell_id < 1) * (xmax - xmin)) + m_xj[prev_cell_id][0];
        const double xj   = m_xj[cell_id][0];
        const double xjp1 = ((cell_id + 1 >= N) * (xmax - xmin)) + m_xj[next_cell_id][0];
        const double xjp2 = ((cell_id + 2 >= N) * (xmax - xmin)) + m_xj[next2_cell_id][0];

        if (m_lambda_vector[i] > 0) {
          deltaFn[cell_id][i] = Fn[prev2_cell_id][i] * (xj - xjm1) / ((xj - xjm2) * (xjm1 - xjm2)) +
                                Fn[prev_cell_id][i] * (-(xj - xjm2) / ((xj - xjm1) * (xjm1 - xjm2))) +
                                Fn[cell_id][i] * (2 * (xj - xjm1) + (xjm1 - xjm2)) / ((xj - xjm1) * (xj - xjm2));
        } else if (m_lambda_vector[i] < 0) {
          deltaFn[cell_id][i] = Fn[cell_id][i] * (1 / (xj - xjp1) + 1 / (xj - xjp2)) +
                                Fn[next_cell_id][i] * (xj - xjp2) / ((xjp1 - xj) * (xjp1 - xjp2)) +
                                Fn[next2_cell_id][i] * (xj - xjp1) / ((xjp2 - xj) * (xjp2 - xjp1));
        } else {
          deltaFn[cell_id][i] = 0;
        }
      }
    }

    return deltaFn;
  }

  DiscreteFunctionP0Vector<double>
  compute_delta3Fn(const DiscreteFunctionP0Vector<const double>& Fn)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    DiscreteFunctionP0Vector<double> deltaFn(p_mesh, nb_waves);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < nb_waves; ++i) {
        const NodeId first_node_id = 0;
        const NodeId last_node_id  = m_mesh.numberOfNodes() - 1;

        const double xmin = m_xr[first_node_id][0];
        const double xmax = m_xr[last_node_id][0];

        const size_t N = m_mesh.numberOfCells();

        const auto& cell_to_node   = cell_to_node_matrix[cell_id];
        const NodeId left_node_id  = cell_to_node[0];
        const NodeId right_node_id = cell_to_node[1];

        const CellId prev2_cell_id = (cell_id + N - 2) % N;
        const CellId prev_cell_id  = (cell_id + N - 1) % N;
        const CellId next_cell_id  = (cell_id + 1) % N;
        const CellId next2_cell_id = (cell_id + 2) % N;

        const double xjm2 = (-1 * (cell_id < 2) * (xmax - xmin)) + m_xj[prev2_cell_id][0];
        const double xjm1 = (-1 * (cell_id < 1) * (xmax - xmin)) + m_xj[prev_cell_id][0];
        const double xj   = m_xj[cell_id][0];
        const double xjp1 = ((cell_id + 1 >= N) * (xmax - xmin)) + m_xj[next_cell_id][0];
        const double xjp2 = ((cell_id + 2 >= N) * (xmax - xmin)) + m_xj[next2_cell_id][0];

        if (m_lambda_vector[i] > 0) {
          deltaFn[cell_id][i] =
            Fn[prev2_cell_id][i] * (xj - xjp1) * (xj - xjm1) / ((xjm2 - xjp1) * (xjm2 - xj) * (xjm2 - xjm1)) +
            Fn[prev_cell_id][i] * (xj - xjp1) * (xj - xjm2) / ((xjm1 - xjp1) * (xjm1 - xj) * (xjm1 - xjm2)) +
            Fn[cell_id][i] * (1. / (xj - xjm2) + 1. / (xj - xjm1) + 1. / (xj - xjp1)) +
            Fn[next_cell_id][i] * (xj - xjm1) * (xj - xjm2) / ((xjp1 - xj) * (xjp1 - xjm1) * (xjp1 - xjm2));

        } else if (m_lambda_vector[i] < 0) {
          deltaFn[cell_id][i] =
            Fn[prev_cell_id][i] * (xj - xjp1) * (xj - xjp2) / ((xjm1 - xj) * (xjm1 - xjp1) * (xjm1 - xjp2)) +
            Fn[cell_id][i] * (1. / (xj - xjp2) + 1. / (xj - xjp1) + 1. / (xj - xjm1)) +
            Fn[next_cell_id][i] * (xj - xjm1) * (xj - xjp2) / ((xjp1 - xjm1) * (xjp1 - xj) * (xjp1 - xjp2)) +
            Fn[next2_cell_id][i] * (xj - xjm1) * (xj - xjp1) / ((xjp2 - xjm1) * (xjp2 - xj) * (xjp2 - xjp1));

        } else {
          deltaFn[cell_id][i] = 0;
        }
      }
    }
    return deltaFn;
  }

  DiscreteFunctionP0Vector<double>
  compute_delta4Fn(const DiscreteFunctionP0Vector<const double>& Fn)
  {
    const size_t nb_waves = m_lambda_vector.size();

    DiscreteFunctionP0Vector<double> deltaFn(p_mesh, nb_waves);
    // const size_t last_cell     =
    const CellId last_cell_id = m_mesh.numberOfCells() - 1;
    ;   // last_cell;
    const CellId first_cell_id = 0;

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < nb_waves; ++i) {
        // deltaFn[cell_id][i] = -(1. / 12) * (Fn[(cell_id + 2) % m_mesh.numberOfCells()][i] -
        //                                     Fn[(cell_id + m_mesh.numberOfCells() - 2) %
        //                                     m_mesh.numberOfCells()][i]) +
        //                       (2. / 3) * (Fn[(cell_id + 1) % m_mesh.numberOfCells()][i] -
        //                                   Fn[(cell_id + m_mesh.numberOfCells() - 1) %
        //                                   m_mesh.numberOfCells()][i]);
        if (m_lambda_vector[i] < 0) {
          if (cell_id == first_cell_id) {
            deltaFn[cell_id][i] = 1. / 12 * Fn[cell_id + 3][i] - 0.5 * Fn[cell_id + 2][i] +
                                  3. / 2 * Fn[cell_id + 1][i] - 5. / 6 * Fn[cell_id][i] - 0.25 * Fn[last_cell_id][i];
          } else if (cell_id == last_cell_id) {
            deltaFn[cell_id][i] = 1. / 12 * Fn[first_cell_id + 2][i] - 0.5 * Fn[first_cell_id + 1][i] +
                                  3. / 2 * Fn[first_cell_id][i] - 5. / 6 * Fn[cell_id][i] - 0.25 * Fn[cell_id - 1][i];
          } else if (cell_id + 1 == last_cell_id) {
            deltaFn[cell_id][i] = 1. / 12 * Fn[first_cell_id + 1][i] - 0.5 * Fn[first_cell_id][i] +
                                  3. / 2 * Fn[cell_id + 1][i] - 5. / 6 * Fn[cell_id][i] - 0.25 * Fn[cell_id - 1][i];
          } else if (cell_id + 2 == last_cell_id) {
            deltaFn[cell_id][i] = 1. / 12 * Fn[first_cell_id][i] - 0.5 * Fn[cell_id + 2][i] +
                                  3. / 2 * Fn[cell_id + 1][i] - 5. / 6 * Fn[cell_id][i] - 0.25 * Fn[cell_id - 1][i];
          } else {
            deltaFn[cell_id][i] = 1. / 12 * Fn[cell_id + 3][i] - 0.5 * Fn[cell_id + 2][i] +
                                  3. / 2 * Fn[cell_id + 1][i] - 5. / 6 * Fn[cell_id][i] - 0.25 * Fn[cell_id - 1][i];
          }

        } else if (m_lambda_vector[i] > 0) {
          if (cell_id == last_cell_id) {
            deltaFn[cell_id][i] = 0.25 * Fn[first_cell_id][i] + 5. / 6 * Fn[cell_id][i] - 3. / 2 * Fn[cell_id - 1][i] +
                                  0.5 * Fn[cell_id - 2][i] - 1. / 12 * Fn[cell_id - 3][i];
          } else if (cell_id == first_cell_id) {
            deltaFn[cell_id][i] = 0.25 * Fn[cell_id + 1][i] + 5. / 6 * Fn[cell_id][i] - 3. / 2 * Fn[last_cell_id][i] +
                                  0.5 * Fn[last_cell_id - 1][i] - 1. / 12 * Fn[last_cell_id - 2][i];
          } else if (cell_id - 1 == first_cell_id) {
            deltaFn[cell_id][i] = 0.25 * Fn[cell_id + 1][i] + 5. / 6 * Fn[cell_id][i] - 3. / 2 * Fn[cell_id - 1][i] +
                                  0.5 * Fn[last_cell_id][i] - 1. / 12 * Fn[last_cell_id - 1][i];
          } else if (cell_id - 2 == first_cell_id) {
            deltaFn[cell_id][i] = 0.25 * Fn[cell_id + 1][i] + 5. / 6 * Fn[cell_id][i] - 3. / 2 * Fn[cell_id - 1][i] +
                                  0.5 * Fn[cell_id - 2][i] - 1. / 12 * Fn[last_cell_id][i];
          } else {
            deltaFn[cell_id][i] = 0.25 * Fn[cell_id + 1][i] + 5. / 6 * Fn[cell_id][i] - 3. / 2 * Fn[cell_id - 1][i] +
                                  0.5 * Fn[cell_id - 2][i] - 1. / 12 * Fn[cell_id - 3][i];
          }
        } else {
          deltaFn[cell_id][i] = 0;
        }
      }
    }

    return deltaFn;
  }

  CellValue<double>
  compute_PFnp1_rho(const DiscreteFunctionP0Vector<const double> F_rho, const double& dt)
  {
    CellValue<double> PFnp1{m_mesh.connectivity()};

    DiscreteFunctionP0Vector<double> deltaF  = compute_delta2Fn(F_rho);
    DiscreteFunctionP0Vector<double> deltaF0 = compute_delta2Fn(m_Fn_rho);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      // const double dx = m_dx_table[cell_id];
      //  const double c  = 0.5 * dt / dx;
      const double c = 0.5 * dt;
      double PFn     = 0;
      for (size_t i = 0; i < m_lambda_vector.size(); ++i) {
        PFn += m_Fn_rho[cell_id][i] - c * m_lambda_vector[i] * (deltaF[cell_id][i] + deltaF0[cell_id][i]);
      }

      PFnp1[cell_id] = PFn;
    }
    return PFnp1;
  }

  CellValue<double>
  compute_PFnp1_rho_u(const DiscreteFunctionP0Vector<const double> F_rho_u, const double& dt)
  {
    CellValue<double> PFnp1{m_mesh.connectivity()};

    DiscreteFunctionP0Vector<double> deltaF  = compute_delta2Fn(F_rho_u);
    DiscreteFunctionP0Vector<double> deltaF0 = compute_delta2Fn(m_Fn_rho_u);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      // const double dx = m_dx_table[cell_id];
      // const double c  = 0.5 * dt / dx;
      const double c = 0.5 * dt;
      double PFn     = 0;
      for (size_t i = 0; i < m_lambda_vector.size(); ++i) {
        PFn += m_Fn_rho_u[cell_id][i] - c * m_lambda_vector[i] * (deltaF[cell_id][i] + deltaF0[cell_id][i]);
      }

      PFnp1[cell_id] = PFn;
    }
    return PFnp1;
  }

  CellValue<double>
  compute_PFnp1_rho_E(const DiscreteFunctionP0Vector<const double> F_rho_E, const double& dt)
  {
    CellValue<double> PFnp1{m_mesh.connectivity()};

    DiscreteFunctionP0Vector<double> deltaF  = compute_delta2Fn(F_rho_E);
    DiscreteFunctionP0Vector<double> deltaF0 = compute_delta2Fn(m_Fn_rho_E);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      // const double dx = m_dx_table[cell_id];
      // const double c  = 0.5 * dt / dx;
      const double c = 0.5 * dt;
      double PFn     = 0;
      for (size_t i = 0; i < m_lambda_vector.size(); ++i) {
        PFn += m_Fn_rho_E[cell_id][i] - c * m_lambda_vector[i] * (deltaF[cell_id][i] + deltaF0[cell_id][i]);
      }

      PFnp1[cell_id] = PFn;
    }
    return PFnp1;
  }

  CellValue<double>
  compute_PFn(const DiscreteFunctionP0Vector<const double> F)
  {
    CellValue<double> PFn{m_mesh.connectivity()};

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      double PF = 0;
      for (size_t i = 0; i < m_lambda_vector.size(); ++i) {
        PF += F[cell_id][i];
      }
      PFn[cell_id] = PF;
    }
    return PFn;
  }

  std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
             std::shared_ptr<const DiscreteFunctionVariant>,
             std::shared_ptr<const DiscreteFunctionVariant>>
  apply(const double& eps)
  {
    // for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
    //   std::cout << "CFL_j = " << m_lambda_vector[0] * m_dt / m_dx_table[cell_id] << "\n";
    // }

    const DiscreteFunctionP0Vector<const double>& F0_rho   = m_Fn_rho;
    const DiscreteFunctionP0Vector<const double>& F0_rho_u = m_Fn_rho_u;
    const DiscreteFunctionP0Vector<const double>& F0_rho_E = m_Fn_rho_E;

    DiscreteFunctionP0Vector<double> Fnp1_rho   = copy(F0_rho);
    DiscreteFunctionP0Vector<double> Fnp1_rho_u = copy(F0_rho_u);
    DiscreteFunctionP0Vector<double> Fnp1_rho_E = copy(F0_rho_E);

    DiscreteFunctionP0Vector<double> deltaFn_rho   = compute_delta2Fn(F0_rho);
    DiscreteFunctionP0Vector<double> deltaFn_rho_u = compute_delta2Fn(F0_rho_u);
    DiscreteFunctionP0Vector<double> deltaFn_rho_E = compute_delta2Fn(F0_rho_E);

    // for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
    //   std::cout << "D_rho = " << deltaFn_rho[cell_id] << "\n";
    // }

    DiscreteFunctionP0Vector<double> F_rho   = copy(F0_rho);
    DiscreteFunctionP0Vector<double> F_rho_u = copy(F0_rho_u);
    DiscreteFunctionP0Vector<double> F_rho_E = copy(F0_rho_E);

    for (size_t p = 0; p < 2; ++p) {
      DiscreteFunctionP0Vector<double> deltaFnp1_rho   = compute_delta2Fn(Fnp1_rho);
      DiscreteFunctionP0Vector<double> deltaFnp1_rho_u = compute_delta2Fn(Fnp1_rho_u);
      DiscreteFunctionP0Vector<double> deltaFnp1_rho_E = compute_delta2Fn(Fnp1_rho_E);

      // double sum_delta = 0;
      // for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      //   sum_delta += deltaFnp1_rho[cell_id][2];
      // }
      // std::cout << "sum delta = " << sum_delta << "\n";

      const CellValue<const double> PFnp1_rho   = compute_PFnp1_rho(Fnp1_rho, m_dt);
      const CellValue<const double> PFnp1_rho_u = compute_PFnp1_rho_u(Fnp1_rho_u, m_dt);
      const CellValue<const double> PFnp1_rho_E = compute_PFnp1_rho_E(Fnp1_rho_E, m_dt);

      const CellValue<const double> PFn_rho   = compute_PFn(F0_rho);
      const CellValue<const double> PFn_rho_u = compute_PFn(F0_rho_u);
      const CellValue<const double> PFn_rho_E = compute_PFn(F0_rho_E);

      const CellArray<const double> APFnp1 = getA(PFnp1_rho, PFnp1_rho_u, PFnp1_rho_E);
      const CellArray<const double> APFn   = getA(PFn_rho, PFn_rho_u, PFn_rho_E);

      const CellValue<double> APFnp1_rho{m_mesh.connectivity()};
      const CellValue<double> APFnp1_rho_u{m_mesh.connectivity()};
      const CellValue<double> APFnp1_rho_E{m_mesh.connectivity()};

      const CellValue<double> APFn_rho{m_mesh.connectivity()};
      const CellValue<double> APFn_rho_u{m_mesh.connectivity()};
      const CellValue<double> APFn_rho_E{m_mesh.connectivity()};

      for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
        APFnp1_rho[cell_id]   = APFnp1[cell_id][0];
        APFnp1_rho_u[cell_id] = APFnp1[cell_id][1];
        APFnp1_rho_E[cell_id] = APFnp1[cell_id][2];
        APFn_rho[cell_id]     = APFn[cell_id][0];
        APFn_rho_u[cell_id]   = APFn[cell_id][1];
        APFn_rho_E[cell_id]   = APFn[cell_id][2];
      }

      const CellArray<const double> MPFnp1_rho   = compute_M(PFnp1_rho, APFnp1_rho);
      const CellArray<const double> MPFnp1_rho_u = compute_M(PFnp1_rho_u, APFnp1_rho_u);
      const CellArray<const double> MPFnp1_rho_E = compute_M(PFnp1_rho_E, APFnp1_rho_E);
      const CellArray<const double> MPFn_rho     = compute_M(PFn_rho, APFn_rho);
      const CellArray<const double> MPFn_rho_u   = compute_M(PFn_rho_u, APFn_rho_u);
      const CellArray<const double> MPFn_rho_E   = compute_M(PFn_rho_E, APFn_rho_E);

      const double c1 = 1. / (2 * eps + m_dt);
      for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
        // const double dx = m_dx_table[cell_id];
        // const double c2 = (eps * m_dt / dx);
        const double c2 = (eps * m_dt);
        for (size_t i = 0; i < m_lambda_vector.size(); ++i) {
          Fnp1_rho[cell_id][i] = c1 * (m_dt * MPFnp1_rho[cell_id][i] + 2 * eps * F0_rho[cell_id][i] -
                                       c2 * m_lambda_vector[i] * (deltaFnp1_rho[cell_id][i] + deltaFn_rho[cell_id][i]) +
                                       m_dt * (MPFn_rho[cell_id][i] - F0_rho[cell_id][i]));
          Fnp1_rho_u[cell_id][i] =
            c1 * (m_dt * MPFnp1_rho_u[cell_id][i] + 2 * eps * F0_rho_u[cell_id][i] -
                  c2 * m_lambda_vector[i] * (deltaFnp1_rho_u[cell_id][i] + deltaFn_rho_u[cell_id][i]) +
                  m_dt * (MPFn_rho_u[cell_id][i] - F0_rho_u[cell_id][i]));
          Fnp1_rho_E[cell_id][i] =
            c1 * (m_dt * MPFnp1_rho_E[cell_id][i] + 2 * eps * F0_rho_E[cell_id][i] -
                  c2 * m_lambda_vector[i] * (deltaFnp1_rho_E[cell_id][i] + deltaFn_rho_E[cell_id][i]) +
                  m_dt * (MPFn_rho_E[cell_id][i] - F0_rho_E[cell_id][i]));
        }
      }
      // std::cout << m_lambda_vector.size() << "; " << Fnp1_rho_u.cellArrays() << "\n";
    }

    return std::make_tuple(std::make_shared<DiscreteFunctionVariant>(Fnp1_rho),
                           std::make_shared<DiscreteFunctionVariant>(Fnp1_rho_u),
                           std::make_shared<DiscreteFunctionVariant>(Fnp1_rho_E));
  }

  EulerKineticSolver(std::shared_ptr<const MeshType> mesh,
                     const double& dt,
                     const double& gamma,
                     const std::vector<double>& lambda_vector,
                     const DiscreteFunctionP0Vector<const double>& Fn_rho,
                     const DiscreteFunctionP0Vector<const double>& Fn_rho_u,
                     const DiscreteFunctionP0Vector<const double>& Fn_rho_E)
    : m_dt{dt},
      m_lambda_vector{lambda_vector},
      m_Fn_rho{Fn_rho},
      m_Fn_rho_u{Fn_rho_u},
      m_Fn_rho_E{Fn_rho_E},
      m_gamma{gamma},
      p_mesh{mesh},
      m_mesh{*mesh}
  {
    if ((lambda_vector.size() != Fn_rho.size()) or
        (lambda_vector.size() != Fn_rho_u.size() or (lambda_vector.size() != Fn_rho_u.size()))) {
      throw NormalError("Incompatible dimensions of lambda vector and Fn");
    }
  }

  ~EulerKineticSolver() = default;
};

std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
eulerKineticSolver(const double& dt,
                   const double& gamma,
                   const std::vector<double>& lambda_vector,
                   const double& eps,
                   const std::shared_ptr<const DiscreteFunctionVariant>& F_rho,
                   const std::shared_ptr<const DiscreteFunctionVariant>& F_rho_u,
                   const std::shared_ptr<const DiscreteFunctionVariant>& F_rho_E)
{
  const std::shared_ptr<const MeshVariant> mesh_v = getCommonMesh({F_rho, F_rho_u, F_rho_E});
  if (mesh_v.use_count() == 0) {
    throw NormalError("F_rho, F_rho_u and F_rho_E have to be defined on the same mesh");
  }

  return std::visit(
    [&](auto&& p_mesh)
      -> std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                    std::shared_ptr<const DiscreteFunctionVariant>> {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      if constexpr (is_polygonal_mesh_v<MeshType> and (MeshType::Dimension == 1)) {
        auto Fn_rho   = F_rho->get<DiscreteFunctionP0Vector<const double>>();
        auto Fn_rho_u = F_rho_u->get<DiscreteFunctionP0Vector<const double>>();
        auto Fn_rho_E = F_rho_E->get<DiscreteFunctionP0Vector<const double>>();

        EulerKineticSolver solver(p_mesh, dt, gamma, lambda_vector, Fn_rho, Fn_rho_u, Fn_rho_E);

        return solver.apply(eps);
      } else {
        throw NotImplementedError("Invalid mesh type for Scalar Kinetic solver");
      }
    },
    mesh_v->variant());
}
