#include <scheme/EulerKineticSolverAcousticLagrangeFV.hpp>

#include <language/utils/EvaluateAtPoints.hpp>
#include <language/utils/InterpolateItemValue.hpp>   //new
#include <mesh/Connectivity.hpp>
#include <mesh/Mesh.hpp>
#include <mesh/MeshVariant.hpp>
#include <scheme/DiscreteFunctionDescriptorP0Vector.hpp>
#include <scheme/DiscreteFunctionP0.hpp>   //new
#include <scheme/DiscreteFunctionP0Vector.hpp>
#include <scheme/DiscreteFunctionUtils.hpp>
#include <scheme/IDiscreteFunctionDescriptor.hpp>

#include <analysis/GaussLegendreQuadratureDescriptor.hpp>
#include <analysis/QuadratureManager.hpp>
#include <geometry/LineTransformation.hpp>
#include <tuple>

template <MeshConcept MeshType>
std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticExactSolutionAcousticLagrangeFV(const std::shared_ptr<const MeshType> p_mesh,
                                               const double t,
                                               const double gamma)
{
  DiscreteFunctionP0<double> rho_ex(p_mesh);
  DiscreteFunctionP0<TinyVector<MeshType::Dimension>> u_ex(p_mesh);
  DiscreteFunctionP0<double> p_ex(p_mesh);
  DiscreteFunctionP0<TinyVector<MeshType::Dimension>> rho_u_ex(p_mesh);
  DiscreteFunctionP0<double> rho_E_ex(p_mesh);

  MeshData<MeshType>& mesh_data                               = MeshDataManager::instance().getMeshData(*p_mesh);
  const auto& cell_to_node_matrix                             = p_mesh->connectivity().cellToNodeMatrix();
  const CellValue<const TinyVector<MeshType::Dimension>> m_xj = mesh_data.xj();
  const NodeValue<const TinyVector<MeshType::Dimension>> m_xr = p_mesh->xr();
  const CellValue<const double> m_Vj                          = mesh_data.Vj();

  const double pi = acos(-1);
  // EvaluateAtPoints<double(double)>::evaluateTo(density_function_id, x, rho0);

  for (CellId cell_id = 0; cell_id < p_mesh->numberOfCells(); ++cell_id) {
    const auto& cell_to_node   = cell_to_node_matrix[cell_id];
    const NodeId left_node_id  = cell_to_node[0];
    const NodeId right_node_id = cell_to_node[1];
    const double dx            = m_Vj[cell_id];

    const LineTransformation<1> line_transformation(m_xr[left_node_id], m_xr[right_node_id]);
    QuadratureFormula<1> quadrature =
      QuadratureManager::instance().getLineFormula(GaussLegendreQuadratureDescriptor(4));

    rho_ex[cell_id]      = 0;
    u_ex[cell_id][0]     = 0;
    p_ex[cell_id]        = 0;
    rho_u_ex[cell_id][0] = 0;
    rho_E_ex[cell_id]    = 0;

    for (size_t i_gauss = 0; i_gauss < quadrature.numberOfPoints(); ++i_gauss) {
      double gauss_point = line_transformation(quadrature.point(i_gauss))[0];

      double a1      = -1.5;
      double b1      = 1.5;
      double c1      = 0.5 * (a1 + b1);
      double x1      = c1;
      double rho0_x1 = 1 + 0.5 * std::sin(2 * pi * x1);
      double rho0_a1 = 1 + 0.5 * std::sin(2 * pi * a1);
      double rho0_b1 = 1 + 0.5 * std::sin(2 * pi * b1);
      double rho0_c1 = 1 + 0.5 * std::sin(2 * pi * c1);

      double fx1 = gauss_point + std::sqrt(3) * rho0_x1 * t - x1;
      double fa1 = gauss_point + std::sqrt(3) * rho0_a1 * t - a1;
      double fb1 = gauss_point + std::sqrt(3) * rho0_b1 * t - b1;
      double fc1 = gauss_point + std::sqrt(3) * rho0_c1 * t - c1;

      while (std::abs(fx1) > 1E-15) {
        if (fa1 * fc1 < 0) {
          b1 = c1;
          c1 = 0.5 * (a1 + b1);
        } else if (fc1 * fb1 < 0) {
          a1 = c1;
          c1 = 0.5 * (a1 + b1);
        }
        x1 = c1;

        rho0_x1 = 1 + 0.5 * std::sin(2 * pi * x1);
        rho0_a1 = 1 + 0.5 * std::sin(2 * pi * a1);
        rho0_b1 = 1 + 0.5 * std::sin(2 * pi * b1);
        rho0_c1 = 1 + 0.5 * std::sin(2 * pi * c1);

        fx1 = gauss_point + std::sqrt(3) * rho0_x1 * t - x1;
        fa1 = gauss_point + std::sqrt(3) * rho0_a1 * t - a1;
        fb1 = gauss_point + std::sqrt(3) * rho0_b1 * t - b1;
        fc1 = gauss_point + std::sqrt(3) * rho0_c1 * t - c1;
      }

      double a2      = -1.5;
      double b2      = 1.5;
      double c2      = 0.5 * (a2 + b2);
      double x2      = c2;
      double rho0_x2 = 1 + 0.5 * std::sin(2 * pi * x2);
      double rho0_a2 = 1 + 0.5 * std::sin(2 * pi * a2);
      double rho0_b2 = 1 + 0.5 * std::sin(2 * pi * b2);
      double rho0_c2 = 1 + 0.5 * std::sin(2 * pi * c2);

      double fx2 = gauss_point - std::sqrt(3) * rho0_x2 * t - x2;
      double fa2 = gauss_point - std::sqrt(3) * rho0_a2 * t - a2;
      double fb2 = gauss_point - std::sqrt(3) * rho0_b2 * t - b2;
      double fc2 = gauss_point - std::sqrt(3) * rho0_c2 * t - c2;

      while (std::abs(fx2) > 1E-15) {
        if (fa2 * fc2 < 0) {
          b2 = c2;
          c2 = 0.5 * (a2 + b2);
        } else if (fc2 * fb2 < 0) {
          a2 = c2;
          c2 = 0.5 * (a2 + b2);
        }
        x2 = c2;

        rho0_x2 = 1 + 0.5 * std::sin(2 * pi * x2);
        rho0_a2 = 1 + 0.5 * std::sin(2 * pi * a2);
        rho0_b2 = 1 + 0.5 * std::sin(2 * pi * b2);
        rho0_c2 = 1 + 0.5 * std::sin(2 * pi * c2);

        fx2 = gauss_point - std::sqrt(3) * rho0_x2 * t - x2;
        fa2 = gauss_point - std::sqrt(3) * rho0_a2 * t - a2;
        fb2 = gauss_point - std::sqrt(3) * rho0_b2 * t - b2;
        fc2 = gauss_point - std::sqrt(3) * rho0_c2 * t - c2;
      }

      rho0_x1 = 1 + 0.5 * std::sin(2 * pi * x1);
      rho0_x2 = 1 + 0.5 * std::sin(2 * pi * x2);

      const double rho_gauss = 0.5 * (rho0_x1 + rho0_x2);
      TinyVector<MeshType::Dimension> u_gauss;
      u_gauss[0]           = std::sqrt(3) * ((0.5 * (rho0_x1 + rho0_x2)) - rho0_x1);
      const double p_gauss = (0.5 * (rho0_x1 + rho0_x2)) * (0.5 * (rho0_x1 + rho0_x2)) * (0.5 * (rho0_x1 + rho0_x2));

      rho_u_ex[cell_id] += quadrature.weight(i_gauss) * line_transformation.jacobianDeterminant() * rho_gauss * u_gauss;
      rho_E_ex[cell_id] += quadrature.weight(i_gauss) * line_transformation.jacobianDeterminant() *
                           (p_gauss / (gamma - 1) + 0.5 * rho_gauss * u_gauss[0] * u_gauss[0]);

      rho_ex[cell_id] += quadrature.weight(i_gauss) * line_transformation.jacobianDeterminant() * rho_gauss;
      u_ex[cell_id] += quadrature.weight(i_gauss) * line_transformation.jacobianDeterminant() * u_gauss;
      p_ex[cell_id] += quadrature.weight(i_gauss) * line_transformation.jacobianDeterminant() * p_gauss;
    }
    rho_u_ex[cell_id][0] /= dx;
    rho_E_ex[cell_id] /= dx;
    rho_ex[cell_id] /= dx;
    u_ex[cell_id][0] /= dx;
    p_ex[cell_id] /= dx;
  }

  return std::make_tuple(std::make_shared<DiscreteFunctionVariant>(rho_ex),
                         std::make_shared<DiscreteFunctionVariant>(u_ex),
                         std::make_shared<DiscreteFunctionVariant>(p_ex),
                         std::make_shared<DiscreteFunctionVariant>(rho_u_ex),
                         std::make_shared<DiscreteFunctionVariant>(rho_E_ex));
}

std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticExactSolutionAcousticLagrangeFV(const double& t,
                                               const double& gamma,
                                               const std::shared_ptr<const DiscreteFunctionVariant>& rho_ex,
                                               const std::shared_ptr<const DiscreteFunctionVariant>& u_ex,
                                               const std::shared_ptr<const DiscreteFunctionVariant>& p_ex,
                                               const std::shared_ptr<const DiscreteFunctionVariant>& rho_u_ex,
                                               const std::shared_ptr<const DiscreteFunctionVariant>& rho_E_ex)
{
  const std::shared_ptr<const MeshVariant> mesh_v = getCommonMesh({u_ex, rho_ex, p_ex, rho_u_ex, rho_E_ex});
  if (mesh_v.use_count() == 0) {
    throw NormalError("rho_ex, u_ex, p_ex, rho_u_ex and rho_E_ex have to be defined on the same mesh");
  }

  return std::visit(
    [&](auto&& p_mesh)
      -> std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                    std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                    std::shared_ptr<const DiscreteFunctionVariant>> {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      if constexpr (is_polygonal_mesh_v<MeshType> and (MeshType::Dimension == 1)) {
        return getEulerKineticExactSolutionAcousticLagrangeFV(p_mesh, t, gamma);
      } else {
        throw NotImplementedError("Invalid mesh type for Scalar Kinetic solver");
      }
    },
    mesh_v->variant());
}

template <MeshConcept MeshType>
std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticWavesAcousticLagrangeFV(const std::shared_ptr<const MeshType> p_mesh,
                                       const std::vector<double> lambda_vector,
                                       DiscreteFunctionP0<const double> p,
                                       DiscreteFunctionP0<const double> u)
{
  if (lambda_vector.size() < 2) {
    throw NormalError("lambda vector must be of dimension greater than 2");
  }

  MeshData<MeshType>& mesh_data                               = MeshDataManager::instance().getMeshData(*p_mesh);
  const NodeValue<const TinyVector<MeshType::Dimension>> m_xr = p_mesh->xr();
  const CellValue<const TinyVector<MeshType::Dimension>> m_xj = mesh_data.xj();
  const CellValue<const double> m_Vj                          = mesh_data.Vj();

  DiscreteFunctionP0Vector<double> Fn_p(p_mesh, lambda_vector.size());
  DiscreteFunctionP0Vector<double> Fn_u(p_mesh, lambda_vector.size());

  CellArray<double> vec_A{p_mesh->connectivity(), 3};

  for (CellId cell_id = 0; cell_id < p_mesh->numberOfCells(); ++cell_id) {
    vec_A[cell_id][0] = u[cell_id];
    vec_A[cell_id][1] = p[cell_id];
  }

  const size_t nb_waves           = lambda_vector.size();
  double inv_L2norm_lambda_vector = 0;
  for (size_t i = 0; i < nb_waves; ++i) {
    inv_L2norm_lambda_vector += std::pow(lambda_vector[i], 2);
  }
  inv_L2norm_lambda_vector = 1. / inv_L2norm_lambda_vector;

  for (CellId cell_id = 0; cell_id < p_mesh->numberOfCells(); ++cell_id) {
    for (size_t i = 0; i < nb_waves; ++i) {
      Fn_p[cell_id][i] = (1. / nb_waves) * p[cell_id] + inv_L2norm_lambda_vector * vec_A[cell_id][0] * lambda_vector[i];
      Fn_u[cell_id][i] = (1. / nb_waves) * u[cell_id] + inv_L2norm_lambda_vector * vec_A[cell_id][1] * lambda_vector[i];
    }
  }

  return std::make_tuple(std::make_shared<DiscreteFunctionVariant>(Fn_p),
                         std::make_shared<DiscreteFunctionVariant>(Fn_u));
}

std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticWavesAcousticLagrangeFV(const std::vector<double>& lambda_vector,
                                       const std::shared_ptr<const DiscreteFunctionVariant>& p,
                                       const std::shared_ptr<const DiscreteFunctionVariant>& u)
{
  const std::shared_ptr<const MeshVariant> mesh_v = getCommonMesh({p, u});
  if (mesh_v.use_count() == 0) {
    throw NormalError("p and u have to be defined on the same mesh");
  }

  return std::visit(
    [&](auto&& p_mesh)
      -> std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>> {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      if constexpr (is_polygonal_mesh_v<MeshType> and (MeshType::Dimension == 1)) {
        auto p_h = p->get<DiscreteFunctionP0<const double>>();
        auto u_h = u->get<DiscreteFunctionP0<const double>>();

        return getEulerKineticWavesAcousticLagrangeFV(p_mesh, lambda_vector, p_h, u_h);
      } else {
        throw NotImplementedError("Invalid mesh type for Scalar Kinetic solver");
      }
    },
    mesh_v->variant());
}

template <MeshConcept MeshType>
class EulerKineticSolverAcousticLagrangeFV
{
 private:
  constexpr static size_t Dimension = MeshType::Dimension;
  using Rd                          = TinyVector<Dimension>;

  const double m_dt;
  const std::vector<double> m_lambda_vector;
  const double m_gamma;
  const size_t m_spaceOrder;
  const size_t m_timeOrder;
  const DiscreteFunctionP0<const double> m_rho;
  const DiscreteFunctionP0<const TinyVector<Dimension>> m_u;
  const DiscreteFunctionP0<const double> m_E;
  const DiscreteFunctionP0<const double> m_p;
  const std::shared_ptr<const MeshType> p_mesh;
  const MeshType& m_mesh;
  const CellValue<const double> m_Vj = MeshDataManager::instance().getMeshData(m_mesh).Vj();
  CellValue<const double> m_inv_Mj;
  CellValue<const double> m_inv_Vj;
  CellValue<const double> m_Mj;
  const CellValue<const TinyVector<Dimension>> m_xj = MeshDataManager::instance().getMeshData(m_mesh).xj();
  const NodeValue<const TinyVector<Dimension>> m_xr = m_mesh.xr();

 public:
  CellArray<double>
  getA(const DiscreteFunctionP0<const double>& p, const DiscreteFunctionP0<const TinyVector<Dimension>>& u) const
  {
    CellArray<double> vec_A{m_mesh.connectivity(), 2};

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      vec_A[cell_id][0] = u[cell_id][0];
      vec_A[cell_id][1] = p[cell_id];
    }

    return vec_A;
  }

  std::tuple<DiscreteFunctionP0Vector<double>, DiscreteFunctionP0Vector<double>>
  compute_M(const DiscreteFunctionP0<const double>& p, const DiscreteFunctionP0<const TinyVector<Dimension>>& u)
  {
    const size_t nb_waves = m_lambda_vector.size();
    DiscreteFunctionP0Vector<double> Mp{p_mesh, nb_waves};
    DiscreteFunctionP0Vector<double> Mu{p_mesh, nb_waves};
    CellArray<double> A = getA(p, u);

    double inv_L2norm_lambda_vector = 0;
    for (size_t i = 0; i < nb_waves; ++i) {
      inv_L2norm_lambda_vector += m_lambda_vector[i] * m_lambda_vector[i];
    }
    inv_L2norm_lambda_vector = 1. / inv_L2norm_lambda_vector;

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < nb_waves; ++i) {
        Mp[cell_id][i] = (1. / nb_waves) * p[cell_id] + inv_L2norm_lambda_vector * A[cell_id][0] * m_lambda_vector[i];
        Mu[cell_id][i] =
          (1. / nb_waves) * u[cell_id][0] + inv_L2norm_lambda_vector * A[cell_id][1] * m_lambda_vector[i];
      }
    }
    return std::make_tuple(DiscreteFunctionP0Vector<double>(Mp), DiscreteFunctionP0Vector<double>(Mu));
  }

  NodeArray<double>
  compute_Flux1(const DiscreteFunctionP0Vector<const double>& F)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    NodeArray<double> Flux_F{m_mesh.connectivity(), nb_waves};
    Flux_F.fill(0);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < nb_waves; ++i) {
        const size_t N = m_mesh.numberOfCells();

        const auto& cell_to_node   = cell_to_node_matrix[cell_id];
        const NodeId left_node_id  = cell_to_node[0];
        const NodeId right_node_id = cell_to_node[1];

        const CellId prev_cell_id = (cell_id + N - 1) % N;
        const CellId next_cell_id = (cell_id + 1) % N;

        if (m_lambda_vector[i] < 0) {
          Flux_F[left_node_id][i]  = F[cell_id][i];
          Flux_F[right_node_id][i] = F[next_cell_id][i];

        } else if (m_lambda_vector[i] > 0) {
          Flux_F[left_node_id][i]  = F[prev_cell_id][i];
          Flux_F[right_node_id][i] = F[cell_id][i];

        } else {
          Flux_F[right_node_id][i] = 0;
          Flux_F[left_node_id][i]  = 0;
        }
      }
    }
    return Flux_F;
  }

  NodeArray<double>
  compute_Polynomial_reconstruction_2(const DiscreteFunctionP0Vector<const double>& F)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    NodeArray<double> Flux_F{m_mesh.connectivity(), nb_waves};
    Flux_F.fill(0);

    parallel_for(
      m_mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
        const NodeId first_node_id = 0;
        const NodeId last_node_id  = m_mesh.numberOfNodes() - 1;

        const double xmin = m_xr[first_node_id][0];
        const double xmax = m_xr[last_node_id][0];

        const size_t N = m_mesh.numberOfCells();

        const auto& cell_to_node   = cell_to_node_matrix[cell_id];
        const NodeId left_node_id  = cell_to_node[0];
        const NodeId right_node_id = cell_to_node[1];

        const CellId prev_cell_id = (cell_id + N - 1) % N;
        const CellId next_cell_id = (cell_id + 1) % N;

        const double xj = m_xj[cell_id][0];

        const auto& prev_cell_to_node = cell_to_node_matrix[prev_cell_id];
        const NodeId left2_node_id    = prev_cell_to_node[0];

        const auto& next_cell_to_node = cell_to_node_matrix[next_cell_id];
        const NodeId right2_node_id   = next_cell_to_node[1];

        const double left_node  = m_xr[left_node_id][0];
        const double right_node = m_xr[right_node_id][0];

        const double left2_node  = (-1 * (cell_id < 1) * (xmax - xmin)) + m_xr[left2_node_id][0];
        const double right2_node = ((cell_id + 1 >= N) * (xmax - xmin)) + m_xr[right2_node_id][0];

        const double left_minus_xj  = left_node - xj;
        const double right_minus_xj = right_node - xj;

        const double left2_node_xj  = left2_node - xj;
        const double right2_node_xj = right2_node - xj;

        constexpr double one_half = 1. / 2;

        for (size_t i = 0; i < nb_waves; ++i) {
          if (m_lambda_vector[i] < 0) {
            const TinyMatrix<1> Qj(one_half * m_inv_Vj[next_cell_id] *
                                   ((right2_node_xj) * (right2_node_xj) - (right_minus_xj) * (right_minus_xj)));

            TinyMatrix inv_Qj = inverse(Qj);

            const TinyVector<1> Bj(F[next_cell_id][i] - F[cell_id][i]);

            const TinyVector<1> Uj = inv_Qj * Bj;

            Flux_F[left_node_id][i] = Uj[0] * left_minus_xj + F[cell_id][i];

            if (cell_id == 0) {
              Flux_F[last_node_id][i] = Flux_F[left_node_id][i];
            }

          } else if (m_lambda_vector[i] > 0) {
            const TinyMatrix<1> Qj(one_half * m_inv_Vj[prev_cell_id] *
                                   ((left_minus_xj) * (left_minus_xj) - (left2_node_xj) * (left2_node_xj)));

            TinyMatrix inv_Qj = inverse(Qj);

            const TinyVector<1> Bj(F[prev_cell_id][i] - F[cell_id][i]);

            const TinyVector<1> Uj = inv_Qj * Bj;

            Flux_F[right_node_id][i] = Uj[0] * right_minus_xj + F[cell_id][i];

            if (cell_id == m_mesh.numberOfCells() - 1) {
              Flux_F[first_node_id][i] = Flux_F[right_node_id][i];
            }
          } else {
            Flux_F[right_node_id][i] = 0;
            Flux_F[left_node_id][i]  = 0;
          }
        }
      });

    return Flux_F;
  }

  NodeArray<double>
  compute_Polynomial_reconstruction_3(const DiscreteFunctionP0Vector<const double>& F)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    NodeArray<double> Flux_F{m_mesh.connectivity(), nb_waves};
    Flux_F.fill(0);

    parallel_for(
      m_mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
        const NodeId first_node_id = 0;
        const NodeId last_node_id  = m_mesh.numberOfNodes() - 1;

        const double xmin = m_xr[first_node_id][0];
        const double xmax = m_xr[last_node_id][0];

        const size_t N = m_mesh.numberOfCells();

        const auto& cell_to_node   = cell_to_node_matrix[cell_id];
        const NodeId left_node_id  = cell_to_node[0];
        const NodeId right_node_id = cell_to_node[1];

        const CellId prev_cell_id = (cell_id + N - 1) % N;
        const CellId next_cell_id = (cell_id + 1) % N;

        const double xj = m_xj[cell_id][0];

        const auto& prev_cell_to_node = cell_to_node_matrix[prev_cell_id];
        const NodeId left2_node_id    = prev_cell_to_node[0];

        const auto& next_cell_to_node = cell_to_node_matrix[next_cell_id];
        const NodeId right2_node_id   = next_cell_to_node[1];

        const double left_node  = m_xr[left_node_id][0];
        const double right_node = m_xr[right_node_id][0];

        const double left2_node  = (-1 * (cell_id < 1) * (xmax - xmin)) + m_xr[left2_node_id][0];
        const double right2_node = ((cell_id + 1 >= N) * (xmax - xmin)) + m_xr[right2_node_id][0];

        const double left_minus_xj  = left_node - xj;
        const double right_minus_xj = right_node - xj;

        const double left_minus_xj_3  = left_minus_xj * left_minus_xj * left_minus_xj;
        const double right_minus_xj_3 = right_minus_xj * right_minus_xj * right_minus_xj;

        const double left2_node_xj  = left2_node - xj;
        const double right2_node_xj = right2_node - xj;

        const double left2_minus_xj_3  = left2_node_xj * left2_node_xj * left2_node_xj;
        const double right2_minus_xj_3 = right2_node_xj * right2_node_xj * right2_node_xj;

        constexpr double one_third = 1. / 3;
        constexpr double one_half  = 1. / 2;

        const TinyMatrix<2> Qj(one_third * m_inv_Vj[prev_cell_id] * (left_minus_xj_3 - left2_minus_xj_3) -
                                 one_third * m_inv_Vj[cell_id] * (right_minus_xj_3 - left_minus_xj_3),
                               one_half * m_inv_Vj[prev_cell_id] *
                                 ((left_minus_xj) * (left_minus_xj) - (left2_node - xj) * (left2_node - xj)),
                               one_third * m_inv_Vj[next_cell_id] * (right2_minus_xj_3 - right_minus_xj_3) -
                                 one_third * m_inv_Vj[cell_id] * (right_minus_xj_3 - left_minus_xj_3),
                               one_half * m_inv_Vj[next_cell_id] *
                                 ((right2_node - xj) * (right2_node - xj) - (right_minus_xj) * (right_minus_xj)));

        TinyMatrix inv_Qj = inverse(Qj);

        for (size_t i = 0; i < nb_waves; ++i) {
          const TinyVector<2> Bj(F[prev_cell_id][i] - F[cell_id][i], F[next_cell_id][i] - F[cell_id][i]);

          const TinyVector<2> Uj = inv_Qj * Bj;

          if (m_lambda_vector[i] < 0) {
            Flux_F[left_node_id][i] = Uj[0] * ((left_minus_xj) * (left_minus_xj)-one_third * m_inv_Vj[cell_id] *
                                               (right_minus_xj_3 - left_minus_xj_3)) +
                                      Uj[1] * (left_minus_xj) + F[cell_id][i];

            if (cell_id == 0) {
              Flux_F[last_node_id][i] = Flux_F[left_node_id][i];
            }

          } else if (m_lambda_vector[i] > 0) {
            Flux_F[right_node_id][i] = Uj[0] * ((right_minus_xj) * (right_minus_xj)-one_third * m_inv_Vj[cell_id] *
                                                (right_minus_xj_3 - left_minus_xj_3)) +
                                       Uj[1] * (right_minus_xj) + F[cell_id][i];

            if (cell_id == m_mesh.numberOfCells() - 1) {
              Flux_F[first_node_id][i] = Flux_F[right_node_id][i];
            }
          } else {
            Flux_F[right_node_id][i] = 0;
            Flux_F[left_node_id][i]  = 0;
          }
        }
      });

    return Flux_F;
  }

  NodeArray<double>
  TVD(NodeArray<double>& Flux_F, const DiscreteFunctionP0Vector<const double>& F, const double c)
  {
    const size_t nb_waves = m_lambda_vector.size();

    parallel_for(
      m_mesh.numberOfNodes(), PUGS_LAMBDA(const NodeId node_id) {
        const size_t N = m_mesh.numberOfCells();

        for (size_t i_wave = 0; i_wave < nb_waves; ++i_wave) {
          if (m_lambda_vector[i_wave] > 0) {
            const size_t j_p_half = node_id;

            const CellId cell_id_jm1 = (j_p_half + N - 2) % N;
            const CellId cell_id_j   = (j_p_half + N - 1) % N;
            const CellId cell_id_jp1 = (j_p_half + N) % N;

            double Fmin = std::min(F[cell_id_j][i_wave], F[cell_id_jp1][i_wave]);
            double Fmax = std::max(F[cell_id_j][i_wave], F[cell_id_jp1][i_wave]);

            double nu   = m_lambda_vector[i_wave] * c * m_dt / (m_Vj[cell_id_j]);
            double F_ul = F[cell_id_j][i_wave] + (1 - nu) / nu * (F[cell_id_j][i_wave] - F[cell_id_jm1][i_wave]);

            double Fmin_ul = std::min(F[cell_id_j][i_wave], F_ul);
            double Fmax_ul = std::max(F[cell_id_j][i_wave], F_ul);

            Fmin = std::max(Fmin, Fmin_ul);
            Fmax = std::min(Fmax, Fmax_ul);

            if (Flux_F[node_id][i_wave] < Fmin) {
              Flux_F[node_id][i_wave] = Fmin;
            } else if (Flux_F[node_id][i_wave] > Fmax) {
              Flux_F[node_id][i_wave] = Fmax;
            }
          } else if (m_lambda_vector[i_wave] < 0) {
            const size_t j_p_half = node_id;

            const CellId cell_id_j   = (j_p_half + N - 1) % N;
            const CellId cell_id_jp1 = (j_p_half + N) % N;
            const CellId cell_id_jp2 = (j_p_half + N + 1) % N;

            double Fmin = std::min(F[cell_id_j][i_wave], F[cell_id_jp1][i_wave]);
            double Fmax = std::max(F[cell_id_j][i_wave], F[cell_id_jp1][i_wave]);

            double nu   = m_lambda_vector[i_wave] * c * m_dt / (m_Vj[cell_id_j]);
            double F_ul = F[cell_id_jp1][i_wave] - (1 + nu) / nu * (F[cell_id_jp1][i_wave] - F[cell_id_jp2][i_wave]);

            double Fmin_ul = std::min(F[cell_id_jp1][i_wave], F_ul);
            double Fmax_ul = std::max(F[cell_id_jp1][i_wave], F_ul);

            Fmin = std::max(Fmin, Fmin_ul);
            Fmax = std::min(Fmax, Fmax_ul);

            if (Flux_F[node_id][i_wave] < Fmin) {
              Flux_F[node_id][i_wave] = Fmin;
            } else if (Flux_F[node_id][i_wave] > Fmax) {
              Flux_F[node_id][i_wave] = Fmax;
            }
          }
        }
      });
    return Flux_F;
  }

  std::tuple<std::shared_ptr<const MeshVariant>,
             std::shared_ptr<const DiscreteFunctionVariant>,
             std::shared_ptr<const DiscreteFunctionVariant>,
             std::shared_ptr<const DiscreteFunctionVariant>>
  apply()
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    SmallArray<const double> B;
    SmallArray<const double> c;

    if (m_timeOrder == 1) {
      SmallArray<double> tmp_B(1);
      SmallArray<double> tmp_c(1);
      tmp_B[0] = 1;
      tmp_c[0] = 1;

      B = tmp_B;
      c = tmp_c;
    } else if (m_timeOrder == 2) {
      SmallArray<double> tmp_B(2);
      SmallArray<double> tmp_c(2);

      tmp_B[0] = 0.5;
      tmp_B[1] = 0.5;
      tmp_c[0] = 0.5;
      tmp_c[1] = 1;

      B = tmp_B;
      c = tmp_c;
    } else {
      throw NotImplementedError("Order higher than two in time not implemented");
    }

    DiscreteFunctionP0Vector<double> F_p(p_mesh, nb_waves);
    DiscreteFunctionP0Vector<double> F_u(p_mesh, nb_waves);
    DiscreteFunctionP0Vector<double> F0_p(p_mesh, nb_waves);
    DiscreteFunctionP0Vector<double> F0_u(p_mesh, nb_waves);

    double lambda = 0;
    for (size_t i_wave = 0; i_wave < m_lambda_vector.size(); ++i_wave) {
      lambda = std::max(lambda, m_lambda_vector[i_wave]);
    }

    std::tuple<DiscreteFunctionP0Vector<double>, DiscreteFunctionP0Vector<double>> vec_F(F_p, F_u);

    CellValue<double> rho                       = copy(m_rho.cellValues());
    DiscreteFunctionP0<TinyVector<Dimension>> u = copy(m_u);
    DiscreteFunctionP0<double> p                = copy(m_p);
    CellValue<double> E                         = copy(m_E.cellValues());
    DiscreteFunctionP0<double> tau{p_mesh};

    CellValue<double> cj = [&] {
      CellValue<double> computed_cj{m_mesh.connectivity()};
      parallel_for(
        m_mesh.numberOfCells(),
        PUGS_LAMBDA(const CellId cell_id) { computed_cj[cell_id] = std::sqrt(m_gamma * p[cell_id] / rho[cell_id]); });
      return computed_cj;
    }();

    NodeArray<TinyVector<Dimension>> u_flux(m_mesh.connectivity(), m_timeOrder);
    NodeArray<double> p_flux(m_mesh.connectivity(), m_timeOrder);

    for (size_t i_time = 0; i_time < m_timeOrder; ++i_time) {
      if (i_time == 0) {
        vec_F = compute_M(m_p, m_u);
      } else {
        vec_F = compute_M(p, u);
      }
      F_p  = get<0>(vec_F);
      F_u  = get<1>(vec_F);
      F0_p = copy(F_p);
      F0_u = copy(F_u);

      NodeArray<double> Flux_Fn_p(m_mesh.connectivity(), nb_waves);
      NodeArray<double> Flux_Fn_u(m_mesh.connectivity(), nb_waves);

      if (m_spaceOrder == 1) {
        Flux_Fn_p = compute_Flux1(F_p);
        Flux_Fn_u = compute_Flux1(F_u);
      } else if (m_spaceOrder == 2) {
        Flux_Fn_p = compute_Polynomial_reconstruction_2(F_p);
        Flux_Fn_u = compute_Polynomial_reconstruction_2(F_u);
      } else if (m_spaceOrder == 3) {
        Flux_Fn_p = compute_Polynomial_reconstruction_3(F_p);
        Flux_Fn_u = compute_Polynomial_reconstruction_3(F_u);
      } else {
        throw NotImplementedError("Only implemented for space orders 1, 2 and 3");
      }

      // Flux_Fn_p = TVD(Flux_Fn_p, F0_p, c[i_time]);
      // Flux_Fn_u = TVD(Flux_Fn_u, F0_u, c[i_time]);

      if (i_time == 0) {
        for (size_t j_time = 0; j_time < m_timeOrder; ++j_time) {
          for (NodeId node_id = 0; node_id < m_mesh.numberOfNodes(); ++node_id) {
            u_flux[node_id][j_time][0] = 0;
            p_flux[node_id][j_time]    = 0;
            for (size_t i_wave = 0; i_wave < nb_waves; ++i_wave) {
              u_flux[node_id][j_time][0] += m_lambda_vector[i_wave] * Flux_Fn_p[node_id][i_wave];
              p_flux[node_id][j_time] += m_lambda_vector[i_wave] * Flux_Fn_u[node_id][i_wave];
            }
          }
        }
      } else {
        for (NodeId node_id = 0; node_id < m_mesh.numberOfNodes(); ++node_id) {
          u_flux[node_id][i_time][0] = 0;
          p_flux[node_id][i_time]    = 0;
          for (size_t i_wave = 0; i_wave < nb_waves; ++i_wave) {
            u_flux[node_id][i_time][0] += m_lambda_vector[i_wave] * Flux_Fn_p[node_id][i_wave];
            p_flux[node_id][i_time] += m_lambda_vector[i_wave] * Flux_Fn_u[node_id][i_wave];
          }
        }
      }

      for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
        const double dt_over_Mj    = (m_dt * m_inv_Mj[cell_id]);
        const auto& cell_nodes     = cell_to_node_matrix[cell_id];
        const NodeId left_node_id  = cell_nodes[0];
        const NodeId right_node_id = cell_nodes[1];
        u[cell_id][0]              = m_u[cell_id][0];
        tau[cell_id]               = 1. / m_rho[cell_id];
        for (size_t j_time = 0; j_time < m_timeOrder; ++j_time) {
          u[cell_id][0] -= B[j_time] * (rho[cell_id] * cj[cell_id]) * (rho[cell_id] * cj[cell_id]) * dt_over_Mj *
                           (p_flux[right_node_id][j_time] - p_flux[left_node_id][j_time]);
          // u[cell_id][0] -= B[j_time] * (rho[cell_id] * cj[cell_id]) * (rho[cell_id] * cj[cell_id]) * dt_over_Mj *
          //                 (p_flux[right_node_id][j_time] - p_flux[left_node_id][j_time]);

          tau[cell_id] += B[j_time] * dt_over_Mj * (u_flux[right_node_id][j_time][0] - u_flux[left_node_id][j_time][0]);
        }
        p[cell_id] = m_p[cell_id] / std::pow(m_rho[cell_id] * tau[cell_id], m_gamma);
      }
    }

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      const double dt_over_Mj    = (m_dt * m_inv_Mj[cell_id]);
      const auto& cell_nodes     = cell_to_node_matrix[cell_id];
      const NodeId left_node_id  = cell_nodes[0];
      const NodeId right_node_id = cell_nodes[1];

      u[cell_id][0] = m_u[cell_id][0];
      E[cell_id]    = m_E[cell_id];
      for (size_t j_time = 0; j_time < m_timeOrder; ++j_time) {
        u[cell_id][0] -= B[j_time] * dt_over_Mj * (p_flux[right_node_id][j_time] - p_flux[left_node_id][j_time]);

        E[cell_id] -= B[j_time] * dt_over_Mj *
                      (p_flux[right_node_id][j_time] * u_flux[right_node_id][j_time][0] -
                       p_flux[left_node_id][j_time] * u_flux[left_node_id][j_time][0]);
      }
    }

    NodeValue<TinyVector<Dimension>> new_xr = copy(m_mesh.xr());
    parallel_for(
      m_mesh.numberOfNodes(), PUGS_LAMBDA(NodeId node_id) {
        for (size_t j_time = 0; j_time < m_timeOrder; ++j_time) {
          new_xr[node_id][0] += B[j_time] * m_dt * u_flux[node_id][j_time][0];
        }
      });

    std::shared_ptr<const MeshType> new_mesh = std::make_shared<MeshType>(m_mesh.shared_connectivity(), new_xr);

    CellValue<const double> new_Vj = MeshDataManager::instance().getMeshData(*new_mesh).Vj();

    parallel_for(
      m_mesh.numberOfCells(), PUGS_LAMBDA(CellId cell_id) { rho[cell_id] = m_Mj[cell_id] / new_Vj[cell_id]; });

    CellValue<TinyVector<Dimension>> new_u = copy(u.cellValues());

    return std::make_tuple(std::make_shared<MeshVariant>(new_mesh),
                           std::make_shared<DiscreteFunctionVariant>(DiscreteFunctionP0<const double>(new_mesh, rho)),
                           std::make_shared<DiscreteFunctionVariant>(
                             DiscreteFunctionP0<const TinyVector<Dimension>>(new_mesh, new_u)),
                           std::make_shared<DiscreteFunctionVariant>(DiscreteFunctionP0<const double>(new_mesh, E)));
  }

  EulerKineticSolverAcousticLagrangeFV(std::shared_ptr<const MeshType> mesh,
                                       const double& dt,
                                       const std::vector<double>& lambda_vector,
                                       const double& gamma,
                                       const size_t& space_order,
                                       const size_t& time_order,
                                       const DiscreteFunctionP0<const double>& rho,
                                       const DiscreteFunctionP0<const TinyVector<MeshType::Dimension>>& u,
                                       const DiscreteFunctionP0<const double>& E,
                                       const DiscreteFunctionP0<const double>& p)
    : m_dt{dt},
      m_lambda_vector{lambda_vector},
      m_gamma{gamma},
      m_spaceOrder{space_order},
      m_timeOrder{time_order},
      m_rho{rho},
      m_u{u},
      m_E{E},
      m_p{p},
      p_mesh{mesh},
      m_mesh{*mesh}
  {
    m_Mj = [&] {
      CellValue<double> Mj{m_mesh.connectivity()};
      parallel_for(
        m_mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) { Mj[cell_id] = m_rho[cell_id] * m_Vj[cell_id]; });
      return Mj;
    }();

    m_inv_Mj = [&] {
      CellValue<double> inv_Mj{m_mesh.connectivity()};
      parallel_for(
        m_mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) { inv_Mj[cell_id] = 1. / m_Mj[cell_id]; });
      return inv_Mj;
    }();

    m_inv_Vj = [&] {
      CellValue<double> inv_Vj{m_mesh.connectivity()};
      parallel_for(
        m_mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) { inv_Vj[cell_id] = 1. / m_Vj[cell_id]; });
      return inv_Vj;
    }();
  }

  ~EulerKineticSolverAcousticLagrangeFV() = default;
};

std::tuple<std::shared_ptr<const MeshVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
eulerKineticSolverAcousticLagrangeFV(const double& dt,
                                     const std::vector<double>& lambda_vector,
                                     const double& gamma,
                                     const double& eps,
                                     const size_t& space_order,
                                     const size_t& time_order,
                                     const std::shared_ptr<const DiscreteFunctionVariant>& rho_n,
                                     const std::shared_ptr<const DiscreteFunctionVariant>& u_n,
                                     const std::shared_ptr<const DiscreteFunctionVariant>& E_n,
                                     const std::shared_ptr<const DiscreteFunctionVariant>& p_n)
{
  const std::shared_ptr<const MeshVariant> mesh_v = getCommonMesh({rho_n, u_n, E_n, p_n});
  if (mesh_v.use_count() == 0) {
    throw NormalError("rho_n, u_n, E_n and p_n have to be defined on the same mesh");
  }
  if (eps != 0) {
    throw NotImplementedError("Euler kinetic solver Lagrange not implemented for eps != 0");
  }

  return std::visit(
    [&](auto&& p_mesh)
      -> std::tuple<std::shared_ptr<const MeshVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                    std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>> {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      if constexpr (is_polygonal_mesh_v<MeshType> and (MeshType::Dimension == 1)) {
        auto rho = rho_n->get<DiscreteFunctionP0<const double>>();
        auto u   = u_n->get<DiscreteFunctionP0<const TinyVector<MeshType::Dimension>>>();
        auto E   = E_n->get<DiscreteFunctionP0<const double>>();
        auto p   = p_n->get<DiscreteFunctionP0<const double>>();

        EulerKineticSolverAcousticLagrangeFV solver(p_mesh, dt, lambda_vector, gamma, space_order, time_order, rho, u,
                                                    E, p);

        return solver.apply();
      } else {
        throw NotImplementedError("Invalid mesh type for Lagrangian acoustic Kinetic solver");
      }
    },
    mesh_v->variant());
}
