#include <scheme/EulerKineticSolverLagrangeFV.hpp>

#include <language/utils/EvaluateAtPoints.hpp>
#include <language/utils/InterpolateItemValue.hpp>   //new
#include <mesh/Connectivity.hpp>
#include <mesh/Mesh.hpp>
#include <mesh/MeshVariant.hpp>
#include <scheme/DiscreteFunctionDescriptorP0Vector.hpp>
#include <scheme/DiscreteFunctionP0.hpp>   //new
#include <scheme/DiscreteFunctionP0Vector.hpp>
#include <scheme/DiscreteFunctionUtils.hpp>
#include <scheme/IDiscreteFunctionDescriptor.hpp>

#include <analysis/GaussLegendreQuadratureDescriptor.hpp>
#include <analysis/QuadratureManager.hpp>
#include <geometry/LineTransformation.hpp>
#include <tuple>

template <MeshConcept MeshType>
std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticExactSolutionLagrangeFV(const std::shared_ptr<const MeshType> p_mesh, const double t, const double gamma)
{
  DiscreteFunctionP0<double> rho_ex(p_mesh);
  DiscreteFunctionP0<double> u_ex(p_mesh);
  DiscreteFunctionP0<double> p_ex(p_mesh);
  DiscreteFunctionP0<double> rho_u_ex(p_mesh);
  DiscreteFunctionP0<double> rho_E_ex(p_mesh);

  MeshData<MeshType>& mesh_data                               = MeshDataManager::instance().getMeshData(*p_mesh);
  const auto& cell_to_node_matrix                             = p_mesh->connectivity().cellToNodeMatrix();
  const CellValue<const TinyVector<MeshType::Dimension>> m_xj = mesh_data.xj();
  const NodeValue<const TinyVector<MeshType::Dimension>> m_xr = p_mesh->xr();
  const CellValue<const double> m_Vj                          = mesh_data.Vj();

  const double pi = acos(-1);
  // EvaluateAtPoints<double(double)>::evaluateTo(density_function_id, x, rho0);

  for (CellId cell_id = 0; cell_id < p_mesh->numberOfCells(); ++cell_id) {
    const auto& cell_to_node   = cell_to_node_matrix[cell_id];
    const NodeId left_node_id  = cell_to_node[0];
    const NodeId right_node_id = cell_to_node[1];
    const double dx            = m_Vj[cell_id];

    const LineTransformation<1> line_transformation(m_xr[left_node_id], m_xr[right_node_id]);
    QuadratureFormula<1> quadrature =
      QuadratureManager::instance().getLineFormula(GaussLegendreQuadratureDescriptor(4));

    rho_ex[cell_id]   = 0;
    u_ex[cell_id]     = 0;
    p_ex[cell_id]     = 0;
    rho_u_ex[cell_id] = 0;
    rho_E_ex[cell_id] = 0;

    for (size_t i_gauss = 0; i_gauss < quadrature.numberOfPoints(); ++i_gauss) {
      double gauss_point = line_transformation(quadrature.point(i_gauss))[0];

      double a1      = -1.5;
      double b1      = 1.5;
      double c1      = 0.5 * (a1 + b1);
      double x1      = c1;
      double rho0_x1 = 1 + 0.5 * std::sin(2 * pi * x1);
      double rho0_a1 = 1 + 0.5 * std::sin(2 * pi * a1);
      double rho0_b1 = 1 + 0.5 * std::sin(2 * pi * b1);
      double rho0_c1 = 1 + 0.5 * std::sin(2 * pi * c1);

      double fx1 = gauss_point + std::sqrt(3) * rho0_x1 * t - x1;
      double fa1 = gauss_point + std::sqrt(3) * rho0_a1 * t - a1;
      double fb1 = gauss_point + std::sqrt(3) * rho0_b1 * t - b1;
      double fc1 = gauss_point + std::sqrt(3) * rho0_c1 * t - c1;

      while (std::abs(fx1) > 1E-15) {
        if (fa1 * fc1 < 0) {
          b1 = c1;
          c1 = 0.5 * (a1 + b1);
        } else if (fc1 * fb1 < 0) {
          a1 = c1;
          c1 = 0.5 * (a1 + b1);
        }
        x1 = c1;

        rho0_x1 = 1 + 0.5 * std::sin(2 * pi * x1);
        rho0_a1 = 1 + 0.5 * std::sin(2 * pi * a1);
        rho0_b1 = 1 + 0.5 * std::sin(2 * pi * b1);
        rho0_c1 = 1 + 0.5 * std::sin(2 * pi * c1);

        fx1 = gauss_point + std::sqrt(3) * rho0_x1 * t - x1;
        fa1 = gauss_point + std::sqrt(3) * rho0_a1 * t - a1;
        fb1 = gauss_point + std::sqrt(3) * rho0_b1 * t - b1;
        fc1 = gauss_point + std::sqrt(3) * rho0_c1 * t - c1;
      }

      double a2      = -1.5;
      double b2      = 1.5;
      double c2      = 0.5 * (a2 + b2);
      double x2      = c2;
      double rho0_x2 = 1 + 0.5 * std::sin(2 * pi * x2);
      double rho0_a2 = 1 + 0.5 * std::sin(2 * pi * a2);
      double rho0_b2 = 1 + 0.5 * std::sin(2 * pi * b2);
      double rho0_c2 = 1 + 0.5 * std::sin(2 * pi * c2);

      double fx2 = gauss_point - std::sqrt(3) * rho0_x2 * t - x2;
      double fa2 = gauss_point - std::sqrt(3) * rho0_a2 * t - a2;
      double fb2 = gauss_point - std::sqrt(3) * rho0_b2 * t - b2;
      double fc2 = gauss_point - std::sqrt(3) * rho0_c2 * t - c2;

      while (std::abs(fx2) > 1E-15) {
        if (fa2 * fc2 < 0) {
          b2 = c2;
          c2 = 0.5 * (a2 + b2);
        } else if (fc2 * fb2 < 0) {
          a2 = c2;
          c2 = 0.5 * (a2 + b2);
        }
        x2 = c2;

        rho0_x2 = 1 + 0.5 * std::sin(2 * pi * x2);
        rho0_a2 = 1 + 0.5 * std::sin(2 * pi * a2);
        rho0_b2 = 1 + 0.5 * std::sin(2 * pi * b2);
        rho0_c2 = 1 + 0.5 * std::sin(2 * pi * c2);

        fx2 = gauss_point - std::sqrt(3) * rho0_x2 * t - x2;
        fa2 = gauss_point - std::sqrt(3) * rho0_a2 * t - a2;
        fb2 = gauss_point - std::sqrt(3) * rho0_b2 * t - b2;
        fc2 = gauss_point - std::sqrt(3) * rho0_c2 * t - c2;
      }

      rho0_x1 = 1 + 0.5 * std::sin(2 * pi * x1);
      rho0_x2 = 1 + 0.5 * std::sin(2 * pi * x2);

      const double rho_gauss = 0.5 * (rho0_x1 + rho0_x2);
      const double u_gauss   = std::sqrt(3) * ((0.5 * (rho0_x1 + rho0_x2)) - rho0_x1);
      const double p_gauss   = (0.5 * (rho0_x1 + rho0_x2)) * (0.5 * (rho0_x1 + rho0_x2)) * (0.5 * (rho0_x1 + rho0_x2));

      rho_u_ex[cell_id] += quadrature.weight(i_gauss) * line_transformation.jacobianDeterminant() * rho_gauss * u_gauss;
      rho_E_ex[cell_id] += quadrature.weight(i_gauss) * line_transformation.jacobianDeterminant() *
                           (p_gauss / (gamma - 1) + 0.5 * rho_gauss * u_gauss * u_gauss);

      rho_ex[cell_id] += quadrature.weight(i_gauss) * line_transformation.jacobianDeterminant() * rho_gauss;
      u_ex[cell_id] += quadrature.weight(i_gauss) * line_transformation.jacobianDeterminant() * u_gauss;
      p_ex[cell_id] += quadrature.weight(i_gauss) * line_transformation.jacobianDeterminant() * p_gauss;
    }
    rho_u_ex[cell_id] /= dx;
    rho_E_ex[cell_id] /= dx;
    rho_ex[cell_id] /= dx;
    u_ex[cell_id] /= dx;
    p_ex[cell_id] /= dx;
  }

  return std::make_tuple(std::make_shared<DiscreteFunctionVariant>(rho_ex),
                         std::make_shared<DiscreteFunctionVariant>(u_ex),
                         std::make_shared<DiscreteFunctionVariant>(p_ex),
                         std::make_shared<DiscreteFunctionVariant>(rho_u_ex),
                         std::make_shared<DiscreteFunctionVariant>(rho_E_ex));
}

std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticExactSolutionLagrangeFV(const double& t,
                                       const double& gamma,
                                       const std::shared_ptr<const DiscreteFunctionVariant>& rho_ex,
                                       const std::shared_ptr<const DiscreteFunctionVariant>& u_ex,
                                       const std::shared_ptr<const DiscreteFunctionVariant>& p_ex,
                                       const std::shared_ptr<const DiscreteFunctionVariant>& rho_u_ex,
                                       const std::shared_ptr<const DiscreteFunctionVariant>& rho_E_ex)
{
  const std::shared_ptr<const MeshVariant> mesh_v = getCommonMesh({u_ex, rho_ex, p_ex, rho_u_ex, rho_E_ex});
  if (mesh_v.use_count() == 0) {
    throw NormalError("rho_ex, u_ex, p_ex, rho_u_ex and rho_E_ex have to be defined on the same mesh");
  }

  return std::visit(
    [&](auto&& p_mesh)
      -> std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                    std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                    std::shared_ptr<const DiscreteFunctionVariant>> {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      if constexpr (is_polygonal_mesh_v<MeshType> and (MeshType::Dimension == 1)) {
        return getEulerKineticExactSolutionLagrangeFV(p_mesh, t, gamma);
      } else {
        throw NotImplementedError("Invalid mesh type for Scalar Kinetic solver");
      }
    },
    mesh_v->variant());
}

template <MeshConcept MeshType>
std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticWavesLagrangeFV(const std::shared_ptr<const MeshType> p_mesh,
                               const std::vector<double> lambda_vector,
                               DiscreteFunctionP0<const double> tau,
                               DiscreteFunctionP0<const TinyVector<MeshType::Dimension>> u,
                               DiscreteFunctionP0<const double> E,
                               const double gamma)
{
  if (lambda_vector.size() < 2) {
    throw NormalError("lambda vector must be of dimension greater than 2");
  }

  MeshData<MeshType>& mesh_data                               = MeshDataManager::instance().getMeshData(*p_mesh);
  const NodeValue<const TinyVector<MeshType::Dimension>> m_xr = p_mesh->xr();
  const CellValue<const TinyVector<MeshType::Dimension>> m_xj = mesh_data.xj();
  const CellValue<const double> m_Vj                          = mesh_data.Vj();

  DiscreteFunctionP0Vector<double> Fn_tau(p_mesh, lambda_vector.size());
  DiscreteFunctionP0Vector<double> Fn_u(p_mesh, lambda_vector.size());
  DiscreteFunctionP0Vector<double> Fn_E(p_mesh, lambda_vector.size());

  CellArray<double> vec_A{p_mesh->connectivity(), 3};
  CellValue<double> p{p_mesh->connectivity()};

  for (CellId cell_id = 0; cell_id < p_mesh->numberOfCells(); ++cell_id) {
    p[cell_id] = (gamma - 1) * (1 / tau[cell_id]) * (E[cell_id] - 0.5 * u[cell_id][0] * u[cell_id][0]);

    vec_A[cell_id][0] = -u[cell_id][0];
    vec_A[cell_id][1] = p[cell_id];
    vec_A[cell_id][2] = p[cell_id] * u[cell_id][0];
  }

  const size_t nb_waves           = lambda_vector.size();
  double inv_L2norm_lambda_vector = 0;
  for (size_t i = 0; i < nb_waves; ++i) {
    inv_L2norm_lambda_vector += std::pow(lambda_vector[i], 2);
  }
  inv_L2norm_lambda_vector = 1. / inv_L2norm_lambda_vector;

  for (CellId cell_id = 0; cell_id < p_mesh->numberOfCells(); ++cell_id) {
    for (size_t i = 0; i < nb_waves; ++i) {
      Fn_tau[cell_id][i] =
        (1. / nb_waves) * tau[cell_id] + inv_L2norm_lambda_vector * vec_A[cell_id][0] * lambda_vector[i];
      Fn_u[cell_id][i] =
        (1. / nb_waves) * u[cell_id][0] + inv_L2norm_lambda_vector * vec_A[cell_id][1] * lambda_vector[i];
      Fn_E[cell_id][i] = (1. / nb_waves) * E[cell_id] + inv_L2norm_lambda_vector * vec_A[cell_id][2] * lambda_vector[i];
    }
  }

  return std::make_tuple(std::make_shared<DiscreteFunctionVariant>(Fn_tau),
                         std::make_shared<DiscreteFunctionVariant>(Fn_u),
                         std::make_shared<DiscreteFunctionVariant>(Fn_E));
}

std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticWavesLagrangeFV(const std::vector<double>& lambda_vector,
                               const std::shared_ptr<const DiscreteFunctionVariant>& tau,
                               const std::shared_ptr<const DiscreteFunctionVariant>& u,
                               const std::shared_ptr<const DiscreteFunctionVariant>& E,
                               const double& gamma)
{
  const std::shared_ptr<const MeshVariant> mesh_v = getCommonMesh({tau, u, E});
  if (mesh_v.use_count() == 0) {
    throw NormalError("tau, u and E have to be defined on the same mesh");
  }

  return std::visit(
    [&](auto&& p_mesh)
      -> std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                    std::shared_ptr<const DiscreteFunctionVariant>> {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      if constexpr (is_polygonal_mesh_v<MeshType> and (MeshType::Dimension == 1)) {
        auto tau_h = tau->get<DiscreteFunctionP0<const double>>();
        auto u_h   = u->get<DiscreteFunctionP0<const TinyVector<MeshType::Dimension>>>();
        auto E_h   = E->get<DiscreteFunctionP0<const double>>();

        return getEulerKineticWavesLagrangeFV(p_mesh, lambda_vector, tau_h, u_h, E_h, gamma);
      } else {
        throw NotImplementedError("Invalid mesh type for Scalar Kinetic solver");
      }
    },
    mesh_v->variant());
}

template <MeshConcept MeshType>
class EulerKineticSolverLagrangeFV
{
 private:
  constexpr static size_t Dimension = MeshType::Dimension;
  using Rd                          = TinyVector<Dimension>;

  const double m_dt;
  const std::vector<double>& m_lambda_vector;
  const size_t m_spaceOrder;
  const size_t m_timeOrder;
  const DiscreteFunctionP0Vector<const double> m_Fn_tau;
  const DiscreteFunctionP0Vector<const double> m_Fn_u;
  const DiscreteFunctionP0Vector<const double> m_Fn_E;
  const double m_gamma;
  const std::shared_ptr<const MeshType> p_mesh;
  const MeshType& m_mesh;
  const CellValue<const double> m_Vj = MeshDataManager::instance().getMeshData(m_mesh).Vj();
  CellValue<const double> m_inv_Vj;
  const CellValue<const TinyVector<Dimension>> m_xj = MeshDataManager::instance().getMeshData(m_mesh).xj();
  const NodeValue<const TinyVector<Dimension>> m_xr = m_mesh.xr();

 public:
  CellArray<double>
  getA(const CellValue<const double>& tau,
       const CellValue<const TinyVector<Dimension>>& u,
       const CellValue<const double>& E) const
  {
    CellArray<double> vec_A{m_mesh.connectivity(), 3};
    CellValue<double> p{m_mesh.connectivity()};

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      p[cell_id] = (m_gamma - 1) * (1 / tau[cell_id]) * (E[cell_id] - 0.5 * u[cell_id][0] * u[cell_id][0]);

      vec_A[cell_id][0] = -u[cell_id][0];
      vec_A[cell_id][1] = p[cell_id];
      vec_A[cell_id][2] = p[cell_id] * u[cell_id][0];
    }

    return vec_A;
  }

  std::tuple<DiscreteFunctionP0Vector<double>, DiscreteFunctionP0Vector<double>, DiscreteFunctionP0Vector<double>>
  compute_M(const CellValue<const double>& tau,
            const CellValue<const TinyVector<Dimension>>& u,
            const CellValue<const double>& E)
  {
    const size_t nb_waves = m_lambda_vector.size();
    DiscreteFunctionP0Vector<double> Mtau{p_mesh, nb_waves};
    DiscreteFunctionP0Vector<double> Mu{p_mesh, nb_waves};
    DiscreteFunctionP0Vector<double> ME{p_mesh, nb_waves};
    CellArray<double> A = getA(tau, u, E);

    double inv_L2norm_lambda_vector = 0;
    for (size_t i = 0; i < nb_waves; ++i) {
      inv_L2norm_lambda_vector += m_lambda_vector[i] * m_lambda_vector[i];
    }
    inv_L2norm_lambda_vector = 1. / inv_L2norm_lambda_vector;

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < nb_waves; ++i) {
        Mtau[cell_id][i] =
          (1. / nb_waves) * tau[cell_id] + inv_L2norm_lambda_vector * A[cell_id][0] * m_lambda_vector[i];
        Mu[cell_id][i] =
          (1. / nb_waves) * u[cell_id][0] + inv_L2norm_lambda_vector * A[cell_id][1] * m_lambda_vector[i];
        ME[cell_id][i] = (1. / nb_waves) * E[cell_id] + inv_L2norm_lambda_vector * A[cell_id][2] * m_lambda_vector[i];
      }
    }
    return std::make_tuple(DiscreteFunctionP0Vector<double>(Mtau), DiscreteFunctionP0Vector<double>(Mu),
                           DiscreteFunctionP0Vector<double>(ME));
  }

  NodeArray<double>
  compute_Flux1(const DiscreteFunctionP0Vector<const double>& F)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    NodeArray<double> Flux_F{m_mesh.connectivity(), nb_waves};
    Flux_F.fill(0);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < nb_waves; ++i) {
        const size_t N = m_mesh.numberOfCells();

        const auto& cell_to_node   = cell_to_node_matrix[cell_id];
        const NodeId left_node_id  = cell_to_node[0];
        const NodeId right_node_id = cell_to_node[1];

        const CellId prev_cell_id = (cell_id + N - 1) % N;
        const CellId next_cell_id = (cell_id + 1) % N;

        if (m_lambda_vector[i] < 0) {
          Flux_F[left_node_id][i]  = F[cell_id][i];
          Flux_F[right_node_id][i] = F[next_cell_id][i];

        } else if (m_lambda_vector[i] > 0) {
          Flux_F[left_node_id][i]  = F[prev_cell_id][i];
          Flux_F[right_node_id][i] = F[cell_id][i];

        } else {
          Flux_F[right_node_id][i] = 0;
          Flux_F[left_node_id][i]  = 0;
        }
      }
    }
    return Flux_F;
  }

  NodeArray<double>
  compute_Polynomial_reconstruction_2(const DiscreteFunctionP0Vector<const double>& F)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    NodeArray<double> Flux_F{m_mesh.connectivity(), nb_waves};
    Flux_F.fill(0);

    parallel_for(
      m_mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
        const NodeId first_node_id = 0;
        const NodeId last_node_id  = m_mesh.numberOfNodes() - 1;

        const double xmin = m_xr[first_node_id][0];
        const double xmax = m_xr[last_node_id][0];

        const size_t N = m_mesh.numberOfCells();

        const auto& cell_to_node   = cell_to_node_matrix[cell_id];
        const NodeId left_node_id  = cell_to_node[0];
        const NodeId right_node_id = cell_to_node[1];

        const CellId prev_cell_id = (cell_id + N - 1) % N;
        const CellId next_cell_id = (cell_id + 1) % N;

        const double xj = m_xj[cell_id][0];

        const auto& prev_cell_to_node = cell_to_node_matrix[prev_cell_id];
        const NodeId left2_node_id    = prev_cell_to_node[0];

        const auto& next_cell_to_node = cell_to_node_matrix[next_cell_id];
        const NodeId right2_node_id   = next_cell_to_node[1];

        const double left_node  = m_xr[left_node_id][0];
        const double right_node = m_xr[right_node_id][0];

        const double left2_node  = (-1 * (cell_id < 1) * (xmax - xmin)) + m_xr[left2_node_id][0];
        const double right2_node = ((cell_id + 1 >= N) * (xmax - xmin)) + m_xr[right2_node_id][0];

        const double left_minus_xj  = left_node - xj;
        const double right_minus_xj = right_node - xj;

        const double left2_node_xj  = left2_node - xj;
        const double right2_node_xj = right2_node - xj;

        constexpr double one_half = 1. / 2;

        for (size_t i = 0; i < nb_waves; ++i) {
          if (m_lambda_vector[i] < 0) {
            const TinyMatrix<1> Qj(one_half * m_inv_Vj[next_cell_id] *
                                   ((right2_node_xj) * (right2_node_xj) - (right_minus_xj) * (right_minus_xj)));

            TinyMatrix inv_Qj = inverse(Qj);

            const TinyVector<1> Bj(F[next_cell_id][i] - F[cell_id][i]);

            const TinyVector<1> Uj = inv_Qj * Bj;

            Flux_F[left_node_id][i] = Uj[0] * left_minus_xj + F[cell_id][i];

            if (cell_id == 0) {
              Flux_F[last_node_id][i] = Flux_F[left_node_id][i];
            }

          } else if (m_lambda_vector[i] > 0) {
            const TinyMatrix<1> Qj(one_half * m_inv_Vj[prev_cell_id] *
                                   ((left_minus_xj) * (left_minus_xj) - (left2_node_xj) * (left2_node_xj)));

            TinyMatrix inv_Qj = inverse(Qj);

            const TinyVector<1> Bj(F[prev_cell_id][i] - F[cell_id][i]);

            const TinyVector<1> Uj = inv_Qj * Bj;

            Flux_F[right_node_id][i] = Uj[0] * right_minus_xj + F[cell_id][i];

            if (cell_id == m_mesh.numberOfCells() - 1) {
              Flux_F[first_node_id][i] = Flux_F[right_node_id][i];
            }
          } else {
            Flux_F[right_node_id][i] = 0;
            Flux_F[left_node_id][i]  = 0;
          }
        }
      });

    return Flux_F;
  }

  NodeArray<double>
  compute_Polynomial_reconstruction_3(const DiscreteFunctionP0Vector<const double>& F)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    NodeArray<double> Flux_F{m_mesh.connectivity(), nb_waves};
    Flux_F.fill(0);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      const NodeId first_node_id = 0;
      const NodeId last_node_id  = m_mesh.numberOfNodes() - 1;

      const double xmin = m_xr[first_node_id][0];
      const double xmax = m_xr[last_node_id][0];

      const size_t N = m_mesh.numberOfCells();

      const auto& cell_to_node   = cell_to_node_matrix[cell_id];
      const NodeId left_node_id  = cell_to_node[0];
      const NodeId right_node_id = cell_to_node[1];

      const CellId prev_cell_id = (cell_id + N - 1) % N;
      const CellId next_cell_id = (cell_id + 1) % N;

      const double xj = m_xj[cell_id][0];

      const auto& prev_cell_to_node = cell_to_node_matrix[prev_cell_id];
      const NodeId left2_node_id    = prev_cell_to_node[0];

      const auto& next_cell_to_node = cell_to_node_matrix[next_cell_id];
      const NodeId right2_node_id   = next_cell_to_node[1];

      const double left_node  = m_xr[left_node_id][0];
      const double right_node = m_xr[right_node_id][0];

      const double left2_node  = (-1 * (cell_id < 1) * (xmax - xmin)) + m_xr[left2_node_id][0];
      const double right2_node = ((cell_id + 1 >= N) * (xmax - xmin)) + m_xr[right2_node_id][0];

      const double left_minus_xj  = left_node - xj;
      const double right_minus_xj = right_node - xj;

      const double left_minus_xj_3  = left_minus_xj * left_minus_xj * left_minus_xj;
      const double right_minus_xj_3 = right_minus_xj * right_minus_xj * right_minus_xj;

      const double left2_node_xj  = left2_node - xj;
      const double right2_node_xj = right2_node - xj;

      const double left2_minus_xj_3  = left2_node_xj * left2_node_xj * left2_node_xj;
      const double right2_minus_xj_3 = right2_node_xj * right2_node_xj * right2_node_xj;

      constexpr double one_third = 1. / 3;
      constexpr double one_half  = 1. / 2;

      const TinyMatrix<2> Qj(one_third * m_inv_Vj[prev_cell_id] * (left_minus_xj_3 - left2_minus_xj_3) -
                               one_third * m_inv_Vj[cell_id] * (right_minus_xj_3 - left_minus_xj_3),
                             one_half * m_inv_Vj[prev_cell_id] *
                               ((left_minus_xj) * (left_minus_xj) - (left2_node - xj) * (left2_node - xj)),
                             one_third * m_inv_Vj[next_cell_id] * (right2_minus_xj_3 - right_minus_xj_3) -
                               one_third * m_inv_Vj[cell_id] * (right_minus_xj_3 - left_minus_xj_3),
                             one_half * m_inv_Vj[next_cell_id] *
                               ((right2_node - xj) * (right2_node - xj) - (right_minus_xj) * (right_minus_xj)));

      TinyMatrix inv_Qj = inverse(Qj);

      for (size_t i = 0; i < nb_waves; ++i) {
        const TinyVector<2> Bj(F[prev_cell_id][i] - F[cell_id][i], F[next_cell_id][i] - F[cell_id][i]);

        const TinyVector<2> Uj = inv_Qj * Bj;

        if (m_lambda_vector[i] < 0) {
          Flux_F[left_node_id][i] = Uj[0] * ((left_minus_xj) * (left_minus_xj)-one_third * m_inv_Vj[cell_id] *
                                             (right_minus_xj_3 - left_minus_xj_3)) +
                                    Uj[1] * (left_minus_xj) + F[cell_id][i];

          if (cell_id == 0) {
            Flux_F[last_node_id][i] = Flux_F[left_node_id][i];
          }

        } else if (m_lambda_vector[i] > 0) {
          Flux_F[right_node_id][i] = Uj[0] * ((right_minus_xj) * (right_minus_xj)-one_third * m_inv_Vj[cell_id] *
                                              (right_minus_xj_3 - left_minus_xj_3)) +
                                     Uj[1] * (right_minus_xj) + F[cell_id][i];

          if (cell_id == m_mesh.numberOfCells() - 1) {
            Flux_F[first_node_id][i] = Flux_F[right_node_id][i];
          }
        } else {
          Flux_F[right_node_id][i] = 0;
          Flux_F[left_node_id][i]  = 0;
        }
      }
    }

    return Flux_F;
  }

  NodeArray<double>
  TVD(NodeArray<double>& Flux_F, const DiscreteFunctionP0Vector<const double>& F)
  {
    const size_t nb_waves = m_lambda_vector.size();
    // return Flux_F;

    for (NodeId node_id = 0; node_id < m_mesh.numberOfNodes(); ++node_id) {
      const size_t N = m_mesh.numberOfCells();

      // const CellId cell_id_jp2 = (next_node_to_cell[0] + 1) % N;

      for (size_t i_wave = 0; i_wave < nb_waves; ++i_wave) {
        if (m_lambda_vector[i_wave] > 0) {
          const size_t j_p_half = node_id;

          const CellId cell_id_jm1 = (j_p_half + N - 2) % N;
          const CellId cell_id_j   = (j_p_half + N - 1) % N;
          const CellId cell_id_jp1 = (j_p_half + N) % N;

          double Fmin = std::min(F[cell_id_j][i_wave], F[cell_id_jp1][i_wave]);
          double Fmax = std::max(F[cell_id_j][i_wave], F[cell_id_jp1][i_wave]);

          double nu   = m_lambda_vector[i_wave] * m_dt / (m_Vj[cell_id_j]);
          double F_ul = F[cell_id_j][i_wave] + (1 - nu) / nu * (F[cell_id_j][i_wave] - F[cell_id_jm1][i_wave]);

          double Fmin_ul = std::min(F[cell_id_j][i_wave], F_ul);
          double Fmax_ul = std::max(F[cell_id_j][i_wave], F_ul);

          Fmin = std::max(Fmin, Fmin_ul);
          Fmax = std::min(Fmax, Fmax_ul);

          if (Flux_F[node_id][i_wave] < Fmin) {
            Flux_F[node_id][i_wave] = Fmin;
          } else if (Flux_F[node_id][i_wave] > Fmax) {
            Flux_F[node_id][i_wave] = Fmax;
          }
        } else if (m_lambda_vector[i_wave] < 0) {
          const size_t j_p_half = node_id;

          const CellId cell_id_j   = (j_p_half + N - 1) % N;
          const CellId cell_id_jp1 = (j_p_half + N) % N;
          const CellId cell_id_jp2 = (j_p_half + N + 1) % N;

          double Fmin = std::min(F[cell_id_j][i_wave], F[cell_id_jp1][i_wave]);
          double Fmax = std::max(F[cell_id_j][i_wave], F[cell_id_jp1][i_wave]);

          double nu   = m_lambda_vector[i_wave] * m_dt / (m_Vj[cell_id_j]);
          double F_ul = F[cell_id_jp1][i_wave] - (1 + nu) / nu * (F[cell_id_jp1][i_wave] - F[cell_id_jp2][i_wave]);

          double Fmin_ul = std::min(F[cell_id_jp1][i_wave], F_ul);
          double Fmax_ul = std::max(F[cell_id_jp1][i_wave], F_ul);

          Fmin = std::max(Fmin, Fmin_ul);
          Fmax = std::min(Fmax, Fmax_ul);

          if (Flux_F[node_id][i_wave] < Fmin) {
            Flux_F[node_id][i_wave] = Fmin;
          } else if (Flux_F[node_id][i_wave] > Fmax) {
            Flux_F[node_id][i_wave] = Fmax;
          }
        }
      }
    }
    return Flux_F;
  }

  std::vector<CellValue<double>>
  compute_PFnp1(const DiscreteFunctionP0Vector<const double>& F0,
                std::vector<DiscreteFunctionP0Vector<double>>& deltaF,
                DiscreteFunctionP0Vector<double>& deltaF0,
                const TinyMatrix<1>& B,
                const TinyVector<1>& b0,   //  !! Impossible to put nb_subtimesteps for the dimension
                const double& dt,
                const CellValue<double>& Mj)
  {
    const size_t nb_subtimesteps = deltaF.size();
    std::vector<CellValue<double>> PFnp1;
    for (size_t i_subtimestep = 0; i_subtimestep < nb_subtimesteps; ++i_subtimestep) {
      PFnp1.emplace_back(m_mesh.connectivity());
    }
    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      const double c = dt / Mj[cell_id];
      for (size_t i_subtimestep = 0; i_subtimestep < nb_subtimesteps; ++i_subtimestep) {
        double PFn = 0;
        for (size_t i_wave = 0; i_wave < m_lambda_vector.size(); ++i_wave) {
          PFn += F0[cell_id][i_wave] -
                 c * m_lambda_vector[i_wave] *
                   (B(i_subtimestep, 0) * deltaF[0][cell_id][i_wave] + b0[i_subtimestep] * deltaF0[cell_id][i_wave]);
        }

        PFnp1[i_subtimestep][cell_id] = PFn;
      }
    }
    return PFnp1;
  }

  std::tuple<std::shared_ptr<const MeshVariant>,
             std::shared_ptr<const DiscreteFunctionVariant>,
             std::shared_ptr<const DiscreteFunctionVariant>,
             std::shared_ptr<const DiscreteFunctionVariant>>
  apply()
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    SmallArray<const double> B;
    SmallArray<const double> c;

    if (m_timeOrder == 1) {
      SmallArray<double> tmp_B(1);
      SmallArray<double> tmp_c(1);
      tmp_B[0] = 1;
      tmp_c[0] = 1;

      B = tmp_B;
      c = tmp_c;
    } else if (m_timeOrder == 2) {
      SmallArray<double> tmp_B(2);
      SmallArray<double> tmp_c(2);

      tmp_B[0] = 0.5;
      tmp_B[1] = 0.5;
      tmp_c[0] = 0.5;
      tmp_c[1] = 1;

      B = tmp_B;
      c = tmp_c;
    } else {
      throw NotImplementedError("Order higher than two in time not implemented");
    }

    CellValue<double> tau{m_mesh.connectivity()};
    CellValue<TinyVector<Dimension>> u{m_mesh.connectivity()};
    CellValue<double> E{m_mesh.connectivity()};

    tau.fill(0);
    u.fill(zero);
    E.fill(0);
    parallel_for(
      m_mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
        for (size_t i_wave = 0; i_wave < nb_waves; ++i_wave) {
          tau[cell_id] += m_Fn_tau[cell_id][i_wave];
          u[cell_id][0] += m_Fn_u[cell_id][i_wave];
          E[cell_id] += m_Fn_E[cell_id][i_wave];
        }
      });

    CellValue<double> tau0              = copy(tau);
    CellValue<TinyVector<Dimension>> u0 = copy(u);
    CellValue<double> E0                = copy(E);

    DiscreteFunctionP0Vector<double> F_tau(p_mesh, nb_waves);
    DiscreteFunctionP0Vector<double> F_u(p_mesh, nb_waves);
    DiscreteFunctionP0Vector<double> F_E(p_mesh, nb_waves);

    F_tau = copy(m_Fn_tau);
    F_u   = copy(m_Fn_u);
    F_E   = copy(m_Fn_E);

    std::tuple<DiscreteFunctionP0Vector<double>, DiscreteFunctionP0Vector<double>, DiscreteFunctionP0Vector<double>>
      vec_F(F_tau, F_u, F_E);

    const CellValue<double> Mj{m_mesh.connectivity()};

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      Mj[cell_id] = m_Vj[cell_id] / tau[cell_id];
    }

    NodeArray<TinyVector<Dimension>> u_flux(m_mesh.connectivity(), m_timeOrder);
    NodeArray<double> p_flux(m_mesh.connectivity(), m_timeOrder);
    NodeArray<TinyVector<Dimension>> pu_flux(m_mesh.connectivity(), m_timeOrder);

    for (size_t i_time = 0; i_time < m_timeOrder; ++i_time) {
      vec_F = compute_M(tau, u, E);
      F_tau = get<0>(vec_F);
      F_u   = get<1>(vec_F);
      F_E   = get<2>(vec_F);

      NodeArray<double> Flux_Fn_tau(m_mesh.connectivity(), nb_waves);
      NodeArray<double> Flux_Fn_u(m_mesh.connectivity(), nb_waves);
      NodeArray<double> Flux_Fn_E(m_mesh.connectivity(), nb_waves);

      if (m_spaceOrder == 1) {
        Flux_Fn_tau = compute_Flux1(F_tau);
        Flux_Fn_u   = compute_Flux1(F_u);
        Flux_Fn_E   = compute_Flux1(F_E);
      } else if (m_spaceOrder == 2) {
        Flux_Fn_tau = compute_Polynomial_reconstruction_2(F_tau);
        Flux_Fn_u   = compute_Polynomial_reconstruction_2(F_u);
        Flux_Fn_E   = compute_Polynomial_reconstruction_2(F_E);
      } else if (m_spaceOrder == 3) {
        Flux_Fn_tau = compute_Polynomial_reconstruction_3(F_tau);
        Flux_Fn_u   = compute_Polynomial_reconstruction_3(F_u);
        Flux_Fn_E   = compute_Polynomial_reconstruction_3(F_E);
      } else {
        throw NotImplementedError("Only implemented for space orders 1, 2 and 3");
      }

      Flux_Fn_tau = TVD(Flux_Fn_tau, m_Fn_tau);
      Flux_Fn_u   = TVD(Flux_Fn_u, m_Fn_u);
      Flux_Fn_E   = TVD(Flux_Fn_E, m_Fn_E);

      if (i_time == 0) {
        for (size_t j_time = 0; j_time < m_timeOrder; ++j_time) {
          for (NodeId node_id = 0; node_id < m_mesh.numberOfNodes(); ++node_id) {
            u_flux[node_id][j_time][0]  = 0;
            p_flux[node_id][j_time]     = 0;
            pu_flux[node_id][j_time][0] = 0;
            for (size_t i_wave = 0; i_wave < nb_waves; ++i_wave) {
              u_flux[node_id][j_time][0] -= m_lambda_vector[i_wave] * Flux_Fn_tau[node_id][i_wave];
              p_flux[node_id][j_time] += m_lambda_vector[i_wave] * Flux_Fn_u[node_id][i_wave];
              pu_flux[node_id][j_time][0] += m_lambda_vector[i_wave] * Flux_Fn_E[node_id][i_wave];
            }
          }
        }
      } else {
        for (NodeId node_id = 0; node_id < m_mesh.numberOfNodes(); ++node_id) {
          u_flux[node_id][i_time][0]  = 0;
          p_flux[node_id][i_time]     = 0;
          pu_flux[node_id][i_time][0] = 0;
          for (size_t i_wave = 0; i_wave < nb_waves; ++i_wave) {
            u_flux[node_id][i_time][0] -= m_lambda_vector[i_wave] * Flux_Fn_tau[node_id][i_wave];
            p_flux[node_id][i_time] += m_lambda_vector[i_wave] * Flux_Fn_u[node_id][i_wave];
            pu_flux[node_id][i_time][0] += m_lambda_vector[i_wave] * Flux_Fn_E[node_id][i_wave];
          }
        }
      }

      for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
        const double dt_over_Mj    = (m_dt / Mj[cell_id]);
        const auto& cell_nodes     = cell_to_node_matrix[cell_id];
        const NodeId left_node_id  = cell_nodes[0];
        const NodeId right_node_id = cell_nodes[1];
        tau[cell_id]               = tau0[cell_id];
        u[cell_id][0]              = u0[cell_id][0];
        E[cell_id]                 = E0[cell_id];
        for (size_t j_time = 0; j_time < m_timeOrder; ++j_time) {
          tau[cell_id] += B[j_time] * dt_over_Mj * (u_flux[right_node_id][j_time][0] - u_flux[left_node_id][j_time][0]);
          u[cell_id][0] -= B[j_time] * dt_over_Mj * (p_flux[right_node_id][j_time] - p_flux[left_node_id][j_time]);

          E[cell_id] -= B[j_time] * dt_over_Mj * (pu_flux[right_node_id][j_time][0] - pu_flux[left_node_id][j_time][0]);
        }
      }
    }

    NodeValue<TinyVector<Dimension>> new_xr = copy(m_mesh.xr());
    parallel_for(
      m_mesh.numberOfNodes(), PUGS_LAMBDA(NodeId node_id) {
        for (size_t j_time = 0; j_time < m_timeOrder; ++j_time) {
          new_xr[node_id][0] += B[j_time] * m_dt * u_flux[node_id][j_time][0];
        }
      });

    std::shared_ptr<const MeshType> new_mesh = std::make_shared<MeshType>(m_mesh.shared_connectivity(), new_xr);

    // CellValue<const double> new_Vj = MeshDataManager::instance().getMeshData(*new_mesh).Vj();

    // parallel_for(
    //   m_mesh.numberOfCells(), PUGS_LAMBDA(CellId cell_id) { tau[cell_id] = new_Vj[cell_id] / Mj[cell_id]; });

    return std::make_tuple(std::make_shared<MeshVariant>(new_mesh),
                           std::make_shared<DiscreteFunctionVariant>(DiscreteFunctionP0<const double>(new_mesh, tau)),
                           std::make_shared<DiscreteFunctionVariant>(
                             DiscreteFunctionP0<const TinyVector<Dimension>>(new_mesh, u)),
                           std::make_shared<DiscreteFunctionVariant>(DiscreteFunctionP0<const double>(new_mesh, E)));
  }

  EulerKineticSolverLagrangeFV(std::shared_ptr<const MeshType> mesh,
                               const double& dt,
                               const double& gamma,
                               const std::vector<double>& lambda_vector,
                               const size_t& space_order,
                               const size_t& time_order,
                               const DiscreteFunctionP0Vector<const double>& Fn_tau,
                               const DiscreteFunctionP0Vector<const double>& Fn_u,
                               const DiscreteFunctionP0Vector<const double>& Fn_E)
    : m_dt{dt},
      m_lambda_vector{lambda_vector},
      m_spaceOrder{space_order},
      m_timeOrder{time_order},
      m_Fn_tau{Fn_tau},
      m_Fn_u{Fn_u},
      m_Fn_E{Fn_E},
      m_gamma{gamma},
      p_mesh{mesh},
      m_mesh{*mesh}
  {
    if ((lambda_vector.size() != Fn_tau.size()) or
        (lambda_vector.size() != Fn_u.size() or (lambda_vector.size() != Fn_E.size()))) {
      throw NormalError("Incompatible dimensions of lambda vector and Fn");
    }

    m_inv_Vj = [&] {
      CellValue<double> inv_Vj{m_mesh.connectivity()};
      parallel_for(
        m_mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) { inv_Vj[cell_id] = 1. / m_Vj[cell_id]; });
      return inv_Vj;
    }();
  }

  ~EulerKineticSolverLagrangeFV() = default;
};

std::tuple<std::shared_ptr<const MeshVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
eulerKineticSolverLagrangeFV(const double& dt,
                             const double& gamma,
                             const std::vector<double>& lambda_vector,
                             const size_t& space_order,
                             const size_t& time_order,
                             const double& eps,
                             const std::shared_ptr<const DiscreteFunctionVariant>& F_tau,
                             const std::shared_ptr<const DiscreteFunctionVariant>& F_u,
                             const std::shared_ptr<const DiscreteFunctionVariant>& F_E)
{
  const std::shared_ptr<const MeshVariant> mesh_v = getCommonMesh({F_tau, F_u, F_E});
  if (mesh_v.use_count() == 0) {
    throw NormalError("F_rho, F_rho_u and F_rho_E have to be defined on the same mesh");
    if (eps != 0) {
      throw NotImplementedError("Euler kinetic solver Lagrange not implemented for eps != 0");
    }
  }

  return std::visit(
    [&](auto&& p_mesh)
      -> std::tuple<std::shared_ptr<const MeshVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                    std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>> {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      if constexpr (is_polygonal_mesh_v<MeshType> and (MeshType::Dimension == 1)) {
        auto Fn_tau = F_tau->get<DiscreteFunctionP0Vector<const double>>();
        auto Fn_u   = F_u->get<DiscreteFunctionP0Vector<const double>>();
        auto Fn_E   = F_E->get<DiscreteFunctionP0Vector<const double>>();

        EulerKineticSolverLagrangeFV solver(p_mesh, dt, gamma, lambda_vector, space_order, time_order, Fn_tau, Fn_u,
                                            Fn_E);

        return solver.apply();
      } else {
        throw NotImplementedError("Invalid mesh type for Scalar Kinetic solver");
      }
    },
    mesh_v->variant());
}
