#include <scheme/EulerKineticSolverMoodFV.hpp>

#include <language/utils/EvaluateAtPoints.hpp>
#include <mesh/Connectivity.hpp>
#include <mesh/Mesh.hpp>
#include <mesh/MeshVariant.hpp>
#include <scheme/DiscreteFunctionDescriptorP0Vector.hpp>
#include <scheme/DiscreteFunctionP0Vector.hpp>
#include <scheme/DiscreteFunctionUtils.hpp>
#include <scheme/IDiscreteFunctionDescriptor.hpp>

#include <analysis/GaussLegendreQuadratureDescriptor.hpp>
#include <analysis/QuadratureManager.hpp>
#include <geometry/LineTransformation.hpp>
#include <tuple>

template <MeshConcept MeshType>
std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticExactSolutionMoodFV(const std::shared_ptr<const MeshType> p_mesh, const double t)
{
  DiscreteFunctionP0<double> rho_ex(p_mesh);
  DiscreteFunctionP0<double> u_ex(p_mesh);

  MeshData<MeshType>& mesh_data                               = MeshDataManager::instance().getMeshData(*p_mesh);
  const auto& cell_to_node_matrix                             = p_mesh->connectivity().cellToNodeMatrix();
  const CellValue<const TinyVector<MeshType::Dimension>> m_xj = mesh_data.xj();
  const NodeValue<const TinyVector<MeshType::Dimension>> m_xr = p_mesh->xr();
  const CellValue<const double> m_dx_table                    = mesh_data.Vj();

  const double pi = acos(-1);
  // EvaluateAtPoints<double(double)>::evaluateTo(density_function_id, x, rho0);

  for (CellId cell_id = 0; cell_id < p_mesh->numberOfCells(); ++cell_id) {
    const auto& cell_to_node   = cell_to_node_matrix[cell_id];
    const NodeId left_node_id  = cell_to_node[0];
    const NodeId right_node_id = cell_to_node[1];
    const double dx            = m_dx_table[cell_id];

    const LineTransformation<1> line_transformation(m_xr[left_node_id], m_xr[right_node_id]);
    QuadratureFormula<1> quadrature =
      QuadratureManager::instance().getLineFormula(GaussLegendreQuadratureDescriptor(2));

    rho_ex[cell_id] = 0;
    u_ex[cell_id]   = 0;

    for (size_t i_gauss = 0; i_gauss < quadrature.numberOfPoints(); ++i_gauss) {
      double gauss_point = line_transformation(quadrature.point(i_gauss))[0];

      double a1      = -1.5;
      double b1      = 1.5;
      double c1      = 0.5 * (a1 + b1);
      double x1      = c1;
      double rho0_x1 = 1 + 0.5 * std::sin(2 * pi * x1);
      double rho0_a1 = 1 + 0.5 * std::sin(2 * pi * a1);
      double rho0_b1 = 1 + 0.5 * std::sin(2 * pi * b1);
      double rho0_c1 = 1 + 0.5 * std::sin(2 * pi * c1);

      double fx1 = gauss_point + sqrt(3) * rho0_x1 * t - x1;
      double fa1 = gauss_point + sqrt(3) * rho0_a1 * t - a1;
      double fb1 = gauss_point + sqrt(3) * rho0_b1 * t - b1;
      double fc1 = gauss_point + sqrt(3) * rho0_c1 * t - c1;

      while (std::abs(fx1) > 1E-15) {
        if (fa1 * fc1 < 0) {
          b1 = c1;
          c1 = 0.5 * (a1 + b1);
        } else if (fc1 * fb1 < 0) {
          a1 = c1;
          c1 = 0.5 * (a1 + b1);
        }
        x1 = c1;

        rho0_x1 = 1 + 0.5 * std::sin(2 * pi * x1);
        rho0_a1 = 1 + 0.5 * std::sin(2 * pi * a1);
        rho0_b1 = 1 + 0.5 * std::sin(2 * pi * b1);
        rho0_c1 = 1 + 0.5 * std::sin(2 * pi * c1);

        fx1 = gauss_point + sqrt(3) * rho0_x1 * t - x1;
        fa1 = gauss_point + sqrt(3) * rho0_a1 * t - a1;
        fb1 = gauss_point + sqrt(3) * rho0_b1 * t - b1;
        fc1 = gauss_point + sqrt(3) * rho0_c1 * t - c1;
      }

      double a2      = -1.5;
      double b2      = 1.5;
      double c2      = 0.5 * (a2 + b2);
      double x2      = c2;
      double rho0_x2 = 1 + 0.5 * std::sin(2 * pi * x2);
      double rho0_a2 = 1 + 0.5 * std::sin(2 * pi * a2);
      double rho0_b2 = 1 + 0.5 * std::sin(2 * pi * b2);
      double rho0_c2 = 1 + 0.5 * std::sin(2 * pi * c2);

      double fx2 = gauss_point - sqrt(3) * rho0_x2 * t - x2;
      double fa2 = gauss_point - sqrt(3) * rho0_a2 * t - a2;
      double fb2 = gauss_point - sqrt(3) * rho0_b2 * t - b2;
      double fc2 = gauss_point - sqrt(3) * rho0_c2 * t - c2;

      while (std::abs(fx2) > 1E-15) {
        if (fa2 * fc2 < 0) {
          b2 = c2;
          c2 = 0.5 * (a2 + b2);
        } else if (fc2 * fb2 < 0) {
          a2 = c2;
          c2 = 0.5 * (a2 + b2);
        }
        x2 = c2;

        rho0_x2 = 1 + 0.5 * std::sin(2 * pi * x2);
        rho0_a2 = 1 + 0.5 * std::sin(2 * pi * a2);
        rho0_b2 = 1 + 0.5 * std::sin(2 * pi * b2);
        rho0_c2 = 1 + 0.5 * std::sin(2 * pi * c2);

        fx2 = gauss_point - sqrt(3) * rho0_x2 * t - x2;
        fa2 = gauss_point - sqrt(3) * rho0_a2 * t - a2;
        fb2 = gauss_point - sqrt(3) * rho0_b2 * t - b2;
        fc2 = gauss_point - sqrt(3) * rho0_c2 * t - c2;
      }

      rho0_x1 = 1 + 0.5 * std::sin(2 * pi * x1);
      rho0_x2 = 1 + 0.5 * std::sin(2 * pi * x2);

      rho_ex[cell_id] +=
        quadrature.weight(i_gauss) * line_transformation.jacobianDeterminant() * (0.5 * (rho0_x1 + rho0_x2));
      u_ex[cell_id] += quadrature.weight(i_gauss) * line_transformation.jacobianDeterminant() *
                       (sqrt(3) * ((0.5 * (rho0_x1 + rho0_x2)) - rho0_x1));
    }
    rho_ex[cell_id] /= dx;
    u_ex[cell_id] /= dx;
  }

  return std::make_tuple(std::make_shared<DiscreteFunctionVariant>(rho_ex),
                         std::make_shared<DiscreteFunctionVariant>(u_ex));
}

std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticExactSolutionMoodFV(const double& t,
                                   const std::shared_ptr<const DiscreteFunctionVariant>& rho_ex,
                                   const std::shared_ptr<const DiscreteFunctionVariant>& u_ex)
{
  const std::shared_ptr<const MeshVariant> mesh_v = getCommonMesh({u_ex, rho_ex});
  if (mesh_v.use_count() == 0) {
    throw NormalError("rho_ex and u_ex have to be defined on the same mesh");
  }

  return std::visit(
    [&](auto&& p_mesh)
      -> std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>> {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      if constexpr (is_polygonal_mesh_v<MeshType> and (MeshType::Dimension == 1)) {
        return getEulerKineticExactSolutionMoodFV(p_mesh, t);
      } else {
        throw NotImplementedError("Invalid mesh type for Scalar Kinetic solver");
      }
    },
    mesh_v->variant());
}

template <MeshConcept MeshType>
std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticWavesMoodFV(const std::shared_ptr<const MeshType> p_mesh,
                           const std::vector<double> lambda_vector,
                           DiscreteFunctionP0<const double> rho,
                           DiscreteFunctionP0<const double> rho_u,
                           DiscreteFunctionP0<const double> rho_E,
                           const double gamma)
{
  if (lambda_vector.size() < 2) {
    throw NormalError("lambda vector must be of dimension greater than 2");
  }

  DiscreteFunctionP0Vector<double> Fn_rho(p_mesh, lambda_vector.size());
  DiscreteFunctionP0Vector<double> Fn_rho_u(p_mesh, lambda_vector.size());
  DiscreteFunctionP0Vector<double> Fn_rho_E(p_mesh, lambda_vector.size());

  CellArray<double> vec_A{p_mesh->connectivity(), 3};
  CellValue<double> p{p_mesh->connectivity()};

  for (CellId cell_id = 0; cell_id < p_mesh->numberOfCells(); ++cell_id) {
    double rho_u_2    = (rho_u[cell_id] * rho_u[cell_id]) / rho[cell_id];
    p[cell_id]        = (gamma - 1) * (rho_E[cell_id] - 0.5 * rho_u_2);
    vec_A[cell_id][0] = rho_u[cell_id];
    vec_A[cell_id][1] = rho_u_2 + p[cell_id];
    vec_A[cell_id][2] = rho_u[cell_id] / rho[cell_id] * (rho_E[cell_id] * p[cell_id]);
  }

  const size_t nb_waves           = lambda_vector.size();
  double inv_L2norm_lambda_vector = 0;
  for (size_t i = 0; i < nb_waves; ++i) {
    inv_L2norm_lambda_vector += std::pow(lambda_vector[i], 2);
  }
  inv_L2norm_lambda_vector = 1. / inv_L2norm_lambda_vector;

  for (CellId cell_id = 0; cell_id < p_mesh->numberOfCells(); ++cell_id) {
    for (size_t i = 0; i < nb_waves; ++i) {
      Fn_rho[cell_id][i] =
        (1. / nb_waves) * rho[cell_id] + inv_L2norm_lambda_vector * vec_A[cell_id][0] * lambda_vector[i];
      Fn_rho_u[cell_id][i] =
        (1. / nb_waves) * rho_u[cell_id] + inv_L2norm_lambda_vector * vec_A[cell_id][1] * lambda_vector[i];
      Fn_rho_E[cell_id][i] =
        (1. / nb_waves) * rho_E[cell_id] + inv_L2norm_lambda_vector * vec_A[cell_id][2] * lambda_vector[i];
    }
  }

  return std::make_tuple(std::make_shared<DiscreteFunctionVariant>(Fn_rho),
                         std::make_shared<DiscreteFunctionVariant>(Fn_rho_u),
                         std::make_shared<DiscreteFunctionVariant>(Fn_rho_E));
}

std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticWavesMoodFV(const std::vector<double>& lambda_vector,
                           const std::shared_ptr<const DiscreteFunctionVariant>& rho,
                           const std::shared_ptr<const DiscreteFunctionVariant>& rho_u,
                           const std::shared_ptr<const DiscreteFunctionVariant>& rho_E,
                           const double& gamma)
{
  const std::shared_ptr<const MeshVariant> mesh_v = getCommonMesh({rho, rho_u, rho_E});
  if (mesh_v.use_count() == 0) {
    throw NormalError("rho rho_u and rho_E have to be defined on the same mesh");
  }

  return std::visit(
    [&](auto&& p_mesh)
      -> std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                    std::shared_ptr<const DiscreteFunctionVariant>> {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      if constexpr (is_polygonal_mesh_v<MeshType> and (MeshType::Dimension == 1)) {
        auto rho_h   = rho->get<DiscreteFunctionP0<const double>>();
        auto rho_u_h = rho_u->get<DiscreteFunctionP0<const double>>();
        auto rho_E_h = rho_E->get<DiscreteFunctionP0<const double>>();

        return getEulerKineticWavesMoodFV(p_mesh, lambda_vector, rho_h, rho_u_h, rho_E_h, gamma);
      } else {
        throw NotImplementedError("Invalid mesh type for Scalar Kinetic solver");
      }
    },
    mesh_v->variant());
}

template <MeshConcept MeshType>
class EulerKineticSolverMoodFV
{
 private:
  constexpr static size_t Dimension = MeshType::Dimension;
  using Rd                          = TinyVector<Dimension>;

  // using DiscreteEulerFunction = DiscreteFunctionP0< double>;

  const double m_dt;
  const std::vector<double>& m_lambda_vector;
  const DiscreteFunctionP0Vector<const double> m_Fn_rho;
  const DiscreteFunctionP0Vector<const double> m_Fn_rho_u;
  const DiscreteFunctionP0Vector<const double> m_Fn_rho_E;
  const double m_gamma;
  const std::shared_ptr<const MeshType> p_mesh;
  const MeshType& m_mesh;
  const CellValue<const double> m_dx_table          = MeshDataManager::instance().getMeshData(m_mesh).Vj();
  const CellValue<const TinyVector<Dimension>> m_xj = MeshDataManager::instance().getMeshData(m_mesh).xj();
  CellValue<const double> m_inv_dx_table;
  // const NodeValue<const TinyVector<Dimension>> m_xr = MeshDataManager::instance().getMeshData(m_mesh).xr();
  const NodeValue<const TinyVector<Dimension>> m_xr = m_mesh.xr();

 public:
  CellArray<const double>
  getA(const CellValue<const double>& rho,
       const CellValue<const double>& rho_u,
       const CellValue<const double>& rho_E) const
  {
    CellArray<double> vec_A{m_mesh.connectivity(), 3};
    CellValue<double> p{m_mesh.connectivity()};
    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      double rho_u_2    = (rho_u[cell_id] * rho_u[cell_id]) / rho[cell_id];
      p[cell_id]        = (m_gamma - 1) * (rho_E[cell_id] - 0.5 * rho_u_2);
      vec_A[cell_id][0] = rho_u[cell_id];
      vec_A[cell_id][1] = rho_u_2 + p[cell_id];
      vec_A[cell_id][2] = rho_u[cell_id] / rho[cell_id] * (rho_E[cell_id] + p[cell_id]);
    }

    return vec_A;
  }

  CellArray<const double>
  compute_M(const CellValue<const double>& u, const CellValue<const double> Au) const
  {
    const size_t nb_waves = m_lambda_vector.size();
    CellArray<double> M{m_mesh.connectivity(), nb_waves};

    double inv_L2norm_lambda_vector = 0;
    for (size_t i = 0; i < nb_waves; ++i) {
      inv_L2norm_lambda_vector += pow(m_lambda_vector[i], 2);
    }
    inv_L2norm_lambda_vector = 1. / inv_L2norm_lambda_vector;

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < nb_waves; ++i) {
        M[cell_id][i] = (1. / nb_waves) * u[cell_id] + inv_L2norm_lambda_vector * Au[cell_id] * m_lambda_vector[i];
      }
    }
    return M;
  }

  NodeArray<double>
  compute_Flux1(const DiscreteFunctionP0Vector<const double>& F)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    NodeArray<double> Flux_F{m_mesh.connectivity(), nb_waves};
    Flux_F.fill(0);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < nb_waves; ++i) {
        const size_t N = m_mesh.numberOfCells();

        const auto& cell_to_node   = cell_to_node_matrix[cell_id];
        const NodeId left_node_id  = cell_to_node[0];
        const NodeId right_node_id = cell_to_node[1];

        const CellId prev_cell_id = (cell_id + N - 1) % N;
        const CellId next_cell_id = (cell_id + 1) % N;

        if (m_lambda_vector[i] < 0) {
          Flux_F[left_node_id][i]  = F[cell_id][i];
          Flux_F[right_node_id][i] = F[next_cell_id][i];

        } else if (m_lambda_vector[i] > 0) {
          Flux_F[left_node_id][i]  = F[prev_cell_id][i];
          Flux_F[right_node_id][i] = F[cell_id][i];

        } else {
          Flux_F[right_node_id][i] = 0;
          Flux_F[left_node_id][i]  = 0;
        }
      }
    }
    return Flux_F;
  }

  // NodeArray<const double>
  // compute_Flux1(const DiscreteFunctionP0Vector< const double>& F,
  //               const NodeValue<const bool>& PF_node_is_valid)
  // {
  //   const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
  //   const size_t nb_waves           = m_lambda_vector.size();

  //   NodeArray<double> Flux_F{m_mesh.connectivity(), nb_waves};
  //   Flux_F.fill(0);

  //   for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
  //     // const double dx = m_dx_table[cell_id];
  //     for (size_t i = 0; i < nb_waves; ++i) {
  //       // const NodeId first_node_id = 0;
  //       // const NodeId last_node_id  = m_mesh.numberOfNodes() - 1;

  //       // const double xmin = m_xr[first_node_id][0];
  //       // const double xmax = m_xr[last_node_id][0];

  //       const size_t N = m_mesh.numberOfCells();

  //       const auto& cell_to_node   = cell_to_node_matrix[cell_id];
  //       const NodeId left_node_id  = cell_to_node[0];
  //       const NodeId right_node_id = cell_to_node[1];

  //       const CellId prev_cell_id = (cell_id + N - 1) % N;
  //       const CellId next_cell_id = (cell_id + 1) % N;

  //       // const double xjm1 = (-1 * (cell_id < 1) * (xmax - xmin)) + m_xj[prev_cell_id][0];
  //       // const double xj   = m_xj[cell_id][0];
  //       // const double xjp1 = ((cell_id + 1 >= N) * (xmax - xmin)) + m_xj[next_cell_id][0];

  //       if (m_lambda_vector[i] < 0) {
  //         // if (PF_node_is_valid[left_node_id] == false) {
  //         //   Flux_F[left_node_id][i] += 0.5 * dx / (xjp1 - xj) * F[cell_id][i];

  //         // } else if (PF_node_is_valid[right_node_id] == false) {
  //         //   Flux_F[right_node_id][i] += 0.5 * dx / (xjp1 - xj) * F[next_cell_id][i];
  //         // }
  //         if (PF_node_is_valid[left_node_id] == false) {
  //           Flux_F[left_node_id][i] = F[cell_id][i];
  //         } else if (PF_node_is_valid[right_node_id] == false) {
  //           Flux_F[right_node_id][i] = F[next_cell_id][i];
  //         }

  //       } else if (m_lambda_vector[i] > 0) {
  //         // if (PF_node_is_valid[left_node_id] == false) {
  //         //   Flux_F[left_node_id][i] += 0.5 * dx / (xj - xjm1) * F[prev_cell_id][i];
  //         // } else if (PF_node_is_valid[right_node_id] == false) {
  //         //   Flux_F[right_node_id][i] += 0.5 * dx / (xj - xjm1) * F[cell_id][i];
  //         // }
  //         if (PF_node_is_valid[left_node_id] == false) {
  //           Flux_F[left_node_id][i] = F[prev_cell_id][i];
  //         } else if (PF_node_is_valid[right_node_id] == false) {
  //           Flux_F[right_node_id][i] = F[cell_id][i];
  //         }
  //       } else {
  //         Flux_F[right_node_id][i] = 0;
  //         Flux_F[left_node_id][i]  = 0;
  //       }
  //     }
  //   }
  //   return Flux_F;
  // }

  NodeArray<const double>
  compute_nodeFlux2(const DiscreteFunctionP0Vector<const double>& F)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    NodeArray<double> Flux_F{m_mesh.connectivity(), nb_waves};
    Flux_F.fill(0);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < nb_waves; ++i) {
        const NodeId first_node_id = 0;
        const NodeId last_node_id  = m_mesh.numberOfNodes() - 1;

        const double xmin = m_xr[first_node_id][0];
        const double xmax = m_xr[last_node_id][0];

        const size_t N = m_mesh.numberOfCells();

        const auto& cell_to_node   = cell_to_node_matrix[cell_id];
        const NodeId left_node_id  = cell_to_node[0];
        const NodeId right_node_id = cell_to_node[1];

        const CellId prev2_cell_id = (cell_id + N - 2) % N;
        const CellId prev_cell_id  = (cell_id + N - 1) % N;
        const CellId next_cell_id  = (cell_id + 1) % N;
        const CellId next2_cell_id = (cell_id + 2) % N;

        const double xjm2 = (-1 * (cell_id < 2) * (xmax - xmin)) + m_xj[prev2_cell_id][0];
        const double xjm1 = (-1 * (cell_id < 1) * (xmax - xmin)) + m_xj[prev_cell_id][0];
        const double xj   = m_xj[cell_id][0];
        const double xjp1 = ((cell_id + 1 >= N) * (xmax - xmin)) + m_xj[next_cell_id][0];
        const double xjp2 = ((cell_id + 2 >= N) * (xmax - xmin)) + m_xj[next2_cell_id][0];

        if (m_lambda_vector[i] < 0) {
          const double left_node  = m_xr[left_node_id][0];
          const double right_node = m_xr[right_node_id][0];

          Flux_F[left_node_id][i] =
            (left_node - xjp1) / (xj - xjp1) * F[cell_id][i] + (left_node - xj) / (xjp1 - xj) * F[next_cell_id][i];
          Flux_F[right_node_id][i] = (right_node - xjp2) / (xjp1 - xjp2) * F[next_cell_id][i] +
                                     (right_node - xjp1) / (xjp2 - xjp1) * F[next2_cell_id][i];

        } else if (m_lambda_vector[i] > 0) {
          const double left_node  = m_xr[left_node_id][0];
          const double right_node = m_xr[right_node_id][0];

          Flux_F[left_node_id][i] = (left_node - xjm1) / (xjm2 - xjm1) * F[prev2_cell_id][i] +
                                    (left_node - xjm2) / (xjm1 - xjm2) * F[prev_cell_id][i];
          Flux_F[right_node_id][i] =
            (right_node - xj) / (xjm1 - xj) * F[prev_cell_id][i] + (right_node - xjm1) / (xj - xjm1) * F[cell_id][i];

        } else {
          Flux_F[right_node_id][i] = 0;
          Flux_F[left_node_id][i]  = 0;
        }
      }
    }

    return Flux_F;
  }

  NodeArray<const double>
  compute_nodeFlux2centered(const DiscreteFunctionP0Vector<const double>& F)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    NodeArray<double> Flux_F{m_mesh.connectivity(), nb_waves};
    Flux_F.fill(0);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      const NodeId first_node_id = 0;
      const NodeId last_node_id  = m_mesh.numberOfNodes() - 1;

      const double xmin = m_xr[first_node_id][0];
      const double xmax = m_xr[last_node_id][0];

      const size_t N = m_mesh.numberOfCells();

      const auto& cell_to_node   = cell_to_node_matrix[cell_id];
      const NodeId left_node_id  = cell_to_node[0];
      const NodeId right_node_id = cell_to_node[1];

      for (size_t i = 0; i < nb_waves; ++i) {
        if (m_lambda_vector[i] != 0) {
          const CellId prev_cell_id = (cell_id + N - 1) % N;
          const CellId next_cell_id = (cell_id + 1) % N;

          const double xjm1 = (-1 * (cell_id < 1) * (xmax - xmin)) + m_xj[prev_cell_id][0];
          const double xj   = m_xj[cell_id][0];
          const double xjp1 = ((cell_id + 1 >= N) * (xmax - xmin)) + m_xj[next_cell_id][0];

          const double left_node  = m_xr[left_node_id][0];
          const double right_node = m_xr[right_node_id][0];

          Flux_F[left_node_id][i] =
            (left_node - xj) / (xjm1 - xj) * F[prev_cell_id][i] + (left_node - xjm1) / (xj - xjm1) * F[cell_id][i];
          Flux_F[right_node_id][i] =
            (right_node - xjp1) / (xj - xjp1) * F[cell_id][i] + (right_node - xj) / (xjp1 - xj) * F[next_cell_id][i];

        } else {
          Flux_F[right_node_id][i] = 0;
          Flux_F[left_node_id][i]  = 0;
        }
      }
    }

    return Flux_F;
  }

  NodeArray<const double>
  compute_nodeFlux3(const DiscreteFunctionP0Vector<const double>& F)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    NodeArray<double> Flux_F{m_mesh.connectivity(), nb_waves};
    Flux_F.fill(0);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < nb_waves; ++i) {
        const NodeId first_node_id = 0;
        const NodeId last_node_id  = m_mesh.numberOfNodes() - 1;

        const double xmin = m_xr[first_node_id][0];
        const double xmax = m_xr[last_node_id][0];

        const size_t N = m_mesh.numberOfCells();

        const auto& cell_to_node   = cell_to_node_matrix[cell_id];
        const NodeId left_node_id  = cell_to_node[0];
        const NodeId right_node_id = cell_to_node[1];

        const CellId prev2_cell_id = (cell_id + N - 2) % N;
        const CellId prev_cell_id  = (cell_id + N - 1) % N;
        const CellId next_cell_id  = (cell_id + 1) % N;
        const CellId next2_cell_id = (cell_id + 2) % N;

        const double xjm2 = (-1 * (cell_id < 2) * (xmax - xmin)) + m_xj[prev2_cell_id][0];
        const double xjm1 = (-1 * (cell_id < 1) * (xmax - xmin)) + m_xj[prev_cell_id][0];
        const double xj   = m_xj[cell_id][0];
        const double xjp1 = ((cell_id + 1 >= N) * (xmax - xmin)) + m_xj[next_cell_id][0];
        const double xjp2 = ((cell_id + 2 >= N) * (xmax - xmin)) + m_xj[next2_cell_id][0];

        const double left_node  = m_xr[left_node_id][0];
        const double right_node = m_xr[right_node_id][0];

        if (m_lambda_vector[i] < 0) {
          Flux_F[left_node_id][i] =
            F[prev_cell_id][i] * (left_node - xj) * (left_node - xjp1) / ((xjm1 - xj) * (xjm1 - xjp1)) +
            F[cell_id][i] * (left_node - xjm1) * (left_node - xjp1) / ((xj - xjm1) * (xj - xjp1)) +
            F[next_cell_id][i] * (left_node - xjm1) * (left_node - xj) / ((xjp1 - xjm1) * (xjp1 - xj));
          Flux_F[right_node_id][i] =
            F[cell_id][i] * (right_node - xjp1) * (right_node - xjp2) / ((xj - xjp1) * (xj - xjp2)) +
            F[next_cell_id][i] * (right_node - xj) * (right_node - xjp2) / ((xjp1 - xj) * (xjp1 - xjp2)) +
            F[next2_cell_id][i] * (right_node - xj) * (right_node - xjp1) / ((xjp2 - xj) * (xjp2 - xjp1));

        } else if (m_lambda_vector[i] > 0) {
          Flux_F[left_node_id][i] =
            F[prev2_cell_id][i] * (left_node - xjm1) * (left_node - xj) / ((xjm2 - xjm1) * (xjm2 - xj)) +
            F[prev_cell_id][i] * (left_node - xjm2) * (left_node - xj) / ((xjm1 - xjm2) * (xjm1 - xj)) +
            F[cell_id][i] * (left_node - xjm2) * (left_node - xjm1) / ((xj - xjm2) * (xj - xjm1));
          Flux_F[right_node_id][i] =
            F[prev_cell_id][i] * (right_node - xj) * (right_node - xjp1) / ((xjm1 - xj) * (xjm1 - xjp1)) +
            F[cell_id][i] * (right_node - xjm1) * (right_node - xjp1) / ((xj - xjm1) * (xj - xjp1)) +
            F[next_cell_id][i] * (right_node - xjm1) * (right_node - xj) / ((xjp1 - xjm1) * (xjp1 - xj));

        } else {
          Flux_F[right_node_id][i] = 0;
          Flux_F[left_node_id][i]  = 0;
        }
      }
    }

    return Flux_F;
  }

  NodeArray<const double>
  compute_Polynomial_reconstruction_2(const DiscreteFunctionP0Vector<const double>& F)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    NodeArray<double> Flux_F{m_mesh.connectivity(), nb_waves};
    Flux_F.fill(0);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      const NodeId first_node_id = 0;
      const NodeId last_node_id  = m_mesh.numberOfNodes() - 1;

      const double xmin = m_xr[first_node_id][0];
      const double xmax = m_xr[last_node_id][0];

      const size_t N = m_mesh.numberOfCells();

      const auto& cell_to_node   = cell_to_node_matrix[cell_id];
      const NodeId left_node_id  = cell_to_node[0];
      const NodeId right_node_id = cell_to_node[1];

      const CellId prev_cell_id = (cell_id + N - 1) % N;
      const CellId next_cell_id = (cell_id + 1) % N;

      const double xj = m_xj[cell_id][0];
      // const double xjm1 = m_xj[prev_cell_id][0];
      // const double xjp1 = m_xj[next_cell_id][0];

      const auto& prev_cell_to_node = cell_to_node_matrix[prev_cell_id];
      const NodeId left2_node_id    = prev_cell_to_node[0];

      const auto& next_cell_to_node = cell_to_node_matrix[next_cell_id];
      const NodeId right2_node_id   = next_cell_to_node[1];

      const double left_node  = m_xr[left_node_id][0];
      const double right_node = m_xr[right_node_id][0];

      const double left2_node  = (-1 * (cell_id < 1) * (xmax - xmin)) + m_xr[left2_node_id][0];
      const double right2_node = ((cell_id + 1 >= N) * (xmax - xmin)) + m_xr[right2_node_id][0];

      const double left_minus_xj  = left_node - xj;
      const double right_minus_xj = right_node - xj;

      const double left2_node_xj  = left2_node - xj;
      const double right2_node_xj = right2_node - xj;

      constexpr double one_half = 1. / 2;

      for (size_t i = 0; i < nb_waves; ++i) {
        if (m_lambda_vector[i] < 0) {
          const TinyMatrix<1> Qj(one_half * m_inv_dx_table[next_cell_id] *
                                 ((right2_node_xj) * (right2_node_xj) - (right_minus_xj) * (right_minus_xj)));

          TinyMatrix inv_Qj = inverse(Qj);

          const TinyVector<1> Bj(F[next_cell_id][i] - F[cell_id][i]);

          const TinyVector<1> Uj = inv_Qj * Bj;

          Flux_F[left_node_id][i] = Uj[0] * left_minus_xj + F[cell_id][i];

          if (cell_id == 0) {
            Flux_F[last_node_id][i] = Flux_F[left_node_id][i];
          }

        } else if (m_lambda_vector[i] > 0) {
          const TinyMatrix<1> Qj(one_half * m_inv_dx_table[prev_cell_id] *
                                 ((left_minus_xj) * (left_minus_xj) - (left2_node_xj) * (left2_node_xj)));

          TinyMatrix inv_Qj = inverse(Qj);

          const TinyVector<1> Bj(F[prev_cell_id][i] - F[cell_id][i]);

          const TinyVector<1> Uj = inv_Qj * Bj;

          Flux_F[right_node_id][i] = Uj[0] * right_minus_xj + F[cell_id][i];

          if (cell_id == m_mesh.numberOfCells() - 1) {
            Flux_F[first_node_id][i] = Flux_F[right_node_id][i];
          }
        } else {
          Flux_F[right_node_id][i] = 0;
          Flux_F[left_node_id][i]  = 0;
        }
      }
    }

    return Flux_F;
  }

  NodeArray<const double>
  compute_Polynomial_reconstruction_3(const DiscreteFunctionP0Vector<const double>& F)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    NodeArray<double> Flux_F{m_mesh.connectivity(), nb_waves};
    Flux_F.fill(0);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < nb_waves; ++i) {
        const NodeId first_node_id = 0;
        const NodeId last_node_id  = m_mesh.numberOfNodes() - 1;

        const double xmin = m_xr[first_node_id][0];
        const double xmax = m_xr[last_node_id][0];

        const size_t N = m_mesh.numberOfCells();

        const auto& cell_to_node   = cell_to_node_matrix[cell_id];
        const NodeId left_node_id  = cell_to_node[0];
        const NodeId right_node_id = cell_to_node[1];

        const CellId prev_cell_id = (cell_id + N - 1) % N;
        const CellId next_cell_id = (cell_id + 1) % N;

        const double xj = m_xj[cell_id][0];

        const auto& prev_cell_to_node = cell_to_node_matrix[prev_cell_id];
        const NodeId left2_node_id    = prev_cell_to_node[0];

        const auto& next_cell_to_node = cell_to_node_matrix[next_cell_id];
        const NodeId right2_node_id   = next_cell_to_node[1];

        const double left_node  = m_xr[left_node_id][0];
        const double right_node = m_xr[right_node_id][0];

        const double left2_node  = (-1 * (cell_id < 1) * (xmax - xmin)) + m_xr[left2_node_id][0];
        const double right2_node = ((cell_id + 1 >= N) * (xmax - xmin)) + m_xr[right2_node_id][0];

        const TinyMatrix<2> Qj(1. / (3 * m_dx_table[prev_cell_id]) *
                                   (pow(left_node - xj, 3) - pow(left2_node - xj, 3)) -
                                 1. / (3 * m_dx_table[cell_id]) * (pow(right_node - xj, 3) - pow(left_node - xj, 3)),
                               1. / (2 * m_dx_table[prev_cell_id]) *
                                 ((left_node - xj) * (left_node - xj) - (left2_node - xj) * (left2_node - xj)),
                               1. / (3 * m_dx_table[next_cell_id]) *
                                   (pow(right2_node - xj, 3) - pow(right_node - xj, 3)) -
                                 1. / (3 * m_dx_table[cell_id]) * (pow(right_node - xj, 3) - pow(left_node - xj, 3)),
                               1. / (2 * m_dx_table[next_cell_id]) *
                                 ((right2_node - xj) * (right2_node - xj) - (right_node - xj) * (right_node - xj)));

        const TinyVector<2> Bj(F[prev_cell_id][i] - F[cell_id][i], F[next_cell_id][i] - F[cell_id][i]);

        const TinyMatrix<2> inv_Qj = inverse(Qj);
        const TinyVector<2> Uj     = inv_Qj * Bj;

        if (m_lambda_vector[i] < 0) {
          Flux_F[left_node_id][i] =
            Uj[0] * ((left_node - xj) * (left_node - xj) -
                     1. / (3 * m_dx_table[cell_id]) * (pow(right_node - xj, 3) - pow(left_node - xj, 3))) +
            Uj[1] * (left_node - xj) + F[cell_id][i];

          if (cell_id == 0) {
            Flux_F[last_node_id][i] = Flux_F[left_node_id][i];
          }

        } else if (m_lambda_vector[i] > 0) {
          Flux_F[right_node_id][i] =
            Uj[0] * ((right_node - xj) * (right_node - xj) -
                     1. / (3 * m_dx_table[cell_id]) * (pow(right_node - xj, 3) - pow(left_node - xj, 3))) +
            Uj[1] * (right_node - xj) + F[cell_id][i];

          if (cell_id == m_mesh.numberOfCells() - 1) {
            Flux_F[first_node_id][i] = Flux_F[right_node_id][i];
          }
        } else {
          Flux_F[right_node_id][i] = 0;
          Flux_F[left_node_id][i]  = 0;
        }
      }
    }

    return Flux_F;
  }

  CellValue<double>
  compute_PFnp1(const DiscreteFunctionP0Vector<const double> F0,
                const DiscreteFunctionP0Vector<const double> deltaF,
                const DiscreteFunctionP0Vector<const double> deltaF0,
                const double& dt)
  {
    CellValue<double> PFnp1{m_mesh.connectivity()};

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      const double dx = m_dx_table[cell_id];
      const double c  = 0.5 * dt / dx;
      double PFn      = 0;
      for (size_t i = 0; i < m_lambda_vector.size(); ++i) {
        PFn += F0[cell_id][i] - c * m_lambda_vector[i] * (deltaF[cell_id][i] + deltaF0[cell_id][i]);
      }

      PFnp1[cell_id] = PFn;
    }
    return PFnp1;
  }

  CellValue<double>
  compute_PFn(const DiscreteFunctionP0Vector<const double> F)
  {
    CellValue<double> PFn{m_mesh.connectivity()};

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      double PF = 0;
      for (size_t i = 0; i < m_lambda_vector.size(); ++i) {
        PF += F[cell_id][i];
      }
      PFn[cell_id] = PF;
    }
    return PFn;
  }

  std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
             std::shared_ptr<const DiscreteFunctionVariant>,
             std::shared_ptr<const DiscreteFunctionVariant>>
  apply(const double& eps)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    CellValue<bool> PFnp1_cell_is_valid{m_mesh.connectivity()};
    PFnp1_cell_is_valid.fill(true);
    NodeValue<bool> PFnp1_node_is_valid{m_mesh.connectivity()};
    PFnp1_node_is_valid.fill(true);

    DiscreteFunctionP0Vector<double> Fnp1_rho   = copy(m_Fn_rho);
    DiscreteFunctionP0Vector<double> Fnp1_rho_u = copy(m_Fn_rho_u);
    DiscreteFunctionP0Vector<double> Fnp1_rho_E = copy(m_Fn_rho_E);

    DiscreteFunctionP0Vector<double> deltaFn_rho(p_mesh, nb_waves);
    DiscreteFunctionP0Vector<double> deltaFn_rho_u(p_mesh, nb_waves);
    DiscreteFunctionP0Vector<double> deltaFn_rho_E(p_mesh, nb_waves);

    PFnp1_node_is_valid.fill(false);
    const NodeArray<const double> Flux_Fn_rho   = compute_Polynomial_reconstruction_2(m_Fn_rho);
    const NodeArray<const double> Flux_Fn_rho_u = compute_Polynomial_reconstruction_2(m_Fn_rho_u);
    const NodeArray<const double> Flux_Fn_rho_E = compute_Polynomial_reconstruction_2(m_Fn_rho_E);
    PFnp1_node_is_valid.fill(true);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      const auto& cell_to_node   = cell_to_node_matrix[cell_id];
      const NodeId left_node_id  = cell_to_node[0];
      const NodeId right_node_id = cell_to_node[1];

      for (size_t i_wave = 0; i_wave < nb_waves; ++i_wave) {
        deltaFn_rho[cell_id][i_wave]   = (Flux_Fn_rho[right_node_id][i_wave] - Flux_Fn_rho[left_node_id][i_wave]);
        deltaFn_rho_u[cell_id][i_wave] = (Flux_Fn_rho_u[right_node_id][i_wave] - Flux_Fn_rho_u[left_node_id][i_wave]);
        deltaFn_rho_E[cell_id][i_wave] = (Flux_Fn_rho_E[right_node_id][i_wave] - Flux_Fn_rho_E[left_node_id][i_wave]);
      }
    }

    DiscreteFunctionP0Vector<double> F_rho   = copy(m_Fn_rho);
    DiscreteFunctionP0Vector<double> F_rho_u = copy(m_Fn_rho_u);
    DiscreteFunctionP0Vector<double> F_rho_E = copy(m_Fn_rho_E);

    for (size_t p = 0; p < 2; ++p) {
      DiscreteFunctionP0Vector<double> deltaFnp1_rho(p_mesh, nb_waves);
      DiscreteFunctionP0Vector<double> deltaFnp1_rho_u(p_mesh, nb_waves);
      DiscreteFunctionP0Vector<double> deltaFnp1_rho_E(p_mesh, nb_waves);

      PFnp1_node_is_valid.fill(false);
      const NodeArray<const double> Flux_Fnp1_rho   = compute_Polynomial_reconstruction_2(Fnp1_rho);
      const NodeArray<const double> Flux_Fnp1_rho_u = compute_Polynomial_reconstruction_2(Fnp1_rho_u);
      const NodeArray<const double> Flux_Fnp1_rho_E = compute_Polynomial_reconstruction_2(Fnp1_rho_E);
      PFnp1_node_is_valid.fill(true);

      for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
        const auto& cell_to_node   = cell_to_node_matrix[cell_id];
        const NodeId left_node_id  = cell_to_node[0];
        const NodeId right_node_id = cell_to_node[1];

        for (size_t i_wave = 0; i_wave < nb_waves; ++i_wave) {
          deltaFnp1_rho[cell_id][i_wave] = (Flux_Fnp1_rho[right_node_id][i_wave] - Flux_Fnp1_rho[left_node_id][i_wave]);
          deltaFnp1_rho_u[cell_id][i_wave] =
            (Flux_Fnp1_rho_u[right_node_id][i_wave] - Flux_Fnp1_rho_u[left_node_id][i_wave]);
          deltaFnp1_rho_E[cell_id][i_wave] =
            (Flux_Fnp1_rho_E[right_node_id][i_wave] - Flux_Fnp1_rho_E[left_node_id][i_wave]);
        }
      }

      const CellValue<const double> PFnp1_rho   = compute_PFnp1(m_Fn_rho, deltaFnp1_rho, deltaFn_rho, m_dt);
      const CellValue<const double> PFnp1_rho_u = compute_PFnp1(m_Fn_rho_u, deltaFnp1_rho_u, deltaFn_rho_u, m_dt);
      const CellValue<const double> PFnp1_rho_E = compute_PFnp1(m_Fn_rho_E, deltaFnp1_rho_E, deltaFn_rho_E, m_dt);

      const CellValue<const double> PFn_rho   = compute_PFn(m_Fn_rho);
      const CellValue<const double> PFn_rho_u = compute_PFn(m_Fn_rho_u);
      const CellValue<const double> PFn_rho_E = compute_PFn(m_Fn_rho_E);

      const CellArray<const double> APFnp1 = getA(PFnp1_rho, PFnp1_rho_u, PFnp1_rho_E);
      const CellArray<const double> APFn   = getA(PFn_rho, PFn_rho_u, PFn_rho_E);

      const CellValue<double> APFnp1_rho{m_mesh.connectivity()};
      const CellValue<double> APFnp1_rho_u{m_mesh.connectivity()};
      const CellValue<double> APFnp1_rho_E{m_mesh.connectivity()};

      const CellValue<double> APFn_rho{m_mesh.connectivity()};
      const CellValue<double> APFn_rho_u{m_mesh.connectivity()};
      const CellValue<double> APFn_rho_E{m_mesh.connectivity()};

      for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
        APFnp1_rho[cell_id]   = APFnp1[cell_id][0];
        APFnp1_rho_u[cell_id] = APFnp1[cell_id][1];
        APFnp1_rho_E[cell_id] = APFnp1[cell_id][2];
        APFn_rho[cell_id]     = APFn[cell_id][0];
        APFn_rho_u[cell_id]   = APFn[cell_id][1];
        APFn_rho_E[cell_id]   = APFn[cell_id][2];
      }

      const CellArray<const double> MPFnp1_rho   = compute_M(PFnp1_rho, APFnp1_rho);
      const CellArray<const double> MPFnp1_rho_u = compute_M(PFnp1_rho_u, APFnp1_rho_u);
      const CellArray<const double> MPFnp1_rho_E = compute_M(PFnp1_rho_E, APFnp1_rho_E);
      const CellArray<const double> MPFn_rho     = compute_M(PFn_rho, APFn_rho);
      const CellArray<const double> MPFn_rho_u   = compute_M(PFn_rho_u, APFn_rho_u);
      const CellArray<const double> MPFn_rho_E   = compute_M(PFn_rho_E, APFn_rho_E);

      const double c1 = 1. / (2 * eps + m_dt);

      for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
        const double dx      = m_dx_table[cell_id];
        const double c2      = (eps * m_dt / dx);
        double rho_to_test   = 0;
        double rho_u_to_test = 0;
        double rho_E_to_test = 0;
        for (size_t i_wave = 0; i_wave < nb_waves; ++i_wave) {
          Fnp1_rho[cell_id][i_wave] =
            c1 * (m_dt * MPFnp1_rho[cell_id][i_wave] + 2 * eps * m_Fn_rho[cell_id][i_wave] -
                  c2 * m_lambda_vector[i_wave] * (deltaFnp1_rho[cell_id][i_wave] + deltaFn_rho[cell_id][i_wave]) +
                  m_dt * (MPFn_rho[cell_id][i_wave] - m_Fn_rho[cell_id][i_wave]));
          Fnp1_rho_u[cell_id][i_wave] =
            c1 * (m_dt * MPFnp1_rho_u[cell_id][i_wave] + 2 * eps * m_Fn_rho_u[cell_id][i_wave] -
                  c2 * m_lambda_vector[i_wave] * (deltaFnp1_rho_u[cell_id][i_wave] + deltaFn_rho_u[cell_id][i_wave]) +
                  m_dt * (MPFn_rho_u[cell_id][i_wave] - m_Fn_rho_u[cell_id][i_wave]));
          Fnp1_rho_E[cell_id][i_wave] =
            c1 * (m_dt * MPFnp1_rho_E[cell_id][i_wave] + 2 * eps * m_Fn_rho_E[cell_id][i_wave] -
                  c2 * m_lambda_vector[i_wave] * (deltaFnp1_rho_E[cell_id][i_wave] + deltaFn_rho_E[cell_id][i_wave]) +
                  m_dt * (MPFn_rho_E[cell_id][i_wave] - m_Fn_rho_E[cell_id][i_wave]));

          rho_to_test += Fnp1_rho[cell_id][i_wave];
          rho_u_to_test += Fnp1_rho_u[cell_id][i_wave];
          rho_E_to_test += Fnp1_rho_E[cell_id][i_wave];
        }

        if (rho_to_test <= 0) {
          PFnp1_cell_is_valid[cell_id] = false;

          std::cout << "cell " << cell_id << " rho is invalid. \n";
          std::cout << " rho = " << rho_to_test << "\n";
          const auto& cell_to_node = cell_to_node_matrix[cell_id];
          for (size_t i_node = 0; i_node < cell_to_node.size(); ++i_node) {
            NodeId node_id               = cell_to_node[i_node];
            PFnp1_node_is_valid[node_id] = false;
          }

        } else {
          double rho_u_2_to_test = (rho_u_to_test * rho_u_to_test) / rho_to_test;
          double p_to_test       = (m_gamma - 1) * (rho_E_to_test - 0.5 * rho_u_2_to_test);
          if (p_to_test <= 0) {
            PFnp1_cell_is_valid[cell_id] = false;
            std::cout << "cell " << cell_id << " p is invalid. \n";
            const auto& cell_to_node = cell_to_node_matrix[cell_id];
            for (size_t i_node = 0; i_node < cell_to_node.size(); ++i_node) {
              NodeId node_id               = cell_to_node[i_node];
              PFnp1_node_is_valid[node_id] = false;
            }
          }
        }
      }

      DiscreteFunctionP0Vector<double> deltaFn_low_order_rho(p_mesh, nb_waves);
      DiscreteFunctionP0Vector<double> deltaFn_low_order_rho_u(p_mesh, nb_waves);
      DiscreteFunctionP0Vector<double> deltaFn_low_order_rho_E(p_mesh, nb_waves);

      DiscreteFunctionP0Vector<double> deltaFnp1_low_order_rho(p_mesh, nb_waves);
      DiscreteFunctionP0Vector<double> deltaFnp1_low_order_rho_u(p_mesh, nb_waves);
      DiscreteFunctionP0Vector<double> deltaFnp1_low_order_rho_E(p_mesh, nb_waves);

      DiscreteFunctionP0Vector<double> corrected_deltaFn_rho(p_mesh, nb_waves);
      DiscreteFunctionP0Vector<double> corrected_deltaFn_rho_u(p_mesh, nb_waves);
      DiscreteFunctionP0Vector<double> corrected_deltaFn_rho_E(p_mesh, nb_waves);

      DiscreteFunctionP0Vector<double> corrected_deltaFnp1_rho(p_mesh, nb_waves);
      DiscreteFunctionP0Vector<double> corrected_deltaFnp1_rho_u(p_mesh, nb_waves);
      DiscreteFunctionP0Vector<double> corrected_deltaFnp1_rho_E(p_mesh, nb_waves);

      const NodeArray<const double> Flux_low_order_Fn_rho   = compute_Flux1(m_Fn_rho);
      const NodeArray<const double> Flux_low_order_Fn_rho_u = compute_Flux1(m_Fn_rho_u);
      const NodeArray<const double> Flux_low_order_Fn_rho_E = compute_Flux1(m_Fn_rho_E);

      const NodeArray<const double> Flux_low_order_Fnp1_rho   = compute_Flux1(F_rho);
      const NodeArray<const double> Flux_low_order_Fnp1_rho_u = compute_Flux1(F_rho_u);
      const NodeArray<const double> Flux_low_order_Fnp1_rho_E = compute_Flux1(F_rho_E);

      for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
        const auto& cell_to_node   = cell_to_node_matrix[cell_id];
        const NodeId left_node_id  = cell_to_node[0];
        const NodeId right_node_id = cell_to_node[1];

        if (PFnp1_node_is_valid[left_node_id] == false or PFnp1_node_is_valid[right_node_id] == false) {
          for (size_t i = 0; i < nb_waves; ++i) {
            deltaFn_low_order_rho[cell_id][i] =
              Flux_low_order_Fn_rho[right_node_id][i] - Flux_low_order_Fn_rho[left_node_id][i];
            deltaFn_low_order_rho_u[cell_id][i] =
              Flux_low_order_Fn_rho_u[right_node_id][i] - Flux_low_order_Fn_rho_u[left_node_id][i];
            deltaFn_low_order_rho_E[cell_id][i] =
              Flux_low_order_Fn_rho_E[right_node_id][i] - Flux_low_order_Fn_rho_E[left_node_id][i];

            deltaFnp1_low_order_rho[cell_id][i] =
              Flux_low_order_Fnp1_rho[right_node_id][i] - Flux_low_order_Fnp1_rho[left_node_id][i];
            deltaFnp1_low_order_rho_u[cell_id][i] =
              Flux_low_order_Fnp1_rho_u[right_node_id][i] - Flux_low_order_Fnp1_rho_u[left_node_id][i];
            deltaFnp1_low_order_rho_E[cell_id][i] =
              Flux_low_order_Fnp1_rho_E[right_node_id][i] - Flux_low_order_Fnp1_rho_E[left_node_id][i];

            corrected_deltaFn_rho[cell_id][i]   = deltaFn_low_order_rho[cell_id][i];
            corrected_deltaFn_rho_u[cell_id][i] = deltaFn_low_order_rho_u[cell_id][i];
            corrected_deltaFn_rho_E[cell_id][i] = deltaFn_low_order_rho_E[cell_id][i];

            corrected_deltaFnp1_rho[cell_id][i]   = deltaFnp1_low_order_rho[cell_id][i];
            corrected_deltaFnp1_rho_u[cell_id][i] = deltaFnp1_low_order_rho_u[cell_id][i];
            corrected_deltaFnp1_rho_E[cell_id][i] = deltaFnp1_low_order_rho_E[cell_id][i];
          }

        } else {
          for (size_t i = 0; i < nb_waves; ++i) {
            corrected_deltaFn_rho[cell_id][i]   = deltaFn_rho[cell_id][i];
            corrected_deltaFn_rho_u[cell_id][i] = deltaFn_rho_u[cell_id][i];
            corrected_deltaFn_rho_E[cell_id][i] = deltaFn_rho_E[cell_id][i];

            corrected_deltaFnp1_rho[cell_id][i]   = deltaFnp1_rho[cell_id][i];
            corrected_deltaFnp1_rho_u[cell_id][i] = deltaFnp1_rho_u[cell_id][i];
            corrected_deltaFnp1_rho_E[cell_id][i] = deltaFnp1_rho_E[cell_id][i];
          }
        }
      }

      const CellValue<const double> corrected_PFnp1_rho =
        compute_PFnp1(m_Fn_rho, corrected_deltaFnp1_rho, corrected_deltaFn_rho, m_dt);
      const CellValue<const double> corrected_PFnp1_rho_u =
        compute_PFnp1(m_Fn_rho_u, corrected_deltaFnp1_rho_u, corrected_deltaFn_rho_u, m_dt);
      const CellValue<const double> corrected_PFnp1_rho_E =
        compute_PFnp1(m_Fn_rho_E, corrected_deltaFnp1_rho_E, corrected_deltaFn_rho_E, m_dt);

      const CellArray<const double> corrected_APFnp1 =
        getA(corrected_PFnp1_rho, corrected_PFnp1_rho_u, corrected_PFnp1_rho_E);

      const CellValue<double> corrected_APFnp1_rho{m_mesh.connectivity()};
      const CellValue<double> corrected_APFnp1_rho_u{m_mesh.connectivity()};
      const CellValue<double> corrected_APFnp1_rho_E{m_mesh.connectivity()};

      for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
        corrected_APFnp1_rho[cell_id]   = corrected_APFnp1[cell_id][0];
        corrected_APFnp1_rho_u[cell_id] = corrected_APFnp1[cell_id][1];
        corrected_APFnp1_rho_E[cell_id] = corrected_APFnp1[cell_id][2];
      }

      const CellArray<const double> corrected_MPFnp1_rho   = compute_M(corrected_PFnp1_rho, corrected_APFnp1_rho);
      const CellArray<const double> corrected_MPFnp1_rho_u = compute_M(corrected_PFnp1_rho_u, corrected_APFnp1_rho_u);
      const CellArray<const double> corrected_MPFnp1_rho_E = compute_M(corrected_PFnp1_rho_E, corrected_APFnp1_rho_E);

      for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
        const double dx = m_dx_table[cell_id];
        const double c2 = (eps * m_dt / dx);

        for (size_t i = 0; i < nb_waves; ++i) {
          Fnp1_rho[cell_id][i] =
            c1 * (m_dt * corrected_MPFnp1_rho[cell_id][i] + 2 * eps * m_Fn_rho[cell_id][i] -
                  c2 * m_lambda_vector[i] * (corrected_deltaFnp1_rho[cell_id][i] + corrected_deltaFn_rho[cell_id][i]) +
                  m_dt * (MPFn_rho[cell_id][i] - m_Fn_rho[cell_id][i]));
          Fnp1_rho_u[cell_id][i] =
            c1 *
            (m_dt * corrected_MPFnp1_rho_u[cell_id][i] + 2 * eps * m_Fn_rho_u[cell_id][i] -
             c2 * m_lambda_vector[i] * (corrected_deltaFnp1_rho_u[cell_id][i] + corrected_deltaFn_rho_u[cell_id][i]) +
             m_dt * (MPFn_rho_u[cell_id][i] - m_Fn_rho_u[cell_id][i]));
          Fnp1_rho_E[cell_id][i] =
            c1 *
            (m_dt * corrected_MPFnp1_rho_E[cell_id][i] + 2 * eps * m_Fn_rho_E[cell_id][i] -
             c2 * m_lambda_vector[i] * (corrected_deltaFnp1_rho_E[cell_id][i] + corrected_deltaFn_rho_E[cell_id][i]) +
             m_dt * (MPFn_rho_E[cell_id][i] - m_Fn_rho_E[cell_id][i]));
        }
      }

      PFnp1_node_is_valid.fill(true);
      F_rho   = copy(Fnp1_rho);
      F_rho_u = copy(Fnp1_rho_u);
      F_rho_E = copy(Fnp1_rho_E);
    }

    return std::make_tuple(std::make_shared<DiscreteFunctionVariant>(Fnp1_rho),
                           std::make_shared<DiscreteFunctionVariant>(Fnp1_rho_u),
                           std::make_shared<DiscreteFunctionVariant>(Fnp1_rho_E));
  }

  EulerKineticSolverMoodFV(std::shared_ptr<const MeshType> mesh,
                           const double& dt,
                           const double& gamma,
                           const std::vector<double>& lambda_vector,
                           const DiscreteFunctionP0Vector<const double>& Fn_rho,
                           const DiscreteFunctionP0Vector<const double>& Fn_rho_u,
                           const DiscreteFunctionP0Vector<const double>& Fn_rho_E)
    : m_dt{dt},
      m_lambda_vector{lambda_vector},
      m_Fn_rho{Fn_rho},
      m_Fn_rho_u{Fn_rho_u},
      m_Fn_rho_E{Fn_rho_E},
      m_gamma{gamma},
      p_mesh{mesh},
      m_mesh{*mesh}
  {
    if ((lambda_vector.size() != Fn_rho.size()) or
        (lambda_vector.size() != Fn_rho_u.size() or (lambda_vector.size() != Fn_rho_u.size()))) {
      throw NormalError("Incompatible dimensions of lambda vector and Fn");
    }
    m_inv_dx_table = [=, this] {
      CellValue<double> inv_dx_table{m_mesh.connectivity()};
      parallel_for(
        m_mesh.numberOfCells(),
        PUGS_LAMBDA(const CellId cell_id) { inv_dx_table[cell_id] = 1. / m_dx_table[cell_id]; });
      return inv_dx_table;
    }();
  }

  ~EulerKineticSolverMoodFV() = default;
};

std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
eulerKineticSolverMoodFV(const double& dt,
                         const double& gamma,
                         const std::vector<double>& lambda_vector,
                         const double& eps,
                         const std::shared_ptr<const DiscreteFunctionVariant>& F_rho,
                         const std::shared_ptr<const DiscreteFunctionVariant>& F_rho_u,
                         const std::shared_ptr<const DiscreteFunctionVariant>& F_rho_E)
{
  const std::shared_ptr<const MeshVariant> mesh_v = getCommonMesh({F_rho, F_rho_u, F_rho_E});
  if (mesh_v.use_count() == 0) {
    throw NormalError("F_rho, F_rho_u and F_rho_E have to be defined on the same mesh");
  }

  return std::visit(
    [&](auto&& p_mesh)
      -> std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                    std::shared_ptr<const DiscreteFunctionVariant>> {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      if constexpr (is_polygonal_mesh_v<MeshType> and (MeshType::Dimension == 1)) {
        auto Fn_rho   = F_rho->get<DiscreteFunctionP0Vector<const double>>();
        auto Fn_rho_u = F_rho_u->get<DiscreteFunctionP0Vector<const double>>();
        auto Fn_rho_E = F_rho_E->get<DiscreteFunctionP0Vector<const double>>();

        EulerKineticSolverMoodFV solver(p_mesh, dt, gamma, lambda_vector, Fn_rho, Fn_rho_u, Fn_rho_E);

        return solver.apply(eps);
      } else {
        throw NotImplementedError("Invalid mesh type for Scalar Kinetic solver");
      }
    },
    mesh_v->variant());
}
