#include <scheme/EulerKineticSolverThirdOrderFV.hpp>

#include <language/utils/EvaluateAtPoints.hpp>
#include <mesh/Connectivity.hpp>
#include <mesh/Mesh.hpp>
#include <mesh/MeshVariant.hpp>
#include <scheme/DiscreteFunctionDescriptorP0Vector.hpp>
#include <scheme/DiscreteFunctionP0Vector.hpp>
#include <scheme/DiscreteFunctionUtils.hpp>
#include <scheme/IDiscreteFunctionDescriptor.hpp>

#include <analysis/GaussLegendreQuadratureDescriptor.hpp>
#include <analysis/QuadratureManager.hpp>
#include <geometry/LineTransformation.hpp>
#include <tuple>

template <MeshConcept MeshType>
std::shared_ptr<const DiscreteFunctionVariant>
refine(const std::shared_ptr<const MeshType> p_old_mesh,
       const std::shared_ptr<const MeshType> p_new_mesh,
       DiscreteFunctionP0<const double> old_u_old_mesh)
{
  DiscreteFunctionP0<double> old_u_new_mesh(p_new_mesh);

  parallel_for(
    p_old_mesh->numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
      old_u_new_mesh[2 * cell_id]     = old_u_old_mesh[cell_id];
      old_u_new_mesh[2 * cell_id + 1] = old_u_old_mesh[cell_id];
    });

  return std::make_shared<DiscreteFunctionVariant>(old_u_new_mesh);
}

std::shared_ptr<const DiscreteFunctionVariant>
refine(const std::shared_ptr<const DiscreteFunctionVariant>& old_u_old_mesh,
       const std::shared_ptr<const DiscreteFunctionVariant>& new_u_new_mesh)
{
  std::shared_ptr<const MeshVariant> old_mesh_v = getCommonMesh({old_u_old_mesh});
  std::shared_ptr<const MeshVariant> new_mesh_v = getCommonMesh({new_u_new_mesh});

  return std::visit(
    [&](auto&& p_old_mesh, auto&& p_new_mesh) -> std::shared_ptr<const DiscreteFunctionVariant> {
      using OldMeshType = mesh_type_t<decltype(p_old_mesh)>;
      using NewMeshType = mesh_type_t<decltype(p_new_mesh)>;
      if constexpr (std::is_same_v<OldMeshType, NewMeshType>) {
        using MeshType = OldMeshType;
        if constexpr (is_polygonal_mesh_v<MeshType> and (MeshType::Dimension == 1)) {
          auto old_u_old_mesh_h = old_u_old_mesh->get<DiscreteFunctionP0<const double>>();

          return refine(p_old_mesh, p_new_mesh, old_u_old_mesh_h);
        } else {
          throw NotImplementedError("Invalid mesh type for Scalar Kinetic solver");
        }
      } else {
        throw NormalError("old and new mesh are not of same kind");
      }
    },
    old_mesh_v->variant(), new_mesh_v->variant());
}

template <MeshConcept MeshType>
std::shared_ptr<const DiscreteFunctionVariant>
coarsen(const std::shared_ptr<const MeshType> p_old_mesh, DiscreteFunctionP0<const double> new_u_new_mesh)
{
  DiscreteFunctionP0<double> new_u_old_mesh(p_old_mesh);

  parallel_for(
    p_old_mesh->numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
      new_u_old_mesh[cell_id] = (new_u_new_mesh[2 * cell_id] + new_u_new_mesh[2 * cell_id + 1]) * 0.5;
    });

  return std::make_shared<DiscreteFunctionVariant>(new_u_old_mesh);
}

std::shared_ptr<const DiscreteFunctionVariant>
coarsen(const std::shared_ptr<const DiscreteFunctionVariant>& u_old_mesh,
        const std::shared_ptr<const DiscreteFunctionVariant>& new_u_new_mesh)
{
  std::shared_ptr<const MeshVariant> old_mesh_v = getCommonMesh({u_old_mesh});
  std::shared_ptr<const MeshVariant> new_mesh_v = getCommonMesh({new_u_new_mesh});

  return std::visit(
    [&](auto&& p_old_mesh, auto&& p_new_mesh) -> std::shared_ptr<const DiscreteFunctionVariant> {
      using OldMeshType = mesh_type_t<decltype(p_old_mesh)>;
      using NewMeshType = mesh_type_t<decltype(p_new_mesh)>;
      if constexpr (std::is_same_v<OldMeshType, NewMeshType>) {
        using MeshType = OldMeshType;
        if constexpr (is_polygonal_mesh_v<MeshType> and (MeshType::Dimension == 1)) {
          auto new_u_new_mesh_h = new_u_new_mesh->get<DiscreteFunctionP0<const double>>();

          return coarsen(p_old_mesh, new_u_new_mesh_h);
        } else {
          throw NotImplementedError("Invalid mesh type for Scalar Kinetic solver");
        }
      } else {
        throw NormalError("old and new mesh are not of same kind");
      }
    },
    old_mesh_v->variant(), new_mesh_v->variant());
}

template <MeshConcept MeshType>
std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticExactSolutionThirdOrderFV(const std::shared_ptr<const MeshType> p_mesh,
                                         const double t,
                                         const double gamma)
{
  DiscreteFunctionP0<double> rho_ex(p_mesh);
  DiscreteFunctionP0<double> u_ex(p_mesh);
  DiscreteFunctionP0<double> p_ex(p_mesh);
  DiscreteFunctionP0<double> rho_u_ex(p_mesh);
  DiscreteFunctionP0<double> rho_E_ex(p_mesh);

  MeshData<MeshType>& mesh_data                               = MeshDataManager::instance().getMeshData(*p_mesh);
  const auto& cell_to_node_matrix                             = p_mesh->connectivity().cellToNodeMatrix();
  const CellValue<const TinyVector<MeshType::Dimension>> m_xj = mesh_data.xj();
  const NodeValue<const TinyVector<MeshType::Dimension>> m_xr = p_mesh->xr();
  const CellValue<const double> m_dx_table                    = mesh_data.Vj();

  const double pi = acos(-1);
  // EvaluateAtPoints<double(double)>::evaluateTo(density_function_id, x, rho0);

  parallel_for(
    p_mesh->numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
      const auto& cell_to_node   = cell_to_node_matrix[cell_id];
      const NodeId left_node_id  = cell_to_node[0];
      const NodeId right_node_id = cell_to_node[1];
      const double dx            = m_dx_table[cell_id];

      const LineTransformation<1> line_transformation(m_xr[left_node_id], m_xr[right_node_id]);
      QuadratureFormula<1> quadrature =
        QuadratureManager::instance().getLineFormula(GaussLegendreQuadratureDescriptor(4));

      rho_ex[cell_id]   = 0;
      u_ex[cell_id]     = 0;
      p_ex[cell_id]     = 0;
      rho_u_ex[cell_id] = 0;
      rho_E_ex[cell_id] = 0;

      for (size_t i_gauss = 0; i_gauss < quadrature.numberOfPoints(); ++i_gauss) {
        double gauss_point = line_transformation(quadrature.point(i_gauss))[0];

        double a1      = -1.5;
        double b1      = 1.5;
        double c1      = 0.5 * (a1 + b1);
        double x1      = c1;
        double rho0_x1 = 1 + 0.5 * std::sin(2 * pi * x1);
        double rho0_a1 = 1 + 0.5 * std::sin(2 * pi * a1);
        double rho0_b1 = 1 + 0.5 * std::sin(2 * pi * b1);
        double rho0_c1 = 1 + 0.5 * std::sin(2 * pi * c1);

        double fx1 = gauss_point + std::sqrt(3) * rho0_x1 * t - x1;
        double fa1 = gauss_point + std::sqrt(3) * rho0_a1 * t - a1;
        double fb1 = gauss_point + std::sqrt(3) * rho0_b1 * t - b1;
        double fc1 = gauss_point + std::sqrt(3) * rho0_c1 * t - c1;

        while (std::abs(fx1) > 1E-15) {
          if (fa1 * fc1 < 0) {
            b1 = c1;
            c1 = 0.5 * (a1 + b1);
          } else if (fc1 * fb1 < 0) {
            a1 = c1;
            c1 = 0.5 * (a1 + b1);
          }
          x1 = c1;

          rho0_x1 = 1 + 0.5 * std::sin(2 * pi * x1);
          rho0_a1 = 1 + 0.5 * std::sin(2 * pi * a1);
          rho0_b1 = 1 + 0.5 * std::sin(2 * pi * b1);
          rho0_c1 = 1 + 0.5 * std::sin(2 * pi * c1);

          fx1 = gauss_point + std::sqrt(3) * rho0_x1 * t - x1;
          fa1 = gauss_point + std::sqrt(3) * rho0_a1 * t - a1;
          fb1 = gauss_point + std::sqrt(3) * rho0_b1 * t - b1;
          fc1 = gauss_point + std::sqrt(3) * rho0_c1 * t - c1;
        }

        double a2      = -1.5;
        double b2      = 1.5;
        double c2      = 0.5 * (a2 + b2);
        double x2      = c2;
        double rho0_x2 = 1 + 0.5 * std::sin(2 * pi * x2);
        double rho0_a2 = 1 + 0.5 * std::sin(2 * pi * a2);
        double rho0_b2 = 1 + 0.5 * std::sin(2 * pi * b2);
        double rho0_c2 = 1 + 0.5 * std::sin(2 * pi * c2);

        double fx2 = gauss_point - std::sqrt(3) * rho0_x2 * t - x2;
        double fa2 = gauss_point - std::sqrt(3) * rho0_a2 * t - a2;
        double fb2 = gauss_point - std::sqrt(3) * rho0_b2 * t - b2;
        double fc2 = gauss_point - std::sqrt(3) * rho0_c2 * t - c2;

        while (std::abs(fx2) > 1E-15) {
          if (fa2 * fc2 < 0) {
            b2 = c2;
            c2 = 0.5 * (a2 + b2);
          } else if (fc2 * fb2 < 0) {
            a2 = c2;
            c2 = 0.5 * (a2 + b2);
          }
          x2 = c2;

          rho0_x2 = 1 + 0.5 * std::sin(2 * pi * x2);
          rho0_a2 = 1 + 0.5 * std::sin(2 * pi * a2);
          rho0_b2 = 1 + 0.5 * std::sin(2 * pi * b2);
          rho0_c2 = 1 + 0.5 * std::sin(2 * pi * c2);

          fx2 = gauss_point - std::sqrt(3) * rho0_x2 * t - x2;
          fa2 = gauss_point - std::sqrt(3) * rho0_a2 * t - a2;
          fb2 = gauss_point - std::sqrt(3) * rho0_b2 * t - b2;
          fc2 = gauss_point - std::sqrt(3) * rho0_c2 * t - c2;
        }

        rho0_x1 = 1 + 0.5 * std::sin(2 * pi * x1);
        rho0_x2 = 1 + 0.5 * std::sin(2 * pi * x2);

        const double rho_gauss = 0.5 * (rho0_x1 + rho0_x2);
        const double u_gauss   = std::sqrt(3) * ((0.5 * (rho0_x1 + rho0_x2)) - rho0_x1);
        const double p_gauss = (0.5 * (rho0_x1 + rho0_x2)) * (0.5 * (rho0_x1 + rho0_x2)) * (0.5 * (rho0_x1 + rho0_x2));

        rho_u_ex[cell_id] +=
          quadrature.weight(i_gauss) * line_transformation.jacobianDeterminant() * rho_gauss * u_gauss;
        rho_E_ex[cell_id] += quadrature.weight(i_gauss) * line_transformation.jacobianDeterminant() *
                             (p_gauss / (gamma - 1) + 0.5 * rho_gauss * u_gauss * u_gauss);

        rho_ex[cell_id] += quadrature.weight(i_gauss) * line_transformation.jacobianDeterminant() * rho_gauss;
        u_ex[cell_id] += quadrature.weight(i_gauss) * line_transformation.jacobianDeterminant() * u_gauss;
        p_ex[cell_id] += quadrature.weight(i_gauss) * line_transformation.jacobianDeterminant() * p_gauss;
      }
      rho_u_ex[cell_id] /= dx;
      rho_E_ex[cell_id] /= dx;
      rho_ex[cell_id] /= dx;
      u_ex[cell_id] /= dx;
      p_ex[cell_id] /= dx;
    });

  return std::make_tuple(std::make_shared<DiscreteFunctionVariant>(rho_ex),
                         std::make_shared<DiscreteFunctionVariant>(u_ex),
                         std::make_shared<DiscreteFunctionVariant>(p_ex),
                         std::make_shared<DiscreteFunctionVariant>(rho_u_ex),
                         std::make_shared<DiscreteFunctionVariant>(rho_E_ex));
}

std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticExactSolutionThirdOrderFV(const double& t,
                                         const double& gamma,
                                         const std::shared_ptr<const DiscreteFunctionVariant>& rho_ex,
                                         const std::shared_ptr<const DiscreteFunctionVariant>& u_ex,
                                         const std::shared_ptr<const DiscreteFunctionVariant>& p_ex,
                                         const std::shared_ptr<const DiscreteFunctionVariant>& rho_u_ex,
                                         const std::shared_ptr<const DiscreteFunctionVariant>& rho_E_ex)
{
  const std::shared_ptr<const MeshVariant> mesh_v = getCommonMesh({u_ex, rho_ex, p_ex, rho_u_ex, rho_E_ex});
  if (mesh_v.use_count() == 0) {
    throw NormalError("rho_ex, u_ex, p_ex, rho_u_ex and rho_E_ex have to be defined on the same mesh");
  }

  return std::visit(
    [&](auto&& p_mesh)
      -> std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                    std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                    std::shared_ptr<const DiscreteFunctionVariant>> {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      if constexpr (is_polygonal_mesh_v<MeshType> and (MeshType::Dimension == 1)) {
        return getEulerKineticExactSolutionThirdOrderFV(p_mesh, t, gamma);
      } else {
        throw NotImplementedError("Invalid mesh type for Scalar Kinetic solver");
      }
    },
    mesh_v->variant());
}

template <MeshConcept MeshType>
std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticWavesThirdOrderFV(const std::shared_ptr<const MeshType> p_mesh,
                                 const std::vector<double> lambda_vector,
                                 DiscreteFunctionP0<const double> rho,
                                 DiscreteFunctionP0<const double> rho_u,
                                 DiscreteFunctionP0<const double> rho_E,
                                 const double gamma)
{
  if (lambda_vector.size() < 2) {
    throw NormalError("lambda vector must be of dimension greater than 2");
  }

  MeshData<MeshType>& mesh_data                               = MeshDataManager::instance().getMeshData(*p_mesh);
  const NodeValue<const TinyVector<MeshType::Dimension>> m_xr = p_mesh->xr();
  const CellValue<const TinyVector<MeshType::Dimension>> m_xj = mesh_data.xj();
  const CellValue<const double> m_dx_table                    = mesh_data.Vj();

  DiscreteFunctionP0Vector<double> Fn_rho(p_mesh, lambda_vector.size());
  DiscreteFunctionP0Vector<double> Fn_rho_u(p_mesh, lambda_vector.size());
  DiscreteFunctionP0Vector<double> Fn_rho_E(p_mesh, lambda_vector.size());

  const auto& cell_to_node_matrix = p_mesh->connectivity().cellToNodeMatrix();
  CellArray<double> vec_A{p_mesh->connectivity(), 3};
  CellValue<double> p{p_mesh->connectivity()};

  size_t Order = 3;

  if (Order != 3) {
    for (CellId cell_id = 0; cell_id < p_mesh->numberOfCells(); ++cell_id) {
      p[cell_id]        = (gamma - 1) * (rho_E[cell_id] - 0.5 * rho_u[cell_id] * rho_u[cell_id] / rho[cell_id]);
      vec_A[cell_id][0] = rho_u[cell_id];
      vec_A[cell_id][1] = rho_u[cell_id] * rho_u[cell_id] / rho[cell_id] + p[cell_id];
      vec_A[cell_id][2] = (rho_E[cell_id] + p[cell_id]) * rho_u[cell_id] / rho[cell_id];
    }
  } else {
    CellValue<double> m_inv_dx_table = [=] {
      CellValue<double> inv_dx_table{p_mesh->connectivity()};
      parallel_for(
        p_mesh->numberOfCells(),
        PUGS_LAMBDA(const CellId cell_id) { inv_dx_table[cell_id] = 1. / m_dx_table[cell_id]; });
      return inv_dx_table;
    }();

    for (CellId cell_id = 0; cell_id < p_mesh->numberOfCells(); ++cell_id) {
      const NodeId first_node_id = 0;
      const NodeId last_node_id  = p_mesh->numberOfNodes() - 1;

      const double xmin = m_xr[first_node_id][0];
      const double xmax = m_xr[last_node_id][0];

      const size_t N = p_mesh->numberOfCells();

      const auto& cell_to_node   = cell_to_node_matrix[cell_id];
      const NodeId left_node_id  = cell_to_node[0];
      const NodeId right_node_id = cell_to_node[1];

      const CellId prev_cell_id = (cell_id + N - 1) % N;
      const CellId next_cell_id = (cell_id + 1) % N;

      const double xj = m_xj[cell_id][0];

      const auto& prev_cell_to_node = cell_to_node_matrix[prev_cell_id];
      const NodeId left2_node_id    = prev_cell_to_node[0];

      const auto& next_cell_to_node = cell_to_node_matrix[next_cell_id];
      const NodeId right2_node_id   = next_cell_to_node[1];

      const double left_node  = m_xr[left_node_id][0];
      const double right_node = m_xr[right_node_id][0];

      const double left2_node  = (-1 * (cell_id < 1) * (xmax - xmin)) + m_xr[left2_node_id][0];
      const double right2_node = ((cell_id + 1 >= N) * (xmax - xmin)) + m_xr[right2_node_id][0];

      const double left_minus_xj  = left_node - xj;
      const double right_minus_xj = right_node - xj;

      const double left_minus_xj_3  = left_minus_xj * left_minus_xj * left_minus_xj;
      const double right_minus_xj_3 = right_minus_xj * right_minus_xj * right_minus_xj;

      const double left2_node_xj  = left2_node - xj;
      const double right2_node_xj = right2_node - xj;

      const double left2_minus_xj_3  = left2_node_xj * left2_node_xj * left2_node_xj;
      const double right2_minus_xj_3 = right2_node_xj * right2_node_xj * right2_node_xj;

      constexpr double one_third = 1. / 3;
      constexpr double one_half  = 1. / 2;

      const TinyMatrix<2> Qj(one_third * m_inv_dx_table[prev_cell_id] * (left_minus_xj_3 - left2_minus_xj_3) -
                               one_third * m_inv_dx_table[cell_id] * (right_minus_xj_3 - left_minus_xj_3),
                             one_half * m_inv_dx_table[prev_cell_id] *
                               ((left_minus_xj) * (left_minus_xj) - (left2_node - xj) * (left2_node - xj)),
                             one_third * m_inv_dx_table[next_cell_id] * (right2_minus_xj_3 - right_minus_xj_3) -
                               one_third * m_inv_dx_table[cell_id] * (right_minus_xj_3 - left_minus_xj_3),
                             one_half * m_inv_dx_table[next_cell_id] *
                               ((right2_node - xj) * (right2_node - xj) - (right_minus_xj) * (right_minus_xj)));

      TinyMatrix inv_Qj = inverse(Qj);

      const TinyVector<2> Bj_rho(rho[prev_cell_id] - rho[cell_id], rho[next_cell_id] - rho[cell_id]);
      const TinyVector<2> Bj_rho_u(rho_u[prev_cell_id] - rho_u[cell_id], rho_u[next_cell_id] - rho_u[cell_id]);
      const TinyVector<2> Bj_rho_E(rho_E[prev_cell_id] - rho_E[cell_id], rho_E[next_cell_id] - rho_E[cell_id]);

      const TinyVector<2> Uj_rho   = inv_Qj * Bj_rho;
      const TinyVector<2> Uj_rho_u = inv_Qj * Bj_rho_u;
      const TinyVector<2> Uj_rho_E = inv_Qj * Bj_rho_E;

      const LineTransformation<1> line_transformation(m_xr[left_node_id], m_xr[right_node_id]);
      QuadratureFormula<1> quadrature =
        QuadratureManager::instance().getLineFormula(GaussLegendreQuadratureDescriptor(4));

      vec_A[cell_id][0] = 0;
      vec_A[cell_id][1] = 0;
      vec_A[cell_id][2] = 0;

      for (size_t i_gauss = 0; i_gauss < quadrature.numberOfPoints(); ++i_gauss) {
        const double gauss_point          = line_transformation(quadrature.point(i_gauss))[0];
        const double gauss_point_minus_xj = gauss_point - xj;

        const double rho_gauss = Uj_rho[0] * ((gauss_point_minus_xj) * (gauss_point_minus_xj)-one_third *
                                              m_inv_dx_table[cell_id] * (right_minus_xj_3 - left_minus_xj_3)) +
                                 Uj_rho[1] * (gauss_point_minus_xj) + rho[cell_id];
        const double rho_u_gauss = Uj_rho_u[0] * ((gauss_point_minus_xj) * (gauss_point_minus_xj)-one_third *
                                                  m_inv_dx_table[cell_id] * (right_minus_xj_3 - left_minus_xj_3)) +
                                   Uj_rho_u[1] * (gauss_point_minus_xj) + rho_u[cell_id];
        const double rho_E_gauss = Uj_rho_E[0] * ((gauss_point_minus_xj) * (gauss_point_minus_xj)-one_third *
                                                  m_inv_dx_table[cell_id] * (right_minus_xj_3 - left_minus_xj_3)) +
                                   Uj_rho_E[1] * (gauss_point_minus_xj) + rho_E[cell_id];

        const double rho_u_2_gauss = (rho_u_gauss * rho_u_gauss) / rho_gauss;
        const double p_gauss       = (gamma - 1) * (rho_E_gauss - 0.5 * rho_u_2_gauss);

        vec_A[cell_id][1] +=
          quadrature.weight(i_gauss) * line_transformation.jacobianDeterminant() * (rho_u_2_gauss + p_gauss);
        vec_A[cell_id][2] += quadrature.weight(i_gauss) * line_transformation.jacobianDeterminant() *
                             (rho_u_gauss / rho_gauss * (rho_E_gauss + p_gauss));
      }

      vec_A[cell_id][0] = rho_u[cell_id];
      vec_A[cell_id][1] /= m_dx_table[cell_id];
      vec_A[cell_id][2] /= m_dx_table[cell_id];
    }
  }
  const size_t nb_waves           = lambda_vector.size();
  double inv_L2norm_lambda_vector = 0;
  for (size_t i = 0; i < nb_waves; ++i) {
    inv_L2norm_lambda_vector += std::pow(lambda_vector[i], 2);
  }
  inv_L2norm_lambda_vector = 1. / inv_L2norm_lambda_vector;

  for (CellId cell_id = 0; cell_id < p_mesh->numberOfCells(); ++cell_id) {
    for (size_t i = 0; i < nb_waves; ++i) {
      Fn_rho[cell_id][i] =
        (1. / nb_waves) * rho[cell_id] + inv_L2norm_lambda_vector * vec_A[cell_id][0] * lambda_vector[i];
      Fn_rho_u[cell_id][i] =
        (1. / nb_waves) * rho_u[cell_id] + inv_L2norm_lambda_vector * vec_A[cell_id][1] * lambda_vector[i];
      Fn_rho_E[cell_id][i] =
        (1. / nb_waves) * rho_E[cell_id] + inv_L2norm_lambda_vector * vec_A[cell_id][2] * lambda_vector[i];
    }
  }

  return std::make_tuple(std::make_shared<DiscreteFunctionVariant>(Fn_rho),
                         std::make_shared<DiscreteFunctionVariant>(Fn_rho_u),
                         std::make_shared<DiscreteFunctionVariant>(Fn_rho_E));
}

std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticWavesThirdOrderFV(const std::vector<double>& lambda_vector,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& rho,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& rho_u,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& rho_E,
                                 const double& gamma)
{
  const std::shared_ptr<const MeshVariant> mesh_v = getCommonMesh({rho, rho_u, rho_E});
  if (mesh_v.use_count() == 0) {
    throw NormalError("rho rho_u and rho_E have to be defined on the same mesh");
  }

  return std::visit(
    [&](auto&& p_mesh)
      -> std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                    std::shared_ptr<const DiscreteFunctionVariant>> {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      if constexpr (is_polygonal_mesh_v<MeshType> and (MeshType::Dimension == 1)) {
        auto rho_h   = rho->get<DiscreteFunctionP0<const double>>();
        auto rho_u_h = rho_u->get<DiscreteFunctionP0<const double>>();
        auto rho_E_h = rho_E->get<DiscreteFunctionP0<const double>>();

        return getEulerKineticWavesThirdOrderFV(p_mesh, lambda_vector, rho_h, rho_u_h, rho_E_h, gamma);
      } else {
        throw NotImplementedError("Invalid mesh type for Scalar Kinetic solver");
      }
    },
    mesh_v->variant());
}

template <MeshConcept MeshType>
class EulerKineticSolverThirdOrderFV
{
 private:
  constexpr static size_t Dimension = MeshType::Dimension;
  using Rd                          = TinyVector<Dimension>;

  const double m_dt;
  const double m_eps;
  const std::vector<double>& m_lambda_vector;
  const size_t m_SpaceOrder;
  const size_t m_Limitation;
  const bool m_MoodLimitation;
  const DiscreteFunctionP0Vector<const double> m_Fn_rho;
  const DiscreteFunctionP0Vector<const double> m_Fn_rho_u;
  const DiscreteFunctionP0Vector<const double> m_Fn_rho_E;
  const double m_gamma;
  const std::shared_ptr<const MeshType> p_mesh;
  const MeshType& m_mesh;
  const CellValue<const double> m_dx_table = MeshDataManager::instance().getMeshData(m_mesh).Vj();
  CellValue<const double> m_inv_dx_table;
  const CellValue<const TinyVector<Dimension>> m_xj = MeshDataManager::instance().getMeshData(m_mesh).xj();
  const NodeValue<const TinyVector<Dimension>> m_xr = m_mesh.xr();

 public:
  CellArray<double>
  getA(const CellValue<const double>& rho,
       const CellValue<const double>& rho_u,
       const CellValue<const double>& rho_E,
       CellValue<bool>& cell_is_valid) const
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();

    CellArray<double> vec_A{m_mesh.connectivity(), 3};
    CellValue<double> p{m_mesh.connectivity()};
    size_t Order = m_SpaceOrder;

    parallel_for(
      m_mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
        if ((Order == 3) and cell_is_valid[cell_id]) {
          const NodeId first_node_id = 0;
          const NodeId last_node_id  = m_mesh.numberOfNodes() - 1;

          const double xmin = m_xr[first_node_id][0];
          const double xmax = m_xr[last_node_id][0];

          const size_t N = m_mesh.numberOfCells();

          const auto& cell_to_node   = cell_to_node_matrix[cell_id];
          const NodeId left_node_id  = cell_to_node[0];
          const NodeId right_node_id = cell_to_node[1];

          const CellId prev_cell_id = (cell_id + N - 1) % N;
          const CellId next_cell_id = (cell_id + 1) % N;

          const double xj = m_xj[cell_id][0];

          const auto& prev_cell_to_node = cell_to_node_matrix[prev_cell_id];
          const NodeId left2_node_id    = prev_cell_to_node[0];

          const auto& next_cell_to_node = cell_to_node_matrix[next_cell_id];
          const NodeId right2_node_id   = next_cell_to_node[1];

          const double left_node  = m_xr[left_node_id][0];
          const double right_node = m_xr[right_node_id][0];

          const double left2_node  = (-1 * (cell_id < 1) * (xmax - xmin)) + m_xr[left2_node_id][0];
          const double right2_node = ((cell_id + 1 >= N) * (xmax - xmin)) + m_xr[right2_node_id][0];

          const double left_minus_xj  = left_node - xj;
          const double right_minus_xj = right_node - xj;

          const double left_minus_xj_3  = left_minus_xj * left_minus_xj * left_minus_xj;
          const double right_minus_xj_3 = right_minus_xj * right_minus_xj * right_minus_xj;

          const double left2_node_xj  = left2_node - xj;
          const double right2_node_xj = right2_node - xj;

          const double left2_minus_xj_3  = left2_node_xj * left2_node_xj * left2_node_xj;
          const double right2_minus_xj_3 = right2_node_xj * right2_node_xj * right2_node_xj;

          constexpr double one_third = 1. / 3;
          constexpr double one_half  = 1. / 2;

          const TinyMatrix<2> Qj(one_third * m_inv_dx_table[prev_cell_id] * (left_minus_xj_3 - left2_minus_xj_3) -
                                   one_third * m_inv_dx_table[cell_id] * (right_minus_xj_3 - left_minus_xj_3),
                                 one_half * m_inv_dx_table[prev_cell_id] *
                                   ((left_minus_xj) * (left_minus_xj) - (left2_node - xj) * (left2_node - xj)),
                                 one_third * m_inv_dx_table[next_cell_id] * (right2_minus_xj_3 - right_minus_xj_3) -
                                   one_third * m_inv_dx_table[cell_id] * (right_minus_xj_3 - left_minus_xj_3),
                                 one_half * m_inv_dx_table[next_cell_id] *
                                   ((right2_node - xj) * (right2_node - xj) - (right_minus_xj) * (right_minus_xj)));

          TinyMatrix inv_Qj = inverse(Qj);

          const TinyVector<2> Bj_rho(rho[prev_cell_id] - rho[cell_id], rho[next_cell_id] - rho[cell_id]);
          const TinyVector<2> Bj_rho_u(rho_u[prev_cell_id] - rho_u[cell_id], rho_u[next_cell_id] - rho_u[cell_id]);
          const TinyVector<2> Bj_rho_E(rho_E[prev_cell_id] - rho_E[cell_id], rho_E[next_cell_id] - rho_E[cell_id]);

          const TinyVector<2> Uj_rho   = inv_Qj * Bj_rho;
          const TinyVector<2> Uj_rho_u = inv_Qj * Bj_rho_u;
          const TinyVector<2> Uj_rho_E = inv_Qj * Bj_rho_E;

          const LineTransformation<1> line_transformation(m_xr[left_node_id], m_xr[right_node_id]);
          QuadratureFormula<1> quadrature =
            QuadratureManager::instance().getLineFormula(GaussLegendreQuadratureDescriptor(4));

          vec_A[cell_id][0] = 0;
          vec_A[cell_id][1] = 0;
          vec_A[cell_id][2] = 0;

          for (size_t i_gauss = 0; i_gauss < quadrature.numberOfPoints(); ++i_gauss) {
            const double gauss_point          = line_transformation(quadrature.point(i_gauss))[0];
            const double gauss_point_minus_xj = gauss_point - xj;

            const double rho_gauss = Uj_rho[0] * ((gauss_point_minus_xj) * (gauss_point_minus_xj)-one_third *
                                                  m_inv_dx_table[cell_id] * (right_minus_xj_3 - left_minus_xj_3)) +
                                     Uj_rho[1] * (gauss_point_minus_xj) + rho[cell_id];
            const double rho_u_gauss = Uj_rho_u[0] * ((gauss_point_minus_xj) * (gauss_point_minus_xj)-one_third *
                                                      m_inv_dx_table[cell_id] * (right_minus_xj_3 - left_minus_xj_3)) +
                                       Uj_rho_u[1] * (gauss_point_minus_xj) + rho_u[cell_id];
            const double rho_E_gauss = Uj_rho_E[0] * ((gauss_point_minus_xj) * (gauss_point_minus_xj)-one_third *
                                                      m_inv_dx_table[cell_id] * (right_minus_xj_3 - left_minus_xj_3)) +
                                       Uj_rho_E[1] * (gauss_point_minus_xj) + rho_E[cell_id];

            const double rho_u_2_gauss = (rho_u_gauss * rho_u_gauss) / rho_gauss;
            const double p_gauss       = (m_gamma - 1) * (rho_E_gauss - 0.5 * rho_u_2_gauss);

            vec_A[cell_id][1] +=
              quadrature.weight(i_gauss) * line_transformation.jacobianDeterminant() * (rho_u_2_gauss + p_gauss);
            vec_A[cell_id][2] += quadrature.weight(i_gauss) * line_transformation.jacobianDeterminant() *
                                 (rho_u_gauss / rho_gauss * (rho_E_gauss + p_gauss));
          }

          vec_A[cell_id][0] = rho_u[cell_id];
          vec_A[cell_id][1] /= m_dx_table[cell_id];
          vec_A[cell_id][2] /= m_dx_table[cell_id];

        } else {
          p[cell_id]        = (m_gamma - 1) * (rho_E[cell_id] - 0.5 * rho_u[cell_id] * rho_u[cell_id] / rho[cell_id]);
          vec_A[cell_id][0] = rho_u[cell_id];
          vec_A[cell_id][1] = rho_u[cell_id] * rho_u[cell_id] / rho[cell_id] + p[cell_id];
          vec_A[cell_id][2] = (rho_E[cell_id] + p[cell_id]) * rho_u[cell_id] / rho[cell_id];
        }
      });
    return vec_A;
  }

  CellArray<double>
  compute_M(const CellValue<const double>& u, const CellValue<const double> Au)
  {
    const size_t nb_waves = m_lambda_vector.size();
    CellArray<double> M{m_mesh.connectivity(), nb_waves};

    double inv_L2norm_lambda_vector = 0;
    for (size_t i = 0; i < nb_waves; ++i) {
      inv_L2norm_lambda_vector += std::pow(m_lambda_vector[i], 2);
    }
    inv_L2norm_lambda_vector = 1. / inv_L2norm_lambda_vector;

    parallel_for(
      m_mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
        for (size_t i = 0; i < nb_waves; ++i) {
          M[cell_id][i] = (1. / nb_waves) * u[cell_id] + inv_L2norm_lambda_vector * Au[cell_id] * m_lambda_vector[i];
        }
      });
    return M;
  }

  NodeArray<double>
  compute_Flux1(const DiscreteFunctionP0Vector<const double>& F)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    NodeArray<double> Flux_F{m_mesh.connectivity(), nb_waves};
    Flux_F.fill(0);

    parallel_for(
      m_mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
        for (size_t i = 0; i < nb_waves; ++i) {
          const size_t N = m_mesh.numberOfCells();

          const auto& cell_to_node   = cell_to_node_matrix[cell_id];
          const NodeId left_node_id  = cell_to_node[0];
          const NodeId right_node_id = cell_to_node[1];

          const CellId prev_cell_id = (cell_id + N - 1) % N;
          const CellId next_cell_id = (cell_id + 1) % N;

          if (m_lambda_vector[i] < 0) {
            Flux_F[left_node_id][i]  = F[cell_id][i];
            Flux_F[right_node_id][i] = F[next_cell_id][i];

          } else if (m_lambda_vector[i] > 0) {
            Flux_F[left_node_id][i] = F[prev_cell_id][i];

            Flux_F[right_node_id][i] = F[cell_id][i];

          } else {
            Flux_F[right_node_id][i] = 0;
            Flux_F[left_node_id][i]  = 0;
          }
        }
      });
    return Flux_F;
  }

  NodeArray<double>
  compute_nodeFlux2(const DiscreteFunctionP0Vector<const double>& F)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    NodeArray<double> Flux_F{m_mesh.connectivity(), nb_waves};
    Flux_F.fill(0);

    parallel_for(
      m_mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
        for (size_t i = 0; i < nb_waves; ++i) {
          const NodeId first_node_id = 0;
          const NodeId last_node_id  = m_mesh.numberOfNodes() - 1;

          const double xmin = m_xr[first_node_id][0];
          const double xmax = m_xr[last_node_id][0];

          const size_t N = m_mesh.numberOfCells();

          const auto& cell_to_node   = cell_to_node_matrix[cell_id];
          const NodeId left_node_id  = cell_to_node[0];
          const NodeId right_node_id = cell_to_node[1];

          // const CellId prev2_cell_id = (cell_id + N - 2) % N;
          const CellId prev_cell_id = (cell_id + N - 1) % N;
          const CellId next_cell_id = (cell_id + 1) % N;
          // const CellId next2_cell_id = (cell_id + 2) % N;

          // const double xjm2 = (-1 * (cell_id < 2) * (xmax - xmin)) + m_xj[prev2_cell_id][0];
          const double xjm1 = (-1 * (cell_id < 1) * (xmax - xmin)) + m_xj[prev_cell_id][0];
          const double xj   = m_xj[cell_id][0];
          const double xjp1 = ((cell_id + 1 >= N) * (xmax - xmin)) + m_xj[next_cell_id][0];
          // const double xjp2 = ((cell_id + 2 >= N) * (xmax - xmin)) + m_xj[next2_cell_id][0];

          if (m_lambda_vector[i] < 0) {
            const double left_node = m_xr[left_node_id][0];

            Flux_F[left_node_id][i] =
              (left_node - xjp1) / (xj - xjp1) * F[cell_id][i] + (left_node - xj) / (xjp1 - xj) * F[next_cell_id][i];
            if (cell_id == 0) {
              Flux_F[last_node_id][i] = Flux_F[left_node_id][i];
            }

          } else if (m_lambda_vector[i] > 0) {
            const double right_node = m_xr[right_node_id][0];

            Flux_F[right_node_id][i] =
              (right_node - xj) / (xjm1 - xj) * F[prev_cell_id][i] + (right_node - xjm1) / (xj - xjm1) * F[cell_id][i];

            if (cell_id == m_mesh.numberOfCells() - 1) {
              Flux_F[first_node_id][i] = Flux_F[right_node_id][i];
            }

          } else {
            Flux_F[right_node_id][i] = 0;
            Flux_F[left_node_id][i]  = 0;
          }
        }
      });

    return Flux_F;
  }

  NodeArray<double>
  compute_nodeFlux2centered(const DiscreteFunctionP0Vector<const double>& F)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    NodeArray<double> Flux_F{m_mesh.connectivity(), nb_waves};
    Flux_F.fill(0);

    parallel_for(
      m_mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
        const NodeId first_node_id = 0;
        const NodeId last_node_id  = m_mesh.numberOfNodes() - 1;

        const double xmin = m_xr[first_node_id][0];
        const double xmax = m_xr[last_node_id][0];

        const size_t N = m_mesh.numberOfCells();

        const auto& cell_to_node   = cell_to_node_matrix[cell_id];
        const NodeId left_node_id  = cell_to_node[0];
        const NodeId right_node_id = cell_to_node[1];

        for (size_t i = 0; i < nb_waves; ++i) {
          const CellId prev_cell_id = (cell_id + N - 1) % N;
          const CellId next_cell_id = (cell_id + 1) % N;

          const double xjm1 = (-1 * (cell_id < 1) * (xmax - xmin)) + m_xj[prev_cell_id][0];
          const double xj   = m_xj[cell_id][0];
          const double xjp1 = ((cell_id + 1 >= N) * (xmax - xmin)) + m_xj[next_cell_id][0];

          const double left_node  = m_xr[left_node_id][0];
          const double right_node = m_xr[right_node_id][0];
          if (m_lambda_vector[i] < 0) {
            Flux_F[left_node_id][i] =
              (left_node - xj) / (xjm1 - xj) * F[prev_cell_id][i] + (left_node - xjm1) / (xj - xjm1) * F[cell_id][i];
            if (cell_id == 0) {
              Flux_F[last_node_id][i] = Flux_F[left_node_id][i];
            }
          } else if (m_lambda_vector[i] > 0) {
            Flux_F[right_node_id][i] =
              (right_node - xjp1) / (xj - xjp1) * F[cell_id][i] + (right_node - xj) / (xjp1 - xj) * F[next_cell_id][i];
            if (cell_id == m_mesh.numberOfCells() - 1) {
              Flux_F[first_node_id][i] = Flux_F[right_node_id][i];
            }
          } else {
            Flux_F[right_node_id][i] = 0;
            Flux_F[left_node_id][i]  = 0;
          }
        }
      });

    return Flux_F;
  }

  NodeArray<double>
  compute_nodeFlux3(const DiscreteFunctionP0Vector<const double>& F)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    NodeArray<double> Flux_F{m_mesh.connectivity(), nb_waves};
    Flux_F.fill(0);

    parallel_for(
      m_mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
        for (size_t i = 0; i < nb_waves; ++i) {
          const NodeId first_node_id = 0;
          const NodeId last_node_id  = m_mesh.numberOfNodes() - 1;

          const double xmin = m_xr[first_node_id][0];
          const double xmax = m_xr[last_node_id][0];

          const size_t N = m_mesh.numberOfCells();

          const auto& cell_to_node   = cell_to_node_matrix[cell_id];
          const NodeId left_node_id  = cell_to_node[0];
          const NodeId right_node_id = cell_to_node[1];

          const CellId prev_cell_id = (cell_id + N - 1) % N;
          const CellId next_cell_id = (cell_id + 1) % N;

          const double xjm1 = (-1 * (cell_id < 1) * (xmax - xmin)) + m_xj[prev_cell_id][0];
          const double xj   = m_xj[cell_id][0];
          const double xjp1 = ((cell_id + 1 >= N) * (xmax - xmin)) + m_xj[next_cell_id][0];

          const double left_node  = m_xr[left_node_id][0];
          const double right_node = m_xr[right_node_id][0];

          if (m_lambda_vector[i] < 0) {
            Flux_F[left_node_id][i] =
              F[prev_cell_id][i] * (left_node - xj) * (left_node - xjp1) / ((xjm1 - xj) * (xjm1 - xjp1)) +
              F[cell_id][i] * (left_node - xjm1) * (left_node - xjp1) / ((xj - xjm1) * (xj - xjp1)) +
              F[next_cell_id][i] * (left_node - xjm1) * (left_node - xj) / ((xjp1 - xjm1) * (xjp1 - xj));

            if (cell_id == 0) {
              Flux_F[last_node_id][i] = Flux_F[left_node_id][i];
            }

          } else if (m_lambda_vector[i] > 0) {
            Flux_F[right_node_id][i] =
              F[prev_cell_id][i] * (right_node - xj) * (right_node - xjp1) / ((xjm1 - xj) * (xjm1 - xjp1)) +
              F[cell_id][i] * (right_node - xjm1) * (right_node - xjp1) / ((xj - xjm1) * (xj - xjp1)) +
              F[next_cell_id][i] * (right_node - xjm1) * (right_node - xj) / ((xjp1 - xjm1) * (xjp1 - xj));

            if (cell_id == m_mesh.numberOfCells() - 1) {
              Flux_F[first_node_id][i] = Flux_F[right_node_id][i];
            }

          } else {
            Flux_F[right_node_id][i] = 0;
            Flux_F[left_node_id][i]  = 0;
          }
        }
      });

    return Flux_F;
  }

  NodeArray<double>
  compute_Polynomial_reconstruction_2(const DiscreteFunctionP0Vector<const double>& F)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    NodeArray<double> Flux_F{m_mesh.connectivity(), nb_waves};
    Flux_F.fill(0);

    parallel_for(
      m_mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
        const NodeId first_node_id = 0;
        const NodeId last_node_id  = m_mesh.numberOfNodes() - 1;

        const double xmin = m_xr[first_node_id][0];
        const double xmax = m_xr[last_node_id][0];

        const size_t N = m_mesh.numberOfCells();

        const auto& cell_to_node   = cell_to_node_matrix[cell_id];
        const NodeId left_node_id  = cell_to_node[0];
        const NodeId right_node_id = cell_to_node[1];

        const CellId prev_cell_id = (cell_id + N - 1) % N;
        const CellId next_cell_id = (cell_id + 1) % N;

        const double xj = m_xj[cell_id][0];

        const auto& prev_cell_to_node = cell_to_node_matrix[prev_cell_id];
        const NodeId left2_node_id    = prev_cell_to_node[0];

        const auto& next_cell_to_node = cell_to_node_matrix[next_cell_id];
        const NodeId right2_node_id   = next_cell_to_node[1];

        const double left_node  = m_xr[left_node_id][0];
        const double right_node = m_xr[right_node_id][0];

        const double left2_node  = (-1 * (cell_id < 1) * (xmax - xmin)) + m_xr[left2_node_id][0];
        const double right2_node = ((cell_id + 1 >= N) * (xmax - xmin)) + m_xr[right2_node_id][0];

        const double left_minus_xj  = left_node - xj;
        const double right_minus_xj = right_node - xj;

        const double left2_node_xj  = left2_node - xj;
        const double right2_node_xj = right2_node - xj;

        constexpr double one_half = 1. / 2;

        for (size_t i = 0; i < nb_waves; ++i) {
          if (m_lambda_vector[i] < 0) {
            const TinyMatrix<1> Qj(one_half * m_inv_dx_table[next_cell_id] *
                                   ((right2_node_xj) * (right2_node_xj) - (right_minus_xj) * (right_minus_xj)));

            TinyMatrix inv_Qj = inverse(Qj);

            const TinyVector<1> Bj(F[next_cell_id][i] - F[cell_id][i]);

            const TinyVector<1> Uj = inv_Qj * Bj;

            Flux_F[left_node_id][i] = Uj[0] * left_minus_xj + F[cell_id][i];

            if (cell_id == 0) {
              Flux_F[last_node_id][i] = Flux_F[left_node_id][i];
            }

          } else if (m_lambda_vector[i] > 0) {
            const TinyMatrix<1> Qj(one_half * m_inv_dx_table[prev_cell_id] *
                                   ((left_minus_xj) * (left_minus_xj) - (left2_node_xj) * (left2_node_xj)));

            TinyMatrix inv_Qj = inverse(Qj);

            const TinyVector<1> Bj(F[prev_cell_id][i] - F[cell_id][i]);

            const TinyVector<1> Uj = inv_Qj * Bj;

            Flux_F[right_node_id][i] = Uj[0] * right_minus_xj + F[cell_id][i];

            if (cell_id == m_mesh.numberOfCells() - 1) {
              Flux_F[first_node_id][i] = Flux_F[right_node_id][i];
            }
          } else {
            Flux_F[right_node_id][i] = 0;
            Flux_F[left_node_id][i]  = 0;
          }
        }
      });

    return Flux_F;
  }

  NodeArray<double>
  compute_Polynomial_reconstruction_3(const DiscreteFunctionP0Vector<const double>& F)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    NodeArray<double> Flux_F{m_mesh.connectivity(), nb_waves};
    Flux_F.fill(0);

    parallel_for(
      m_mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
        const NodeId first_node_id = 0;
        const NodeId last_node_id  = m_mesh.numberOfNodes() - 1;

        const double xmin = m_xr[first_node_id][0];
        const double xmax = m_xr[last_node_id][0];

        const size_t N = m_mesh.numberOfCells();

        const auto& cell_to_node   = cell_to_node_matrix[cell_id];
        const NodeId left_node_id  = cell_to_node[0];
        const NodeId right_node_id = cell_to_node[1];

        const CellId prev_cell_id = (cell_id + N - 1) % N;
        const CellId next_cell_id = (cell_id + 1) % N;

        const double xj = m_xj[cell_id][0];

        const auto& prev_cell_to_node = cell_to_node_matrix[prev_cell_id];
        const NodeId left2_node_id    = prev_cell_to_node[0];

        const auto& next_cell_to_node = cell_to_node_matrix[next_cell_id];
        const NodeId right2_node_id   = next_cell_to_node[1];

        const double left_node  = m_xr[left_node_id][0];
        const double right_node = m_xr[right_node_id][0];

        const double left2_node  = (-1 * (cell_id < 1) * (xmax - xmin)) + m_xr[left2_node_id][0];
        const double right2_node = ((cell_id + 1 >= N) * (xmax - xmin)) + m_xr[right2_node_id][0];

        const double left_minus_xj  = left_node - xj;
        const double right_minus_xj = right_node - xj;

        const double left_minus_xj_3  = left_minus_xj * left_minus_xj * left_minus_xj;
        const double right_minus_xj_3 = right_minus_xj * right_minus_xj * right_minus_xj;

        const double left2_node_xj  = left2_node - xj;
        const double right2_node_xj = right2_node - xj;

        const double left2_minus_xj_3  = left2_node_xj * left2_node_xj * left2_node_xj;
        const double right2_minus_xj_3 = right2_node_xj * right2_node_xj * right2_node_xj;

        constexpr double one_third = 1. / 3;
        constexpr double one_half  = 1. / 2;

        const TinyMatrix<2> Qj(one_third * m_inv_dx_table[prev_cell_id] * (left_minus_xj_3 - left2_minus_xj_3) -
                                 one_third * m_inv_dx_table[cell_id] * (right_minus_xj_3 - left_minus_xj_3),
                               one_half * m_inv_dx_table[prev_cell_id] *
                                 ((left_minus_xj) * (left_minus_xj) - (left2_node - xj) * (left2_node - xj)),
                               one_third * m_inv_dx_table[next_cell_id] * (right2_minus_xj_3 - right_minus_xj_3) -
                                 one_third * m_inv_dx_table[cell_id] * (right_minus_xj_3 - left_minus_xj_3),
                               one_half * m_inv_dx_table[next_cell_id] *
                                 ((right2_node - xj) * (right2_node - xj) - (right_minus_xj) * (right_minus_xj)));

        TinyMatrix inv_Qj = inverse(Qj);

        for (size_t i = 0; i < nb_waves; ++i) {
          const TinyVector<2> Bj(F[prev_cell_id][i] - F[cell_id][i], F[next_cell_id][i] - F[cell_id][i]);

          const TinyVector<2> Uj = inv_Qj * Bj;

          if (m_lambda_vector[i] < 0) {
            Flux_F[left_node_id][i] = Uj[0] * ((left_minus_xj) * (left_minus_xj)-one_third * m_inv_dx_table[cell_id] *
                                               (right_minus_xj_3 - left_minus_xj_3)) +
                                      Uj[1] * (left_minus_xj) + F[cell_id][i];

            if (cell_id == 0) {
              Flux_F[last_node_id][i] = Flux_F[left_node_id][i];
            }

          } else if (m_lambda_vector[i] > 0) {
            Flux_F[right_node_id][i] = Uj[0] * ((right_minus_xj) * (right_minus_xj)-one_third *
                                                m_inv_dx_table[cell_id] * (right_minus_xj_3 - left_minus_xj_3)) +
                                       Uj[1] * (right_minus_xj) + F[cell_id][i];

            if (cell_id == m_mesh.numberOfCells() - 1) {
              Flux_F[first_node_id][i] = Flux_F[right_node_id][i];
            }
          } else {
            Flux_F[right_node_id][i] = 0;
            Flux_F[left_node_id][i]  = 0;
          }
        }
      });

    return Flux_F;
  }

  NodeArray<double>
  TVD(NodeArray<double>& Flux_F, const DiscreteFunctionP0Vector<const double>& F, const double c)
  {
    const size_t nb_waves = m_lambda_vector.size();

    parallel_for(
      m_mesh.numberOfNodes(), PUGS_LAMBDA(const NodeId node_id) {
        const size_t N = m_mesh.numberOfCells();

        for (size_t i_wave = 0; i_wave < nb_waves; ++i_wave) {
          if (m_lambda_vector[i_wave] > 0) {
            const size_t j_p_half = node_id;

            const CellId cell_id_jm1 = (j_p_half + N - 2) % N;
            const CellId cell_id_j   = (j_p_half + N - 1) % N;
            const CellId cell_id_jp1 = (j_p_half + N) % N;

            double Fmin = std::min(F[cell_id_j][i_wave], F[cell_id_jp1][i_wave]);
            double Fmax = std::max(F[cell_id_j][i_wave], F[cell_id_jp1][i_wave]);

            double nu   = m_lambda_vector[i_wave] * c * m_dt / (m_dx_table[cell_id_j]);
            double F_ul = F[cell_id_j][i_wave] + (1 - nu) / nu * (F[cell_id_j][i_wave] - F[cell_id_jm1][i_wave]);

            double Fmin_ul = std::min(F[cell_id_j][i_wave], F_ul);
            double Fmax_ul = std::max(F[cell_id_j][i_wave], F_ul);

            Fmin = std::max(Fmin, Fmin_ul);
            Fmax = std::min(Fmax, Fmax_ul);

            if (Flux_F[node_id][i_wave] < Fmin) {
              Flux_F[node_id][i_wave] = Fmin;
            } else if (Flux_F[node_id][i_wave] > Fmax) {
              Flux_F[node_id][i_wave] = Fmax;
            }
          } else if (m_lambda_vector[i_wave] < 0) {
            const size_t j_p_half = node_id;

            const CellId cell_id_j   = (j_p_half + N - 1) % N;
            const CellId cell_id_jp1 = (j_p_half + N) % N;
            const CellId cell_id_jp2 = (j_p_half + N + 1) % N;

            double Fmin = std::min(F[cell_id_j][i_wave], F[cell_id_jp1][i_wave]);
            double Fmax = std::max(F[cell_id_j][i_wave], F[cell_id_jp1][i_wave]);

            double nu   = m_lambda_vector[i_wave] * c * m_dt / (m_dx_table[cell_id_j]);
            double F_ul = F[cell_id_jp1][i_wave] - (1 + nu) / nu * (F[cell_id_jp1][i_wave] - F[cell_id_jp2][i_wave]);

            double Fmin_ul = std::min(F[cell_id_jp1][i_wave], F_ul);
            double Fmax_ul = std::max(F[cell_id_jp1][i_wave], F_ul);

            Fmin = std::max(Fmin, Fmin_ul);
            Fmax = std::min(Fmax, Fmax_ul);

            if (Flux_F[node_id][i_wave] < Fmin) {
              Flux_F[node_id][i_wave] = Fmin;
            } else if (Flux_F[node_id][i_wave] > Fmax) {
              Flux_F[node_id][i_wave] = Fmax;
            }
          }
        }
      });
    return Flux_F;
  }

  template <size_t nb_subtimesteps>
  std::vector<CellValue<double>>
  compute_PFnp1(const DiscreteFunctionP0Vector<const double>& F0,
                std::vector<DiscreteFunctionP0Vector<double>>& deltaF,
                DiscreteFunctionP0Vector<double>& deltaF0,
                const TinyMatrix<nb_subtimesteps>& B,
                const TinyVector<nb_subtimesteps>& b0)

  {
    std::vector<CellValue<double>> PFnp1;
    for (size_t i_subtimestep = 0; i_subtimestep < nb_subtimesteps; ++i_subtimestep) {
      PFnp1.emplace_back(m_mesh.connectivity());
    }

    for (size_t i_subtimestep = 0; i_subtimestep < nb_subtimesteps; ++i_subtimestep) {
      parallel_for(
        m_mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
          const double inv_dx = m_inv_dx_table[cell_id];
          const double c      = m_dt * inv_dx;

          double test_rho_j = 0;

          PFnp1[i_subtimestep][cell_id] = 0;
          for (size_t i_wave = 0; i_wave < m_lambda_vector.size(); ++i_wave) {
            const double lambda = m_lambda_vector[i_wave];
            TinyVector<nb_subtimesteps> deltaF_vec;
            TinyVector<nb_subtimesteps> b0_deltaF0;

            for (size_t j_subtimestep = 0; j_subtimestep < nb_subtimesteps; ++j_subtimestep) {
              deltaF_vec[j_subtimestep] = deltaF[j_subtimestep][cell_id][i_wave];
              b0_deltaF0[j_subtimestep] = b0[j_subtimestep] * deltaF0[cell_id][i_wave];
            }
            const TinyVector<nb_subtimesteps> B_deltaF = B * deltaF_vec;

            PFnp1[i_subtimestep][cell_id] +=
              F0[cell_id][i_wave] - c * lambda * (B_deltaF[i_subtimestep] + b0_deltaF0[i_subtimestep]);

            test_rho_j += -c * lambda * B_deltaF[i_subtimestep];
          }
        });
    }
    return PFnp1;
  }

  CellValue<double>
  compute_PFn(const DiscreteFunctionP0Vector<const double> F)
  {
    CellValue<double> PFn{m_mesh.connectivity()};

    // parallel_for(
    //   m_mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      double PF = 0;
      for (size_t i = 0; i < m_lambda_vector.size(); ++i) {
        PF += F[cell_id][i];
      }
      PFn[cell_id] = PF;
    }   //    });
    return PFn;
  }

  template <size_t nb_subtimesteps>
  std::vector<DiscreteFunctionP0Vector<double>>
  compute_Fnp1(std::vector<CellArray<double>>& MPFnp1,
               const CellArray<const double>& MPFn,
               std::vector<DiscreteFunctionP0Vector<double>>& deltaFnp1,
               DiscreteFunctionP0Vector<double>& deltaFn,
               const DiscreteFunctionP0Vector<const double> Fn,
               const TinyMatrix<nb_subtimesteps>& B,
               const TinyVector<nb_subtimesteps>& b0)
  {
    const size_t nb_waves = m_lambda_vector.size();

    std::vector<DiscreteFunctionP0Vector<double>> Fnp1;

    for (size_t i_subtimestep = 0; i_subtimestep < nb_subtimesteps; ++i_subtimestep) {
      Fnp1.emplace_back(p_mesh, nb_waves);
    }

    const double c1 = 1. / (m_eps * m_eps + m_eps * m_dt * 0.5 + (1. / 12.) * m_dt * m_dt);
    const TinyMatrix<nb_subtimesteps> I_muB_inv(c1 * (m_eps + m_dt / 6.), c1 * m_dt / 24., c1 * (-m_dt * 2. / 3.),
                                                c1 * (m_eps + m_dt / 3.));

    for (size_t i_subtimestep = 0; i_subtimestep < nb_subtimesteps; ++i_subtimestep) {
      parallel_for(
        m_mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
          const double dx = m_dx_table[cell_id];
          const double c2 = (m_eps * m_dt / dx);

          for (size_t i_wave = 0; i_wave < nb_waves; ++i_wave) {
            TinyVector<nb_subtimesteps> MPFnp1_vec;
            TinyVector<nb_subtimesteps> deltaFnp1_vec;
            TinyVector<nb_subtimesteps> b0_deltaFn;
            TinyVector<nb_subtimesteps> b0_MPFn_Fn;

            for (size_t j_subtimestep = 0; j_subtimestep < nb_subtimesteps; ++j_subtimestep) {
              MPFnp1_vec[j_subtimestep]    = MPFnp1[j_subtimestep][cell_id][i_wave];
              deltaFnp1_vec[j_subtimestep] = deltaFnp1[j_subtimestep][cell_id][i_wave];
              b0_deltaFn[j_subtimestep]    = b0[j_subtimestep] * deltaFn[cell_id][i_wave];
              b0_MPFn_Fn[j_subtimestep]    = b0[j_subtimestep] * (MPFn[cell_id][i_wave] - Fn[cell_id][i_wave]);
            }
            const TinyVector<nb_subtimesteps> B_MPFnp1    = B * MPFnp1_vec;
            const TinyVector<nb_subtimesteps> B_deltaFnp1 = B * deltaFnp1_vec;
            Fnp1[i_subtimestep][cell_id][i_wave]          = 0;

            for (size_t j_subtimestep = 0; j_subtimestep < nb_subtimesteps; ++j_subtimestep) {
              Fnp1[i_subtimestep][cell_id][i_wave] +=
                I_muB_inv(i_subtimestep, j_subtimestep) *
                (m_dt * B_MPFnp1[j_subtimestep] + m_eps * Fn[cell_id][i_wave] -
                 c2 * m_lambda_vector[i_wave] * (B_deltaFnp1[j_subtimestep] + b0_deltaFn[j_subtimestep]) +
                 m_dt * b0_MPFn_Fn[j_subtimestep]);
            }
          }
        });
    }
    return Fnp1;
  }

  std::tuple<NodeValue<bool>, CellValue<bool>, bool>
  test_validity_mood(DiscreteFunctionP0Vector<double> Fnp1_rho,
                     DiscreteFunctionP0Vector<double> Fnp1_rho_u,
                     DiscreteFunctionP0Vector<double> Fnp1_rho_E,
                     NodeValue<bool> node_is_valid,
                     CellValue<bool> cell_is_valid)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    bool Mood_activation            = false;

    const CellValue<double> rho   = compute_PFn(Fnp1_rho);
    const CellValue<double> rho_u = compute_PFn(Fnp1_rho_u);
    const CellValue<double> rho_E = compute_PFn(Fnp1_rho_E);
    const CellValue<double> p{m_mesh.connectivity()};

    const size_t N = m_mesh.numberOfCells();

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      if (rho[cell_id] <= 0) {
        Mood_activation           = true;
        cell_is_valid[cell_id]    = false;
        const CellId prev_cell_id = (cell_id + N - 1) % N;
        const CellId next_cell_id = (cell_id + 1) % N;

        const auto& cell_to_node      = cell_to_node_matrix[cell_id];
        const auto& prev_cell_to_node = cell_to_node_matrix[prev_cell_id];
        const auto& next_cell_to_node = cell_to_node_matrix[next_cell_id];

        for (size_t i_node = 0; i_node < cell_to_node.size(); ++i_node) {
          const NodeId node_id        = cell_to_node[i_node];
          const NodeId next_node_id   = next_cell_to_node[i_node];
          const NodeId prev_node_id   = prev_cell_to_node[i_node];
          node_is_valid[node_id]      = false;
          node_is_valid[next_node_id] = false;
          node_is_valid[prev_node_id] = false;
          // std::cout << "Mood activation in cell " << cell_id << "\n";
        }
      } else {
        p[cell_id] = (m_gamma - 1) * (rho_E[cell_id] - 0.5 * rho_u[cell_id] * rho_u[cell_id] / rho[cell_id]);
        if (p[cell_id] <= 0) {
          Mood_activation           = true;
          cell_is_valid[cell_id]    = false;
          const CellId prev_cell_id = (cell_id + N - 1) % N;
          const CellId next_cell_id = (cell_id + 1) % N;

          const auto& cell_to_node      = cell_to_node_matrix[cell_id];
          const auto& prev_cell_to_node = cell_to_node_matrix[prev_cell_id];
          const auto& next_cell_to_node = cell_to_node_matrix[next_cell_id];

          for (size_t i_node = 0; i_node < cell_to_node.size(); ++i_node) {
            const NodeId node_id      = cell_to_node[i_node];
            const NodeId next_node_id = next_cell_to_node[i_node];
            const NodeId prev_node_id = prev_cell_to_node[i_node];
            node_is_valid[node_id]    = false;

            node_is_valid[next_node_id] = false;
            node_is_valid[prev_node_id] = false;
          }
        }
      }
    }

    return std::make_tuple(NodeValue<bool>(node_is_valid), CellValue<bool>(cell_is_valid), bool(Mood_activation));
  }

  std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
             std::shared_ptr<const DiscreteFunctionVariant>,
             std::shared_ptr<const DiscreteFunctionVariant>>
  apply()
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    // const auto& node_to_cell_matrix = m_mesh.connectivity().nodeToCellMatrix();
    const size_t nb_waves        = m_lambda_vector.size();
    const size_t nb_subtimesteps = 2;
    const size_t nb_equations    = 3;

    const TinyVector<nb_subtimesteps> c(0.5, 1);
    CellValue<bool> cell_is_valid{m_mesh.connectivity()};
    cell_is_valid.fill(true);

    // if (m_MoodLimitation) {
    NodeValue<bool> node_is_valid{m_mesh.connectivity()};
    node_is_valid.fill(true);
    bool Mood_activation = true;

    DiscreteFunctionP0Vector<double> F0_rho   = copy(m_Fn_rho);
    DiscreteFunctionP0Vector<double> F0_rho_u = copy(m_Fn_rho_u);
    DiscreteFunctionP0Vector<double> F0_rho_E = copy(m_Fn_rho_E);

    std::vector<DiscreteFunctionP0Vector<double>> F_rho;
    std::vector<DiscreteFunctionP0Vector<double>> F_rho_u;
    std::vector<DiscreteFunctionP0Vector<double>> F_rho_E;

    std::vector<DiscreteFunctionP0Vector<double>> Fnp1_rho;
    std::vector<DiscreteFunctionP0Vector<double>> Fnp1_rho_u;
    std::vector<DiscreteFunctionP0Vector<double>> Fnp1_rho_E;

    // Compute first order F

    const CellValue<const double> rho   = compute_PFn(F0_rho);
    const CellValue<const double> rho_u = compute_PFn(F0_rho_u);
    const CellValue<const double> rho_E = compute_PFn(F0_rho_E);

    cell_is_valid.fill(false);
    const CellArray<const double> APFn_O1 = getA(rho, rho_u, rho_E, cell_is_valid);
    cell_is_valid.fill(true);
    const CellValue<double> APFn_rho_O1{m_mesh.connectivity()};
    const CellValue<double> APFn_rho_u_O1{m_mesh.connectivity()};
    const CellValue<double> APFn_rho_E_O1{m_mesh.connectivity()};

    parallel_for(
      m_mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
        APFn_rho_O1[cell_id]   = APFn_O1[cell_id][0];
        APFn_rho_u_O1[cell_id] = APFn_O1[cell_id][1];
        APFn_rho_E_O1[cell_id] = APFn_O1[cell_id][2];
      });

    const CellArray<const double> Fn_rho_O1   = compute_M(rho, APFn_rho_O1);
    const CellArray<const double> Fn_rho_u_O1 = compute_M(rho_u, APFn_rho_u_O1);
    const CellArray<const double> Fn_rho_E_O1 = compute_M(rho_E, APFn_rho_E_O1);

    size_t count_mood = 0;
    while (Mood_activation) {
      Mood_activation = false;

      // if (count_mood > 0) {
      //   std::cout << "MOOD activation number " << count_mood << "\n";
      // }

      count_mood += 1;

      for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
        for (size_t i_wave = 0; i_wave < nb_waves; ++i_wave) {
          if (not cell_is_valid[cell_id]) {
            F0_rho[cell_id][i_wave]   = Fn_rho_O1[cell_id][i_wave];
            F0_rho_u[cell_id][i_wave] = Fn_rho_u_O1[cell_id][i_wave];
            F0_rho_E[cell_id][i_wave] = Fn_rho_E_O1[cell_id][i_wave];
          }
        }
      }

      for (size_t i_subtimestep = 0; i_subtimestep < nb_subtimesteps; ++i_subtimestep) {
        F_rho.emplace_back(p_mesh, nb_waves);
        F_rho_u.emplace_back(p_mesh, nb_waves);
        F_rho_E.emplace_back(p_mesh, nb_waves);

        Fnp1_rho.emplace_back(p_mesh, nb_waves);
        Fnp1_rho_u.emplace_back(p_mesh, nb_waves);
        Fnp1_rho_E.emplace_back(p_mesh, nb_waves);

        F_rho[i_subtimestep]   = copy(F0_rho);
        F_rho_u[i_subtimestep] = copy(F0_rho_u);
        F_rho_E[i_subtimestep] = copy(F0_rho_E);

        Fnp1_rho[i_subtimestep]   = copy(F0_rho);
        Fnp1_rho_u[i_subtimestep] = copy(F0_rho_u);
        Fnp1_rho_E[i_subtimestep] = copy(F0_rho_E);
      }

      DiscreteFunctionP0Vector<double> deltaFn_rho(p_mesh, nb_waves);
      DiscreteFunctionP0Vector<double> deltaFn_rho_u(p_mesh, nb_waves);
      DiscreteFunctionP0Vector<double> deltaFn_rho_E(p_mesh, nb_waves);

      NodeArray<double> Flux_Fn_rho{m_mesh.connectivity(), nb_waves};
      NodeArray<double> Flux_Fn_rho_u{m_mesh.connectivity(), nb_waves};
      NodeArray<double> Flux_Fn_rho_E{m_mesh.connectivity(), nb_waves};

      NodeArray<double> Flux1_Fn_rho{m_mesh.connectivity(), nb_waves};
      NodeArray<double> Flux1_Fn_rho_u{m_mesh.connectivity(), nb_waves};
      NodeArray<double> Flux1_Fn_rho_E{m_mesh.connectivity(), nb_waves};

      if (m_SpaceOrder == 3) {
        Flux_Fn_rho   = compute_Polynomial_reconstruction_3(F0_rho);
        Flux_Fn_rho_u = compute_Polynomial_reconstruction_3(F0_rho_u);
        Flux_Fn_rho_E = compute_Polynomial_reconstruction_3(F0_rho_E);
      } else if (m_SpaceOrder == 2) {
        Flux_Fn_rho   = compute_Polynomial_reconstruction_2(F0_rho);
        Flux_Fn_rho_u = compute_Polynomial_reconstruction_2(F0_rho_u);
        Flux_Fn_rho_E = compute_Polynomial_reconstruction_2(F0_rho_E);
      } else if (m_SpaceOrder == 1) {
        Flux_Fn_rho   = compute_Flux1(F0_rho);
        Flux_Fn_rho_u = compute_Flux1(F0_rho_u);
        Flux_Fn_rho_E = compute_Flux1(F0_rho_E);
      }

      if (m_Limitation == 1) {
        Flux_Fn_rho   = TVD(Flux_Fn_rho, F0_rho, c[0]);
        Flux_Fn_rho_u = TVD(Flux_Fn_rho_u, F0_rho_u, c[0]);
        Flux_Fn_rho_E = TVD(Flux_Fn_rho_E, F0_rho_E, c[0]);
      }

      Flux1_Fn_rho   = compute_Flux1(F0_rho);
      Flux1_Fn_rho_u = compute_Flux1(F0_rho_u);
      Flux1_Fn_rho_E = compute_Flux1(F0_rho_E);

      const CellValue<const double> PFn_rho   = compute_PFn(F0_rho);
      const CellValue<const double> PFn_rho_u = compute_PFn(F0_rho_u);
      const CellValue<const double> PFn_rho_E = compute_PFn(F0_rho_E);

      // for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      //   std::cout << PFn_rho[cell_id] << "  "
      //             << (m_gamma - 1) *
      //                  (PFn_rho_E[cell_id] - 0.5 * PFn_rho_u[cell_id] * PFn_rho_u[cell_id] / PFn_rho[cell_id])
      //             << "\n";
      // }

      const CellArray<const double> APFn = getA(PFn_rho, PFn_rho_u, PFn_rho_E, cell_is_valid);
      const CellValue<double> APFn_rho{m_mesh.connectivity()};
      const CellValue<double> APFn_rho_u{m_mesh.connectivity()};
      const CellValue<double> APFn_rho_E{m_mesh.connectivity()};

      parallel_for(
        m_mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
          APFn_rho[cell_id]   = APFn[cell_id][0];
          APFn_rho_u[cell_id] = APFn[cell_id][1];
          APFn_rho_E[cell_id] = APFn[cell_id][2];
        });

      const CellArray<const double> MPFn_rho   = compute_M(PFn_rho, APFn_rho);
      const CellArray<const double> MPFn_rho_u = compute_M(PFn_rho_u, APFn_rho_u);
      const CellArray<const double> MPFn_rho_E = compute_M(PFn_rho_E, APFn_rho_E);

      for (size_t p = 0; p < 3; ++p) {
        // std::cout << "DeC iteration n° " << p << "\n";
        const TinyMatrix<nb_subtimesteps> B(1. / 3., -1. / 24., 2. / 3., 1. / 6.);
        const TinyVector<nb_subtimesteps> b0(5. / 24., 1. / 6.);

        std::vector<DiscreteFunctionP0Vector<double>> deltaFnp1_rho;
        std::vector<DiscreteFunctionP0Vector<double>> deltaFnp1_rho_u;
        std::vector<DiscreteFunctionP0Vector<double>> deltaFnp1_rho_E;

        std::vector<CellValue<double>> PFnp1_rho;
        std::vector<CellValue<double>> PFnp1_rho_u;
        std::vector<CellValue<double>> PFnp1_rho_E;

        std::vector<NodeArray<double>> Flux_Fnp1_rho;
        std::vector<NodeArray<double>> Flux_Fnp1_rho_u;
        std::vector<NodeArray<double>> Flux_Fnp1_rho_E;

        std::vector<NodeArray<double>> Flux1_Fnp1_rho;
        std::vector<NodeArray<double>> Flux1_Fnp1_rho_u;
        std::vector<NodeArray<double>> Flux1_Fnp1_rho_E;

        for (size_t i_subtimestep = 0; i_subtimestep < nb_subtimesteps; ++i_subtimestep) {
          deltaFnp1_rho.emplace_back(p_mesh, nb_waves);
          deltaFnp1_rho_u.emplace_back(p_mesh, nb_waves);
          deltaFnp1_rho_E.emplace_back(p_mesh, nb_waves);

          Flux_Fnp1_rho.emplace_back(m_mesh.connectivity(), nb_waves);
          Flux_Fnp1_rho_u.emplace_back(m_mesh.connectivity(), nb_waves);
          Flux_Fnp1_rho_E.emplace_back(m_mesh.connectivity(), nb_waves);

          Flux1_Fnp1_rho.emplace_back(m_mesh.connectivity(), nb_waves);
          Flux1_Fnp1_rho_u.emplace_back(m_mesh.connectivity(), nb_waves);
          Flux1_Fnp1_rho_E.emplace_back(m_mesh.connectivity(), nb_waves);

          PFnp1_rho.emplace_back(m_mesh.connectivity());
          PFnp1_rho_u.emplace_back(m_mesh.connectivity());
          PFnp1_rho_E.emplace_back(m_mesh.connectivity());

          Flux1_Fnp1_rho[i_subtimestep]   = compute_Flux1(Fnp1_rho[i_subtimestep]);
          Flux1_Fnp1_rho_u[i_subtimestep] = compute_Flux1(Fnp1_rho_u[i_subtimestep]);
          Flux1_Fnp1_rho_E[i_subtimestep] = compute_Flux1(Fnp1_rho_E[i_subtimestep]);

          if (m_SpaceOrder == 3) {
            Flux_Fnp1_rho[i_subtimestep]   = compute_Polynomial_reconstruction_3(Fnp1_rho[i_subtimestep]);
            Flux_Fnp1_rho_u[i_subtimestep] = compute_Polynomial_reconstruction_3(Fnp1_rho_u[i_subtimestep]);
            Flux_Fnp1_rho_E[i_subtimestep] = compute_Polynomial_reconstruction_3(Fnp1_rho_E[i_subtimestep]);
          } else if (m_SpaceOrder == 2) {
            Flux_Fnp1_rho[i_subtimestep]   = compute_Polynomial_reconstruction_2(Fnp1_rho[i_subtimestep]);
            Flux_Fnp1_rho_u[i_subtimestep] = compute_Polynomial_reconstruction_2(Fnp1_rho_u[i_subtimestep]);
            Flux_Fnp1_rho_E[i_subtimestep] = compute_Polynomial_reconstruction_2(Fnp1_rho_E[i_subtimestep]);
          } else if (m_SpaceOrder == 1) {
            Flux_Fnp1_rho[i_subtimestep]   = Flux1_Fnp1_rho[i_subtimestep];
            Flux_Fnp1_rho_u[i_subtimestep] = Flux1_Fnp1_rho_u[i_subtimestep];
            Flux_Fnp1_rho_E[i_subtimestep] = Flux1_Fnp1_rho_E[i_subtimestep];
          }

          if (m_Limitation == 1) {
            Flux_Fnp1_rho[i_subtimestep] = TVD(Flux_Fnp1_rho[i_subtimestep], Fnp1_rho[i_subtimestep], c[i_subtimestep]);
            Flux_Fnp1_rho_u[i_subtimestep] =
              TVD(Flux_Fnp1_rho_u[i_subtimestep], Fnp1_rho_u[i_subtimestep], c[i_subtimestep]);
            Flux_Fnp1_rho_E[i_subtimestep] =
              TVD(Flux_Fnp1_rho_E[i_subtimestep], Fnp1_rho_E[i_subtimestep], c[i_subtimestep]);
          }
        }

        for (NodeId node_id = 0; node_id < m_mesh.numberOfNodes(); ++node_id) {
          if (not node_is_valid[node_id]) {
            for (size_t i_wave = 0; i_wave < nb_waves; ++i_wave) {
              Flux_Fn_rho[node_id][i_wave]   = Flux1_Fn_rho[node_id][i_wave];
              Flux_Fn_rho_u[node_id][i_wave] = Flux1_Fn_rho_u[node_id][i_wave];
              Flux_Fn_rho_E[node_id][i_wave] = Flux1_Fn_rho_E[node_id][i_wave];
              for (size_t i_subtimestep = 0; i_subtimestep < nb_subtimesteps; ++i_subtimestep) {
                Flux_Fnp1_rho[i_subtimestep][node_id][i_wave]   = Flux1_Fn_rho[node_id][i_wave];
                Flux_Fnp1_rho_u[i_subtimestep][node_id][i_wave] = Flux1_Fn_rho_u[node_id][i_wave];
                Flux_Fnp1_rho_E[i_subtimestep][node_id][i_wave] = Flux1_Fn_rho_E[node_id][i_wave];
              }
            }
          }
        }

        parallel_for(
          m_mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
            const auto& cell_to_node   = cell_to_node_matrix[cell_id];
            const NodeId left_node_id  = cell_to_node[0];
            const NodeId right_node_id = cell_to_node[1];

            for (size_t i_wave = 0; i_wave < nb_waves; ++i_wave) {
              deltaFn_rho[cell_id][i_wave] = (Flux_Fn_rho[right_node_id][i_wave] - Flux_Fn_rho[left_node_id][i_wave]);
              deltaFn_rho_u[cell_id][i_wave] =
                (Flux_Fn_rho_u[right_node_id][i_wave] - Flux_Fn_rho_u[left_node_id][i_wave]);
              deltaFn_rho_E[cell_id][i_wave] =
                (Flux_Fn_rho_E[right_node_id][i_wave] - Flux_Fn_rho_E[left_node_id][i_wave]);
            }
          });

        parallel_for(
          m_mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
            const auto& cell_to_node   = cell_to_node_matrix[cell_id];
            const NodeId left_node_id  = cell_to_node[0];
            const NodeId right_node_id = cell_to_node[1];

            for (size_t i_subtimestep = 0; i_subtimestep < nb_subtimesteps; ++i_subtimestep) {
              for (size_t i_wave = 0; i_wave < nb_waves; ++i_wave) {
                deltaFnp1_rho[i_subtimestep][cell_id][i_wave] = (Flux_Fnp1_rho[i_subtimestep][right_node_id][i_wave] -
                                                                 Flux_Fnp1_rho[i_subtimestep][left_node_id][i_wave]);
                deltaFnp1_rho_u[i_subtimestep][cell_id][i_wave] =
                  (Flux_Fnp1_rho_u[i_subtimestep][right_node_id][i_wave] -
                   Flux_Fnp1_rho_u[i_subtimestep][left_node_id][i_wave]);
                deltaFnp1_rho_E[i_subtimestep][cell_id][i_wave] =
                  (Flux_Fnp1_rho_E[i_subtimestep][right_node_id][i_wave] -
                   Flux_Fnp1_rho_E[i_subtimestep][left_node_id][i_wave]);
              }
            }
          });

        PFnp1_rho   = compute_PFnp1<nb_subtimesteps>(F0_rho, deltaFnp1_rho, deltaFn_rho, B, b0);
        PFnp1_rho_u = compute_PFnp1<nb_subtimesteps>(F0_rho_u, deltaFnp1_rho_u, deltaFn_rho_u, B, b0);
        PFnp1_rho_E = compute_PFnp1<nb_subtimesteps>(F0_rho_E, deltaFnp1_rho_E, deltaFn_rho_E, B, b0);

        std::vector<CellArray<double>> APFnp1;
        for (size_t i_subtimestep = 0; i_subtimestep < nb_subtimesteps; ++i_subtimestep) {
          APFnp1.emplace_back(m_mesh.connectivity(), nb_equations);

          APFnp1[i_subtimestep] =
            getA(PFnp1_rho[i_subtimestep], PFnp1_rho_u[i_subtimestep], PFnp1_rho_E[i_subtimestep], cell_is_valid);
        }

        std::vector<CellValue<double>> APFnp1_rho;
        std::vector<CellValue<double>> APFnp1_rho_u;
        std::vector<CellValue<double>> APFnp1_rho_E;
        for (size_t i_subtimestep = 0; i_subtimestep < nb_subtimesteps; ++i_subtimestep) {
          APFnp1_rho.emplace_back(m_mesh.connectivity());
          APFnp1_rho_u.emplace_back(m_mesh.connectivity());
          APFnp1_rho_E.emplace_back(m_mesh.connectivity());
        }

        parallel_for(
          m_mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
            for (size_t i_subtimestep = 0; i_subtimestep < nb_subtimesteps; ++i_subtimestep) {
              APFnp1_rho[i_subtimestep][cell_id]   = APFnp1[i_subtimestep][cell_id][0];
              APFnp1_rho_u[i_subtimestep][cell_id] = APFnp1[i_subtimestep][cell_id][1];
              APFnp1_rho_E[i_subtimestep][cell_id] = APFnp1[i_subtimestep][cell_id][2];
            }
          });

        std::vector<CellArray<double>> MPFnp1_rho;
        std::vector<CellArray<double>> MPFnp1_rho_u;
        std::vector<CellArray<double>> MPFnp1_rho_E;

        for (size_t i_subtimestep = 0; i_subtimestep < nb_subtimesteps; ++i_subtimestep) {
          MPFnp1_rho.emplace_back(m_mesh.connectivity(), nb_waves);
          MPFnp1_rho_u.emplace_back(m_mesh.connectivity(), nb_waves);
          MPFnp1_rho_E.emplace_back(m_mesh.connectivity(), nb_waves);

          MPFnp1_rho[i_subtimestep]   = compute_M(PFnp1_rho[i_subtimestep], APFnp1_rho[i_subtimestep]);
          MPFnp1_rho_u[i_subtimestep] = compute_M(PFnp1_rho_u[i_subtimestep], APFnp1_rho_u[i_subtimestep]);
          MPFnp1_rho_E[i_subtimestep] = compute_M(PFnp1_rho_E[i_subtimestep], APFnp1_rho_E[i_subtimestep]);
        }

        Fnp1_rho = compute_Fnp1<nb_subtimesteps>(MPFnp1_rho, MPFn_rho, deltaFnp1_rho, deltaFn_rho, F0_rho, B, b0);
        Fnp1_rho_u =
          compute_Fnp1<nb_subtimesteps>(MPFnp1_rho_u, MPFn_rho_u, deltaFnp1_rho_u, deltaFn_rho_u, F0_rho_u, B, b0);
        Fnp1_rho_E =
          compute_Fnp1<nb_subtimesteps>(MPFnp1_rho_E, MPFn_rho_E, deltaFnp1_rho_E, deltaFn_rho_E, F0_rho_E, B, b0);

        for (size_t i_subtimestep = 0; i_subtimestep < nb_subtimesteps; ++i_subtimestep) {
          F_rho[i_subtimestep]   = copy(Fnp1_rho[i_subtimestep]);
          F_rho_u[i_subtimestep] = copy(Fnp1_rho_u[i_subtimestep]);
          F_rho_E[i_subtimestep] = copy(Fnp1_rho_E[i_subtimestep]);
        }

        if (m_MoodLimitation) {
          for (size_t i_subtimestep = 0; i_subtimestep < nb_subtimesteps; ++i_subtimestep) {
            for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
              const auto& cell_to_node = cell_to_node_matrix[cell_id];
              NodeId left_node         = cell_to_node[0];
              NodeId right_node        = cell_to_node[1];

              const size_t N = m_mesh.numberOfCells();

              const CellId prev_cell_id = (cell_id + N - 1) % N;
              const CellId next_cell_id = (cell_id + 1) % N;

              const double rho_np1   = PFnp1_rho[i_subtimestep][cell_id];
              const double rho_u_np1 = PFnp1_rho_u[i_subtimestep][cell_id];
              const double rho_E_np1 = PFnp1_rho_E[i_subtimestep][cell_id];

              if (rho_np1 <= 0) {
                Mood_activation             = true;
                cell_is_valid[cell_id]      = false;
                cell_is_valid[prev_cell_id] = false;
                cell_is_valid[next_cell_id] = false;
                node_is_valid[left_node]    = false;
                node_is_valid[right_node]   = false;
                // cell_is_valid.fill(false);
                // node_is_valid.fill(false);
                // std::cout << "In cell " << cell_id << ", rho = " << rho << "\n";
              } else {
                const double pressure = (m_gamma - 1) * (rho_E_np1 - 0.5 * rho_u_np1 * rho_u_np1 / rho_np1);
                if (pressure <= 0) {
                  Mood_activation             = true;
                  cell_is_valid[cell_id]      = false;
                  cell_is_valid[prev_cell_id] = false;
                  cell_is_valid[next_cell_id] = false;
                  node_is_valid[left_node]    = false;
                  node_is_valid[right_node]   = false;
                  // cell_is_valid.fill(false);
                  // node_is_valid.fill(false);
                  // std::cout << "In cell " << cell_id << ", p = " << pressure << "\n";
                }
              }
            }
          }
        }

      }   // end DeC loop (p)

      if (count_mood >= 100) {
        std::cout << "Too much MOOD iterations. \n";
        exit(0);
      }
    }   // end MOOD while
    // }     // end MOOD

    return std::make_tuple(std::make_shared<DiscreteFunctionVariant>(Fnp1_rho[nb_subtimesteps - 1]),
                           std::make_shared<DiscreteFunctionVariant>(Fnp1_rho_u[nb_subtimesteps - 1]),
                           std::make_shared<DiscreteFunctionVariant>(Fnp1_rho_E[nb_subtimesteps - 1]));
  }

  EulerKineticSolverThirdOrderFV(std::shared_ptr<const MeshType> mesh,
                                 const double& dt,
                                 const double& eps,
                                 const double& gamma,
                                 const std::vector<double>& lambda_vector,
                                 const size_t& SpaceOrder,
                                 const size_t& Limitation,
                                 const bool& MoodLimitation,
                                 const DiscreteFunctionP0Vector<const double>& Fn_rho,
                                 const DiscreteFunctionP0Vector<const double>& Fn_rho_u,
                                 const DiscreteFunctionP0Vector<const double>& Fn_rho_E)
    : m_dt{dt},
      m_eps{eps},
      m_lambda_vector{lambda_vector},
      m_SpaceOrder{SpaceOrder},
      m_Limitation{Limitation},
      m_MoodLimitation{MoodLimitation},
      m_Fn_rho{Fn_rho},
      m_Fn_rho_u{Fn_rho_u},
      m_Fn_rho_E{Fn_rho_E},
      m_gamma{gamma},
      p_mesh{mesh},
      m_mesh{*mesh}
  {
    if ((lambda_vector.size() != Fn_rho.size()) or
        (lambda_vector.size() != Fn_rho_u.size() or (lambda_vector.size() != Fn_rho_E.size()))) {
      throw NormalError("Incompatible dimensions of lambda vector and Fn");
    }

    m_inv_dx_table = [&] {
      CellValue<double> inv_dx_table{m_mesh.connectivity()};
      parallel_for(
        m_mesh.numberOfCells(),
        PUGS_LAMBDA(const CellId cell_id) { inv_dx_table[cell_id] = 1. / m_dx_table[cell_id]; });
      return inv_dx_table;
    }();
  }

  ~EulerKineticSolverThirdOrderFV() = default;
};

std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
eulerKineticSolverThirdOrderFV(const double& dt,
                               const double& gamma,
                               const std::vector<double>& lambda_vector,
                               const double& eps,
                               const size_t& SpaceOrder,
                               const size_t& Limitation,
                               const bool& MoodLimitation,
                               const std::shared_ptr<const DiscreteFunctionVariant>& F_rho,
                               const std::shared_ptr<const DiscreteFunctionVariant>& F_rho_u,
                               const std::shared_ptr<const DiscreteFunctionVariant>& F_rho_E)
{
  const std::shared_ptr<const MeshVariant> mesh_v = getCommonMesh({F_rho, F_rho_u, F_rho_E});
  if (mesh_v.use_count() == 0) {
    throw NormalError("F_rho, F_rho_u and F_rho_E have to be defined on the same mesh");
  }
  if (SpaceOrder > 3) {
    throw NotImplementedError("Euler kinetic solver not implemented at order > 3");
  }
  if (Limitation >= 2) {
    throw NotImplementedError("This limitation reference does not exist");
  }

  return std::visit(
    [&](auto&& p_mesh)
      -> std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                    std::shared_ptr<const DiscreteFunctionVariant>> {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      if constexpr (is_polygonal_mesh_v<MeshType> and (MeshType::Dimension == 1)) {
        auto Fn_rho   = F_rho->get<DiscreteFunctionP0Vector<const double>>();
        auto Fn_rho_u = F_rho_u->get<DiscreteFunctionP0Vector<const double>>();
        auto Fn_rho_E = F_rho_E->get<DiscreteFunctionP0Vector<const double>>();

        EulerKineticSolverThirdOrderFV solver(p_mesh, dt, eps, gamma, lambda_vector, SpaceOrder, Limitation,
                                              MoodLimitation, Fn_rho, Fn_rho_u, Fn_rho_E);

        return solver.apply();
      } else {
        throw NotImplementedError("Invalid mesh type for Scalar Kinetic solver");
      }
    },
    mesh_v->variant());
}
