#include <scheme/Scalar3WavesKineticSolver.hpp>

#include <language/utils/EvaluateAtPoints.hpp>
#include <mesh/Connectivity.hpp>
#include <mesh/Mesh.hpp>
#include <mesh/MeshVariant.hpp>
#include <scheme/DiscreteFunctionDescriptorP0Vector.hpp>
#include <scheme/DiscreteFunctionP0Vector.hpp>
#include <scheme/DiscreteFunctionUtils.hpp>
#include <scheme/IDiscreteFunctionDescriptor.hpp>

template <MeshConcept MeshType>
class Scalar3WavesKineticSolver
{
 private:
  constexpr static size_t Dimension = MeshType::Dimension;
  using Rd                          = TinyVector<Dimension>;

  using DiscreteScalarFunction = DiscreteFunctionP0<double>;

  const FunctionSymbolId m_flux_function_id;
  const double m_lambda;
  const DiscreteFunctionP0Vector<const double> m_Fn;
  const std::shared_ptr<const MeshType> p_mesh;
  const MeshType& m_mesh;
  const CellValue<const double> m_dx_table          = MeshDataManager::instance().getMeshData(m_mesh).Vj();
  const CellValue<const TinyVector<Dimension>> m_xj = MeshDataManager::instance().getMeshData(m_mesh).xj();

 public:
  CellValue<const double>
  getA(const CellValue<const double>& u) const
  {
    CellValue<double> Au{m_mesh.connectivity()};
    EvaluateAtPoints<double(double)>::evaluateTo(m_flux_function_id, u, Au);
    return Au;
  }

  CellArray<const double>
  compute_M(const CellValue<const double>& u, const CellValue<const double> Au) const
  {
    CellArray<double> M{m_mesh.connectivity(), 3};

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      // M[cell_id][0] = (1. / m_lambda) * Au[cell_id];
      // M[cell_id][1] = u[cell_id] - (1. / m_lambda) * Au[cell_id];
      // M[cell_id][2] = 0;
      // if (m_xj[cell_id][0] >= 0.5) {
      //   M[cell_id][0] = (1. / m_lambda) * Au[cell_id];
      //   M[cell_id][1] = u[cell_id] - (1. / m_lambda) * Au[cell_id];
      //   M[cell_id][2] = 0;
      // } else {
      //   M[cell_id][0] = 0;
      //   M[cell_id][1] = u[cell_id] + (1. / m_lambda) * Au[cell_id];
      //   M[cell_id][2] = -(1. / m_lambda) * Au[cell_id];
      // }
      M[cell_id][0] =
        (1. / 3) * u[cell_id] + (1. / (pow((m_lambda - 1), 2) + 1 + pow(m_lambda, 2))) * Au[cell_id] * (m_lambda - 1);
      M[cell_id][1] =
        (1. / 3) * u[cell_id] + (1. / (pow((m_lambda - 1), 2) + 1 + pow(m_lambda, 2))) * Au[cell_id] * (1);
      M[cell_id][2] =
        (1. / 3) * u[cell_id] + (1. / (pow((m_lambda - 1), 2) + 1 + pow(m_lambda, 2))) * Au[cell_id] * (-m_lambda);
    }
    return M;
  }

  DiscreteFunctionP0Vector<double>
  compute_delta1Fn(const DiscreteFunctionP0Vector<const double>& Fn)
  {
    const size_t nb_components = m_Fn.size();

    DiscreteFunctionP0Vector<double> deltaFn(p_mesh, nb_components);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      deltaFn[cell_id][0] = Fn[cell_id][0] - Fn[(cell_id + m_mesh.numberOfCells() - 1) % m_mesh.numberOfCells()][0];
      // deltaFn[cell_id][1] = 0;
      deltaFn[cell_id][1] = Fn[cell_id][1] - Fn[(cell_id + m_mesh.numberOfCells() - 1) % m_mesh.numberOfCells()][1];
      deltaFn[cell_id][2] = Fn[(cell_id + 1) % m_mesh.numberOfCells()][2] - Fn[cell_id][2];
    }

    return deltaFn;
  }

  DiscreteFunctionP0Vector<double>
  compute_delta2Fn(const DiscreteFunctionP0Vector<const double>& Fn)
  {
    const size_t nb_components = m_Fn.size();

    DiscreteFunctionP0Vector<double> deltaFn(p_mesh, nb_components);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      deltaFn[cell_id][0] = 1. / 3 * Fn[(cell_id + 1) % m_mesh.numberOfCells()][0] + 0.5 * Fn[cell_id][0] -
                            Fn[(cell_id + m_mesh.numberOfCells() - 1) % m_mesh.numberOfCells()][0] +
                            1. / 6 * Fn[(cell_id + m_mesh.numberOfCells() - 2) % m_mesh.numberOfCells()][0];
      // deltaFn[cell_id][1] = 0;
      deltaFn[cell_id][1] = 1. / 3 * Fn[(cell_id + 1) % m_mesh.numberOfCells()][1] + 0.5 * Fn[cell_id][1] -
                            Fn[(cell_id + m_mesh.numberOfCells() - 1) % m_mesh.numberOfCells()][1] +
                            1. / 6 * Fn[(cell_id + m_mesh.numberOfCells() - 2) % m_mesh.numberOfCells()][1];

      deltaFn[cell_id][2] = -1. / 3 * Fn[(cell_id + m_mesh.numberOfCells() - 1) % m_mesh.numberOfCells()][2] -
                            0.5 * Fn[cell_id][2] + Fn[(cell_id + 1) % m_mesh.numberOfCells()][2] -
                            1. / 6 * Fn[(cell_id + 2) % m_mesh.numberOfCells()][2];
    }

    return deltaFn;
  }

  CellValue<double>
  compute_PFnp1(const DiscreteFunctionP0Vector<const double> F, const double& dt, const double& dx)
  {
    CellValue<double> PFnp1{m_mesh.connectivity()};

    DiscreteFunctionP0Vector<double> deltaF  = compute_delta2Fn(F);
    DiscreteFunctionP0Vector<double> deltaF0 = compute_delta2Fn(m_Fn);

    const double c = 0.5 * dt / dx;

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      // PFnp1[cell_id] = m_Fn[cell_id][0] + m_Fn[cell_id][1] + m_Fn[cell_id][2] -
      //                  c * m_lambda * (deltaF[cell_id][0] - deltaF[cell_id][2]) -
      //                  c * m_lambda * (deltaF0[cell_id][0] - deltaF0[cell_id][2]);
      PFnp1[cell_id] =
        m_Fn[cell_id][0] + m_Fn[cell_id][1] + m_Fn[cell_id][2] -
        c * ((m_lambda - 1) * deltaF[cell_id][0] + deltaF[cell_id][1] - m_lambda * deltaF[cell_id][2]) -
        c * ((m_lambda - 1) * deltaF0[cell_id][0] + deltaF0[cell_id][1] - m_lambda * deltaF0[cell_id][2]);
    }
    return PFnp1;
  }

  CellValue<double>
  compute_PFn(const DiscreteFunctionP0Vector<const double> F)
  {
    CellValue<double> PFn{m_mesh.connectivity()};

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      PFn[cell_id] = F[cell_id][0] + F[cell_id][1] + F[cell_id][2];
    }
    return PFn;
  }

  DiscreteFunctionP0Vector<double>
  apply(const double& dt, const double& eps)
  {
    const DiscreteFunctionP0Vector<const double>& F0 = m_Fn;
    DiscreteFunctionP0Vector<double> Fnp1            = copy(F0);
    DiscreteFunctionP0Vector<double> deltaFn         = compute_delta2Fn(F0);

    for (size_t p = 0; p < 2; ++p) {
      CellId first_cell_id = 0;
      const double dx      = m_dx_table[first_cell_id];

      DiscreteFunctionP0Vector<double> deltaFnp1 = compute_delta2Fn(Fnp1);

      const CellValue<const double> PFnp1  = compute_PFnp1(Fnp1, dt, dx);
      const CellValue<const double> APFnp1 = getA(PFnp1);
      const CellArray<const double> MPFnp1 = compute_M(PFnp1, APFnp1);
      const CellValue<const double> PFn    = compute_PFn(F0);
      const CellValue<const double> APFn   = getA(PFn);
      const CellArray<const double> MPFn   = compute_M(PFn, APFn);

      const double c     = 0.5 * dt / dx;
      const double c_eps = 0.5 * dt / eps;

      for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
        // Fnp1[cell_id][0] =
        //   1. / (1 + c_eps) *
        //   (c_eps * MPFnp1[cell_id][0] + F0[cell_id][0] - c * m_lambda * (deltaFnp1[cell_id][0] + deltaFn[cell_id][0])
        //   +
        //    c_eps * (MPFn[cell_id][0] - F0[cell_id][0]));
        // Fnp1[cell_id][1] = 1. / (1 + c_eps) *
        //                    (c_eps * MPFnp1[cell_id][1] + F0[cell_id][1] + c_eps * (MPFn[cell_id][1] -
        //                    F0[cell_id][1]));
        // Fnp1[cell_id][2] =
        //   1. / (1 + c_eps) *
        //   (c_eps * MPFnp1[cell_id][2] + F0[cell_id][2] + c * m_lambda * (deltaFnp1[cell_id][2] +
        //   deltaFn[cell_id][2])
        //   +
        //    c_eps * (MPFn[cell_id][2] - F0[cell_id][2]));

        Fnp1[cell_id][0] = 1. / (1 + c_eps) *
                           (c_eps * MPFnp1[cell_id][0] + F0[cell_id][0] -
                            c * (m_lambda - 1) * (deltaFnp1[cell_id][0] + deltaFn[cell_id][0]) +
                            c_eps * (MPFn[cell_id][0] - F0[cell_id][0]));
        Fnp1[cell_id][1] =
          1. / (1 + c_eps) *
          (c_eps * MPFnp1[cell_id][1] + F0[cell_id][1] - c * (deltaFnp1[cell_id][1] + deltaFn[cell_id][1]) +
           c_eps * (MPFn[cell_id][1] - F0[cell_id][1]));
        Fnp1[cell_id][2] =
          1. / (1 + c_eps) *
          (c_eps * MPFnp1[cell_id][2] + F0[cell_id][2] + c * m_lambda * (deltaFnp1[cell_id][2] + deltaFn[cell_id][2]) +
           c_eps * (MPFn[cell_id][2] - F0[cell_id][2]));
      }
    }

    return Fnp1;
  }

  Scalar3WavesKineticSolver(std::shared_ptr<const MeshType> mesh,
                            const FunctionSymbolId& function_symbol_id,
                            const double& lambda,
                            const DiscreteFunctionP0Vector<const double>& Fn)
    : m_flux_function_id{function_symbol_id}, m_lambda{lambda}, m_Fn{Fn}, p_mesh{mesh}, m_mesh{*mesh}
  {}

  ~Scalar3WavesKineticSolver() = default;
};

std::shared_ptr<const DiscreteFunctionVariant>
scalar3WavesKineticSolver(const double& dt,
                          const FunctionSymbolId& flux_function_id,
                          const double& lambda,
                          const double& eps,
                          const std::shared_ptr<const DiscreteFunctionVariant>& F)
{
  const std::shared_ptr<const MeshVariant> mesh_v = getCommonMesh({F});

  std::shared_ptr<const DiscreteFunctionVariant> new_F;

  std::visit(
    [&](auto&& p_mesh) {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      if constexpr (is_polygonal_mesh_v<MeshType> and (MeshType::Dimension == 1)) {
        const auto& Fn = F->get<DiscreteFunctionP0Vector<const double>>();

        Scalar3WavesKineticSolver solver(p_mesh, flux_function_id, lambda, Fn);

        new_F = std::make_shared<DiscreteFunctionVariant>(solver.apply(dt, eps));
      } else {
        throw NotImplementedError("Invalid mesh type for Scalar Kinetic solver");
      }
    },
    mesh_v->variant());

  return new_F;
}
