#include <scheme/ScalarKineticSolver.hpp>

#include <language/utils/EvaluateAtPoints.hpp>
#include <mesh/Connectivity.hpp>
#include <mesh/Mesh.hpp>
#include <mesh/MeshVariant.hpp>
#include <scheme/DiscreteFunctionDescriptorP0Vector.hpp>
#include <scheme/DiscreteFunctionP0Vector.hpp>
#include <scheme/DiscreteFunctionUtils.hpp>
#include <scheme/IDiscreteFunctionDescriptor.hpp>

template <MeshConcept MeshType>
std::shared_ptr<const DiscreteFunctionVariant>
getScalarKineticWaves(const std::shared_ptr<const MeshType> p_mesh,
                      const std::vector<double>& lambda_vector,
                      DiscreteFunctionP0<const double> u,
                      DiscreteFunctionP0<const double> Au)
{
  if (lambda_vector.size() < 2) {
    throw NormalError("lambda vector must be of dimension greater than 2");
  }

  DiscreteFunctionP0Vector<double> Fn(p_mesh, lambda_vector.size());

  for (CellId cell_id = 0; cell_id < p_mesh->numberOfCells(); ++cell_id) {
    const size_t nb_waves       = lambda_vector.size();
    double L2norm_lambda_vector = 0;
    for (size_t i = 0; i < nb_waves; ++i) {
      L2norm_lambda_vector += std::pow(lambda_vector[i], 2);
    }
    L2norm_lambda_vector = std::sqrt(L2norm_lambda_vector);

    for (size_t i = 0; i < nb_waves; ++i) {
      Fn[cell_id][i] =
        (1. / nb_waves) * u[cell_id] + (1. / pow(L2norm_lambda_vector, 2)) * Au[cell_id] * lambda_vector[i];
    }
  }

  return std::make_shared<DiscreteFunctionVariant>(Fn);
}

std::shared_ptr<const DiscreteFunctionVariant>
getScalarKineticWaves(const std::vector<double>& lambda_vector,
                      const std::shared_ptr<const DiscreteFunctionVariant>& u,
                      const std::shared_ptr<const DiscreteFunctionVariant>& Au)
{
  const std::shared_ptr<const MeshVariant> mesh_v = getCommonMesh({u, Au});
  if (mesh_v.use_count() == 0) {
    throw NormalError("u and Au have to be defined on the same mesh");
  }

  return std::visit(
    [&](auto&& p_mesh) -> std::shared_ptr<const DiscreteFunctionVariant> {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      if constexpr (is_polygonal_mesh_v<MeshType> and (MeshType::Dimension == 1)) {
        auto uh = u->get<DiscreteFunctionP0<const double>>();

        auto Auh = Au->get<DiscreteFunctionP0<const double>>();

        return getScalarKineticWaves(p_mesh, lambda_vector, uh, Auh);
      } else {
        throw NotImplementedError("Invalid mesh type for Scalar Kinetic solver");
      }
    },
    mesh_v->variant());
}

template <MeshConcept MeshType>
class ScalarKineticSolver
{
 private:
  constexpr static size_t Dimension = MeshType::Dimension;
  using Rd                          = TinyVector<Dimension>;

  using DiscreteScalarFunction = DiscreteFunctionP0<double>;

  const FunctionSymbolId m_flux_function_id;
  const std::vector<double>& m_lambda_vector;
  const DiscreteFunctionP0Vector<const double> m_Fn;
  const std::shared_ptr<const MeshType> p_mesh;
  const MeshType& m_mesh;
  const CellValue<const double> m_dx_table          = MeshDataManager::instance().getMeshData(m_mesh).Vj();
  const CellValue<const TinyVector<Dimension>> m_xj = MeshDataManager::instance().getMeshData(m_mesh).xj();

 public:
  CellValue<const double>
  getA(const CellValue<const double>& u) const
  {
    CellValue<double> Au{m_mesh.connectivity()};
    EvaluateAtPoints<double(double)>::evaluateTo(m_flux_function_id, u, Au);
    return Au;
  }

  CellArray<const double>
  compute_M(const CellValue<const double>& u, const CellValue<const double> Au) const
  {
    const size_t nb_components = m_Fn.size();
    CellArray<double> M{m_mesh.connectivity(), nb_components};

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      const size_t nb_waves       = m_lambda_vector.size();
      double L2norm_lambda_vector = 0;
      for (size_t i = 0; i < nb_waves; ++i) {
        L2norm_lambda_vector += pow(m_lambda_vector[i], 2);
      }
      L2norm_lambda_vector = sqrt(L2norm_lambda_vector);

      for (size_t i = 0; i < nb_waves; ++i) {
        M[cell_id][i] =
          (1. / nb_waves) * u[cell_id] + (1. / pow(L2norm_lambda_vector, 2)) * Au[cell_id] * m_lambda_vector[i];
      }
    }
    return M;
  }

  DiscreteFunctionP0Vector<double>
  compute_delta1Fn(const DiscreteFunctionP0Vector<const double>& Fn)
  {
    const size_t nb_components = m_Fn.size();

    DiscreteFunctionP0Vector<double> deltaFn(p_mesh, nb_components);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < m_lambda_vector.size(); ++i) {
        if (m_lambda_vector[i] < 0) {
          deltaFn[cell_id][i] = Fn[(cell_id + 1) % m_mesh.numberOfCells()][i] - Fn[cell_id][i];
        } else if (m_lambda_vector[i] > 0) {
          deltaFn[cell_id][i] = Fn[cell_id][i] - Fn[(cell_id + m_mesh.numberOfCells() - 1) % m_mesh.numberOfCells()][i];
        } else {
          deltaFn[cell_id][i] = 0;
        }
      }
    }

    return deltaFn;
  }

  DiscreteFunctionP0Vector<double>
  compute_delta2Fn(const DiscreteFunctionP0Vector<const double>& Fn)
  {
    const size_t nb_components = m_Fn.size();

    DiscreteFunctionP0Vector<double> deltaFn(p_mesh, nb_components);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < m_lambda_vector.size(); ++i) {
        if (m_lambda_vector[i] < 0) {
          deltaFn[cell_id][i] = -1. / 3 * Fn[(cell_id + m_mesh.numberOfCells() - 1) % m_mesh.numberOfCells()][i] -
                                0.5 * Fn[cell_id][i] + Fn[(cell_id + 1) % m_mesh.numberOfCells()][i] -
                                1. / 6 * Fn[(cell_id + 2) % m_mesh.numberOfCells()][i];
        } else if (m_lambda_vector[i] > 0) {
          deltaFn[cell_id][i] = 1. / 3 * Fn[(cell_id + 1) % m_mesh.numberOfCells()][i] + 0.5 * Fn[cell_id][i] -
                                Fn[(cell_id + m_mesh.numberOfCells() - 1) % m_mesh.numberOfCells()][i] +
                                1. / 6 * Fn[(cell_id + m_mesh.numberOfCells() - 2) % m_mesh.numberOfCells()][i];
        } else {
          deltaFn[cell_id][i] = 0;
        }
      }
    }

    return deltaFn;
  }

  DiscreteFunctionP0Vector<double>
  compute_delta3Fn(const DiscreteFunctionP0Vector<const double>& Fn)
  {
    const size_t nb_waves = m_lambda_vector.size();

    DiscreteFunctionP0Vector<double> deltaFn(p_mesh, nb_waves);
    const size_t last_cell     = m_mesh.numberOfCells() - 1;
    const size_t last2_cell    = m_mesh.numberOfCells() - 2;
    const CellId last_cell_id  = last_cell;
    const CellId last2_cell_id = last2_cell;
    const CellId first_cell_id = 0;
    const CellId scnd_cell_id  = 1;

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      const CellId prev2_cell_id = cell_id - 2;
      const CellId prev_cell_id  = cell_id - 1;
      const CellId next_cell_id  = cell_id + 1;
      const CellId next2_cell_id = cell_id + 2;
      const double dx            = m_dx_table[cell_id];

      for (size_t i = 0; i < nb_waves; ++i) {
        if (m_lambda_vector[i] > 0) {
          if (cell_id == last_cell_id) {
            double last_dx = 0.5 * (dx + m_dx_table[first_cell_id]);
            double xj      = m_xj[cell_id][0];
            double xjm1    = m_xj[prev_cell_id][0];
            double xjm2    = m_xj[prev2_cell_id][0];

            deltaFn[cell_id][i] =
              Fn[prev2_cell_id][i] * (-last_dx) * (xj - xjm1) / ((xjm2 - xj - last_dx) * (xjm2 - xj) * (xjm2 - xjm1)) +
              Fn[prev_cell_id][i] * (-last_dx) * (xj - xjm2) / ((xjm1 - xj - last_dx) * (xjm1 - xj) * (xjm1 - xjm2)) +
              Fn[cell_id][i] * (1. / (xj - xjm2) + 1. / (xj - xjm1) + 1. / (-last_dx)) +
              Fn[first_cell_id][i] * (xj - xjm1) * (xj - xjm2) /
                ((last_dx) * (last_dx + xj - xjm1) * (last_dx + xj - xjm2));

          } else if (cell_id == first_cell_id) {
            double first_dx = 0.5 * (dx + m_dx_table[last_cell_id]);
            double xj       = m_xj[cell_id][0];
            double xjp1     = m_xj[next_cell_id][0];
            double xjm1     = m_xj[last_cell_id][0];
            double xjm2     = m_xj[last2_cell_id][0];

            deltaFn[cell_id][i] =
              Fn[last2_cell_id][i] * (xj - xjp1) * (first_dx) /
                ((xjm2 - xjm1 - first_dx + xj - xjp1) * (xjm2 - xjm1 - first_dx) * (xjm2 - xjm1)) +
              Fn[last_cell_id][i] * (xj - xjp1) * (first_dx + xjm1 - xjm2) /
                ((-first_dx + xj - xjp1) * (-first_dx) * (xjm1 - xjm2)) +
              Fn[cell_id][i] * (1. / (first_dx + xjm1 - xjm2) + 1. / (first_dx) + 1. / (xj - xjp1)) +
              Fn[next_cell_id][i] * (first_dx) * (first_dx + xjm1 - xjm2) /
                ((xjp1 - xj) * (xjp1 - xj + first_dx) * (xjp1 - xj + first_dx + xjm1 - xjm2));

          } else if (cell_id - 1 == first_cell_id) {
            double first_dx = 0.5 * (m_dx_table[first_cell_id] + m_dx_table[last_cell_id]);
            double xj       = m_xj[cell_id][0];
            double xjp1     = m_xj[next_cell_id][0];
            double xjm1     = m_xj[prev_cell_id][0];

            deltaFn[cell_id][i] = Fn[last_cell_id][i] * (xj - xjp1) * (xj - xjm1) /
                                    ((-first_dx + xjm1 - xjp1) * (-first_dx + xjm1 - xj) * (-first_dx)) +
                                  Fn[prev_cell_id][i] * (xj - xjp1) * (xj - xjm1 + first_dx) /
                                    ((xjm1 - xjp1) * (xjm1 - xj) * (xjm1 - xjm1 + first_dx)) +
                                  Fn[cell_id][i] * (1. / (xj - xjm1 + first_dx) + 1. / (xj - xjm1) + 1. / (xj - xjp1)) +
                                  Fn[next_cell_id][i] * (xj - xjm1) * (xj - xjm1 + first_dx) /
                                    ((xjp1 - xj) * (xjp1 - xjm1) * (xjp1 - xjm1 + first_dx));

          } else {
            double xj   = m_xj[cell_id][0];
            double xjp1 = m_xj[next_cell_id][0];
            double xjm1 = m_xj[prev_cell_id][0];
            double xjm2 = m_xj[prev2_cell_id][0];

            deltaFn[cell_id][i] =
              Fn[prev2_cell_id][i] * (xj - xjp1) * (xj - xjm1) / ((xjm2 - xjp1) * (xjm2 - xj) * (xjm2 - xjm1)) +
              Fn[prev_cell_id][i] * (xj - xjp1) * (xj - xjm2) / ((xjm1 - xjp1) * (xjm1 - xj) * (xjm1 - xjm2)) +
              Fn[cell_id][i] * (1. / (xj - xjm2) + 1. / (xj - xjm1) + 1. / (xj - xjp1)) +
              Fn[next_cell_id][i] * (xj - xjm1) * (xj - xjm2) / ((xjp1 - xj) * (xjp1 - xjm1) * (xjp1 - xjm2));
          }
        } else {
          if (cell_id + 1 == last_cell_id) {
            double last_dx = 0.5 * (m_dx_table[last_cell_id] + m_dx_table[first_cell_id]);
            double xj      = m_xj[cell_id][0];
            double xjp1    = m_xj[next_cell_id][0];
            double xjm1    = m_xj[prev_cell_id][0];

            deltaFn[cell_id][i] =
              Fn[prev_cell_id][i] * (xj - xjp1) * (xj - xjp1 - last_dx) /
                ((xjm1 - xj) * (xjm1 - xjp1) * (xjm1 - xjp1 - last_dx)) +
              Fn[cell_id][i] * (1. / (xj - xjp1 - last_dx) + 1. / (xj - xjp1) + 1. / (xj - xjm1)) +
              Fn[next_cell_id][i] * (xj - xjm1) * (xj - xjp1 - last_dx) / ((xjp1 - xjm1) * (xjp1 - xj) * (-last_dx)) +
              Fn[first_cell_id][i] * (xj - xjm1) * (xj - xjp1) /
                ((last_dx + xjp1 - xjm1) * (last_dx + xjp1 - xj) * (last_dx));

          } else if (cell_id == last_cell_id) {
            double last_dx = 0.5 * (dx + m_dx_table[first_cell_id]);
            double xj      = m_xj[cell_id][0];
            double xjp1    = m_xj[first_cell_id][0];
            double xjm1    = m_xj[prev_cell_id][0];
            double xjp2    = m_xj[scnd_cell_id][0];

            deltaFn[cell_id][i] =
              Fn[prev_cell_id][i] * (-last_dx) * (-last_dx + xjp1 - xjp2) /
                ((xjm1 - xj) * (xjm1 - xj - last_dx) * (xjm1 - xj - last_dx + xjp1 - xjp2)) +
              Fn[cell_id][i] * (1. / (-last_dx + xjp1 - xjp2) + 1. / (-last_dx) + 1. / (xj - xjm1)) +
              Fn[first_cell_id][i] * (xj - xjm1) * (-last_dx + xjp1 - xjp2) /
                ((last_dx + xj - xjm1) * (last_dx) * (xjp1 - xjp2)) +
              Fn[scnd_cell_id][i] * (xj - xjm1) * (-last_dx) /
                ((xjp2 - xjp1 + last_dx + xj - xjm1) * (last_dx - xjp1 + xjp2) * (xjp2 - xjp1));

          } else if (cell_id == first_cell_id) {
            double first_dx = 0.5 * (dx + m_dx_table[last_cell_id]);
            double xj       = m_xj[cell_id][0];
            double xjp1     = m_xj[next_cell_id][0];
            double xjp2     = m_xj[next2_cell_id][0];

            deltaFn[cell_id][i] =
              Fn[last_cell_id][i] * (xj - xjp1) * (xj - xjp2) /
                ((-first_dx) * (-first_dx + xj - xjp1) * (-first_dx + xj - xjp2)) +
              Fn[cell_id][i] * (1. / (xj - xjp2) + 1. / (xj - xjp1) + 1. / (first_dx)) +
              Fn[next_cell_id][i] * (first_dx) * (xj - xjp2) / ((xjp1 - xj + first_dx) * (xjp1 - xj) * (xjp1 - xjp2)) +
              Fn[next2_cell_id][i] * (first_dx) * (xj - xjp1) / ((xjp2 - xj + first_dx) * (xjp2 - xj) * (xjp2 - xjp1));

          } else {
            double xj   = m_xj[cell_id][0];
            double xjp1 = m_xj[next_cell_id][0];
            double xjm1 = m_xj[prev_cell_id][0];
            double xjp2 = m_xj[next2_cell_id][0];

            deltaFn[cell_id][i] =
              Fn[prev_cell_id][i] * (xj - xjp1) * (xj - xjp2) / ((xjm1 - xj) * (xjm1 - xjp1) * (xjm1 - xjp2)) +
              Fn[cell_id][i] * (1. / (xj - xjp2) + 1. / (xj - xjp1) + 1. / (xj - xjm1)) +
              Fn[next_cell_id][i] * (xj - xjm1) * (xj - xjp2) / ((xjp1 - xjm1) * (xjp1 - xj) * (xjp1 - xjp2)) +
              Fn[next2_cell_id][i] * (xj - xjm1) * (xj - xjp1) / ((xjp2 - xjm1) * (xjp2 - xj) * (xjp2 - xjp1));
          }
        }
      }
    }
    return deltaFn;
  }

  CellValue<double>
  compute_PFnp1(const DiscreteFunctionP0Vector<const double> F, const double& dt, const double& dx)
  {
    CellValue<double> PFnp1{m_mesh.connectivity()};

    DiscreteFunctionP0Vector<double> deltaF  = compute_delta2Fn(F);
    DiscreteFunctionP0Vector<double> deltaF0 = compute_delta2Fn(m_Fn);

    const double c = 0.5 * dt / dx;

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      double PFn = 0;
      for (size_t i = 0; i < m_lambda_vector.size(); ++i) {
        PFn += m_Fn[cell_id][i] - c * m_lambda_vector[i] * (deltaF[cell_id][i] + deltaF0[cell_id][i]);
      }

      PFnp1[cell_id] = PFn;
    }
    return PFnp1;
  }

  CellValue<double>
  compute_PFn(const DiscreteFunctionP0Vector<const double> F)
  {
    CellValue<double> PFn{m_mesh.connectivity()};

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      double PF = 0;
      for (size_t i = 0; i < m_lambda_vector.size(); ++i) {
        PF += F[cell_id][i];
      }
      PFn[cell_id] = PF;
    }
    return PFn;
  }

  DiscreteFunctionP0Vector<double>
  apply(const double& dt, const double& eps)
  {
    const DiscreteFunctionP0Vector<const double>& F0 = m_Fn;
    DiscreteFunctionP0Vector<double> Fnp1            = copy(F0);
    DiscreteFunctionP0Vector<double> deltaFn         = compute_delta2Fn(F0);

    for (size_t p = 0; p < 2; ++p) {
      CellId first_cell_id = 0;
      const double dx      = m_dx_table[first_cell_id];

      DiscreteFunctionP0Vector<double> deltaFnp1 = compute_delta2Fn(Fnp1);

      const CellValue<const double> PFnp1  = compute_PFnp1(Fnp1, dt, dx);
      const CellValue<const double> APFnp1 = getA(PFnp1);
      const CellArray<const double> MPFnp1 = compute_M(PFnp1, APFnp1);
      const CellValue<const double> PFn    = compute_PFn(F0);
      const CellValue<const double> APFn   = getA(PFn);
      const CellArray<const double> MPFn   = compute_M(PFn, APFn);

      const double c2 = (eps * dt / dx);
      const double c1 = 1. / (2 * eps + dt);

      for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
        for (size_t i = 0; i < m_lambda_vector.size(); ++i) {
          Fnp1[cell_id][i] = c1 * (dt * MPFnp1[cell_id][i] + 2 * eps * F0[cell_id][i] -
                                   c2 * m_lambda_vector[i] * (deltaFnp1[cell_id][i] + deltaFn[cell_id][i]) +
                                   dt * (MPFn[cell_id][i] - F0[cell_id][i]));
        }
      }
    }

    return Fnp1;
  }

  ScalarKineticSolver(std::shared_ptr<const MeshType> mesh,
                      const FunctionSymbolId& function_symbol_id,
                      const std::vector<double>& lambda_vector,
                      const DiscreteFunctionP0Vector<const double>& Fn)
    : m_flux_function_id{function_symbol_id}, m_lambda_vector{lambda_vector}, m_Fn{Fn}, p_mesh{mesh}, m_mesh{*mesh}
  {
    if (lambda_vector.size() != Fn.size()) {
      throw NormalError("Incompatible dimensions of lambda vector and Fn");
    }
  }

  ~ScalarKineticSolver() = default;
};

std::shared_ptr<const DiscreteFunctionVariant>
scalarKineticSolver(const double& dt,
                    const FunctionSymbolId& flux_function_id,
                    const std::vector<double>& lambda_vector,
                    const double& eps,
                    const std::shared_ptr<const DiscreteFunctionVariant>& F)
{
  const std::shared_ptr<const MeshVariant> mesh_v = getCommonMesh({F});

  std::shared_ptr<const DiscreteFunctionVariant> new_F;

  std::visit(
    [&](auto&& p_mesh) {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      if constexpr (is_polygonal_mesh_v<MeshType> and (MeshType::Dimension == 1)) {
        const auto& Fn = F->get<DiscreteFunctionP0Vector<const double>>();

        ScalarKineticSolver solver(p_mesh, flux_function_id, lambda_vector, Fn);

        new_F = std::make_shared<DiscreteFunctionVariant>(solver.apply(dt, eps));
      } else {
        throw NotImplementedError("Invalid mesh type for Scalar Kinetic solver");
      }
    },
    mesh_v->variant());

  return new_F;
}
