#include <scheme/WavesEquationKineticSolver.hpp>

#include <language/utils/EvaluateAtPoints.hpp>
#include <mesh/Connectivity.hpp>
#include <mesh/Mesh.hpp>
#include <mesh/MeshVariant.hpp>
#include <scheme/DiscreteFunctionDescriptorP0Vector.hpp>
#include <scheme/DiscreteFunctionP0Vector.hpp>
#include <scheme/DiscreteFunctionUtils.hpp>
#include <scheme/IDiscreteFunctionDescriptor.hpp>

#include <tuple>

template <MeshConcept MeshType>
std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>>
getWavesEquationKineticWaves(const std::shared_ptr<const MeshType> p_mesh,
                             const std::vector<double>& lambda_vector,
                             DiscreteFunctionP0<const double> u,
                             DiscreteFunctionP0<const double> p,
                             TinyMatrix<2> A)
{
  if (lambda_vector.size() < 2) {
    throw NormalError("lambda vector must be of dimension greater than 2");
  }

  DiscreteFunctionP0Vector<double> Fn_u(p_mesh, lambda_vector.size());
  DiscreteFunctionP0Vector<double> Fn_p(p_mesh, lambda_vector.size());

  const size_t nb_waves           = lambda_vector.size();
  double inv_L2norm_lambda_vector = 0;
  for (size_t i = 0; i < nb_waves; ++i) {
    inv_L2norm_lambda_vector += std::pow(lambda_vector[i], 2);
  }
  inv_L2norm_lambda_vector = 1. / inv_L2norm_lambda_vector;

  for (CellId cell_id = 0; cell_id < p_mesh->numberOfCells(); ++cell_id) {
    for (size_t i = 0; i < nb_waves; ++i) {
      Fn_u[cell_id][i] = (1. / nb_waves) * u[cell_id] +
                         inv_L2norm_lambda_vector * (A(0, 0) * u[cell_id] + A(0, 1) * p[cell_id]) * lambda_vector[i];
      Fn_p[cell_id][i] = (1. / nb_waves) * p[cell_id] +
                         inv_L2norm_lambda_vector * (A(1, 0) * u[cell_id] + A(1, 1) * p[cell_id]) * lambda_vector[i];
      // std::cout << "lambda = " << lambda_vector[i] << "; l/|L|^2 = " << inv_L2norm_lambda_vector * lambda_vector[i]
      // << "\n";
    }
  }

  return std::make_tuple(std::make_shared<DiscreteFunctionVariant>(Fn_u),
                         std::make_shared<DiscreteFunctionVariant>(Fn_p));
}

std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>>
getWavesEquationKineticWaves(const std::vector<double>& lambda_vector,
                             const std::shared_ptr<const DiscreteFunctionVariant>& u,
                             const std::shared_ptr<const DiscreteFunctionVariant>& p,
                             const TinyMatrix<2>& A)
{
  const std::shared_ptr<const MeshVariant> mesh_v = getCommonMesh({u, p});
  if (mesh_v.use_count() == 0) {
    throw NormalError("u and p have to be defined on the same mesh");
  }

  return std::visit(
    [&](auto&& p_mesh)
      -> std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>> {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      if constexpr (is_polygonal_mesh_v<MeshType> and (MeshType::Dimension == 1)) {
        auto uh = u->get<DiscreteFunctionP0<const double>>();
        auto ph = p->get<DiscreteFunctionP0<const double>>();

        return getWavesEquationKineticWaves(p_mesh, lambda_vector, uh, ph, A);
      } else {
        throw NotImplementedError("Invalid mesh type for Scalar Kinetic solver");
      }
    },
    mesh_v->variant());
}

template <MeshConcept MeshType>
class WavesEquationKineticSolver
{
 private:
  constexpr static size_t Dimension = MeshType::Dimension;
  using Rd                          = TinyVector<Dimension>;

  using DiscreteWavesEquationFunction = DiscreteFunctionP0<double>;

  const double m_dt;
  const TinyMatrix<2> m_A;
  const std::vector<double>& m_lambda_vector;
  const DiscreteFunctionP0Vector<const double> m_Fn_u;
  const DiscreteFunctionP0Vector<const double> m_Fn_p;
  const std::shared_ptr<const MeshType> p_mesh;
  const MeshType& m_mesh;
  const CellValue<const double> m_dx_table          = MeshDataManager::instance().getMeshData(m_mesh).Vj();
  const CellValue<const TinyVector<Dimension>> m_xj = MeshDataManager::instance().getMeshData(m_mesh).xj();
  const NodeValue<const TinyVector<Dimension>> m_xr = m_mesh.xr();

 public:
  CellArray<const double>
  getA(const CellValue<const double>& u, const CellValue<const double>& p) const
  {
    CellArray<double> vec_A{m_mesh.connectivity(), 2};
    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < 2; ++i) {
        vec_A[cell_id][i] = m_A(i, 0) * u[cell_id] + m_A(i, 1) * p[cell_id];
      }
    }
    return vec_A;
  }

  CellArray<const double>
  compute_M(const CellValue<const double>& u, const CellValue<const double> Au) const
  {
    const size_t nb_components = m_Fn_u.size();
    CellArray<double> M{m_mesh.connectivity(), nb_components};

    double inv_L2norm_lambda_vector = 0;
    for (size_t i = 0; i < nb_components; ++i) {
      inv_L2norm_lambda_vector += pow(m_lambda_vector[i], 2);
    }
    inv_L2norm_lambda_vector = 1. / inv_L2norm_lambda_vector;

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < nb_components; ++i) {
        M[cell_id][i] = (1. / nb_components) * u[cell_id] + inv_L2norm_lambda_vector * Au[cell_id] * m_lambda_vector[i];
      }
    }
    return M;
  }

  DiscreteFunctionP0Vector<double>
  compute_delta1Fn(const DiscreteFunctionP0Vector<const double>& Fn)
  {
    const size_t nb_components = m_Fn_u.size();

    DiscreteFunctionP0Vector<double> deltaFn(p_mesh, nb_components);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < m_lambda_vector.size(); ++i) {
        if (m_lambda_vector[i] < 0) {
          deltaFn[cell_id][i] = Fn[(cell_id + 1) % m_mesh.numberOfCells()][i] - Fn[cell_id][i];
        } else if (m_lambda_vector[i] > 0) {
          deltaFn[cell_id][i] = Fn[cell_id][i] - Fn[(cell_id + m_mesh.numberOfCells() - 1) % m_mesh.numberOfCells()][i];
        } else {
          deltaFn[cell_id][i] = 0;
        }
      }
    }

    return deltaFn;
  }

  DiscreteFunctionP0Vector<double>
  compute_delta3Fn(const DiscreteFunctionP0Vector<const double>& Fn)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    DiscreteFunctionP0Vector<double> deltaFn(p_mesh, nb_waves);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < nb_waves; ++i) {
        const NodeId first_node_id = 0;
        const NodeId last_node_id  = m_mesh.numberOfNodes() - 1;

        const double xmin = m_xr[first_node_id][0];
        const double xmax = m_xr[last_node_id][0];

        const size_t N = m_mesh.numberOfCells();

        const auto& cell_to_node   = cell_to_node_matrix[cell_id];
        const NodeId left_node_id  = cell_to_node[0];
        const NodeId right_node_id = cell_to_node[1];

        const CellId prev2_cell_id = (cell_id + N - 2) % N;
        const CellId prev_cell_id  = (cell_id + N - 1) % N;
        const CellId next_cell_id  = (cell_id + 1) % N;
        const CellId next2_cell_id = (cell_id + 2) % N;

        const double xjm2 = (-1 * (cell_id < 2) * (xmax - xmin)) + m_xj[prev2_cell_id][0];
        const double xjm1 = (-1 * (cell_id < 1) * (xmax - xmin)) + m_xj[prev_cell_id][0];
        const double xj   = m_xj[cell_id][0];
        const double xjp1 = ((cell_id + 1 >= N) * (xmax - xmin)) + m_xj[next_cell_id][0];
        const double xjp2 = ((cell_id + 2 >= N) * (xmax - xmin)) + m_xj[next2_cell_id][0];

        if (m_lambda_vector[i] > 0) {
          deltaFn[cell_id][i] =
            Fn[prev2_cell_id][i] * (xj - xjp1) * (xj - xjm1) / ((xjm2 - xjp1) * (xjm2 - xj) * (xjm2 - xjm1)) +
            Fn[prev_cell_id][i] * (xj - xjp1) * (xj - xjm2) / ((xjm1 - xjp1) * (xjm1 - xj) * (xjm1 - xjm2)) +
            Fn[cell_id][i] * (1. / (xj - xjm2) + 1. / (xj - xjm1) + 1. / (xj - xjp1)) +
            Fn[next_cell_id][i] * (xj - xjm1) * (xj - xjm2) / ((xjp1 - xj) * (xjp1 - xjm1) * (xjp1 - xjm2));

        } else if (m_lambda_vector[i] < 0) {
          deltaFn[cell_id][i] =
            Fn[prev_cell_id][i] * (xj - xjp1) * (xj - xjp2) / ((xjm1 - xj) * (xjm1 - xjp1) * (xjm1 - xjp2)) +
            Fn[cell_id][i] * (1. / (xj - xjp2) + 1. / (xj - xjp1) + 1. / (xj - xjm1)) +
            Fn[next_cell_id][i] * (xj - xjm1) * (xj - xjp2) / ((xjp1 - xjm1) * (xjp1 - xj) * (xjp1 - xjp2)) +
            Fn[next2_cell_id][i] * (xj - xjm1) * (xj - xjp1) / ((xjp2 - xjm1) * (xjp2 - xj) * (xjp2 - xjp1));

        } else {
          deltaFn[cell_id][i] = 0;
        }
      }
    }
    return deltaFn;
  }

  NodeArray<const double>
  compute_nodeFlux1(const DiscreteFunctionP0Vector<const double>& F)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    NodeArray<double> Flux_F{m_mesh.connectivity(), nb_waves};
    Flux_F.fill(0);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < nb_waves; ++i) {
        const NodeId first_node_id = 0;
        const NodeId last_node_id  = m_mesh.numberOfNodes() - 1;

        const auto& cell_to_node   = cell_to_node_matrix[cell_id];
        const NodeId left_node_id  = cell_to_node[0];
        const NodeId right_node_id = cell_to_node[1];

        if (m_lambda_vector[i] < 0) {
          Flux_F[left_node_id][i] = F[cell_id][i];
          if (cell_id == 0) {
            Flux_F[last_node_id][i] = Flux_F[left_node_id][i];
          }
        } else if (m_lambda_vector[i] > 0) {
          Flux_F[right_node_id][i] = F[cell_id][i];
          if (cell_id == m_mesh.numberOfCells() - 1) {
            Flux_F[first_node_id][i] = Flux_F[right_node_id][i];
          }
        } else {
          Flux_F[right_node_id][i] = 0;
          Flux_F[left_node_id][i]  = 0;
        }
      }
    }

    return Flux_F;
  }

  NodeArray<const double>
  compute_nodeFlux3(const DiscreteFunctionP0Vector<const double>& F)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    NodeArray<double> Flux_F{m_mesh.connectivity(), nb_waves};
    Flux_F.fill(0);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < nb_waves; ++i) {
        const NodeId first_node_id = 0;
        const NodeId last_node_id  = m_mesh.numberOfNodes() - 1;

        const double xmin = m_xr[first_node_id][0];
        const double xmax = m_xr[last_node_id][0];

        const size_t N = m_mesh.numberOfCells();

        const auto& cell_to_node   = cell_to_node_matrix[cell_id];
        const NodeId left_node_id  = cell_to_node[0];
        const NodeId right_node_id = cell_to_node[1];

        const CellId prev2_cell_id = (cell_id + N - 2) % N;
        const CellId prev_cell_id  = (cell_id + N - 1) % N;
        const CellId next_cell_id  = (cell_id + 1) % N;
        const CellId next2_cell_id = (cell_id + 2) % N;

        const double xjm2 = (-1 * (cell_id < 2) * (xmax - xmin)) + m_xj[prev2_cell_id][0];
        const double xjm1 = (-1 * (cell_id < 1) * (xmax - xmin)) + m_xj[prev_cell_id][0];
        const double xj   = m_xj[cell_id][0];
        const double xjp1 = ((cell_id + 1 >= N) * (xmax - xmin)) + m_xj[next_cell_id][0];
        const double xjp2 = ((cell_id + 2 >= N) * (xmax - xmin)) + m_xj[next2_cell_id][0];

        const double left_node  = m_xr[left_node_id][0];
        const double right_node = m_xr[right_node_id][0];

        if (m_lambda_vector[i] < 0) {
          Flux_F[left_node_id][i] =
            F[prev_cell_id][i] * (left_node - xj) * (left_node - xjp1) / ((xjm1 - xj) * (xjm1 - xjp1)) +
            F[cell_id][i] * (left_node - xjm1) * (left_node - xjp1) / ((xj - xjm1) * (xj - xjp1)) +
            F[next_cell_id][i] * (left_node - xjm1) * (left_node - xj) / ((xjp1 - xjm1) * (xjp1 - xj));
          Flux_F[right_node_id][i] =
            F[cell_id][i] * (right_node - xjp1) * (right_node - xjp2) / ((xj - xjp1) * (xj - xjp2)) +
            F[next_cell_id][i] * (right_node - xj) * (right_node - xjp2) / ((xjp1 - xj) * (xjp1 - xjp2)) +
            F[next2_cell_id][i] * (right_node - xj) * (right_node - xjp1) / ((xjp2 - xj) * (xjp2 - xjp1));

        } else if (m_lambda_vector[i] > 0) {
          Flux_F[left_node_id][i] =
            F[prev2_cell_id][i] * (left_node - xjm1) * (left_node - xj) / ((xjm2 - xjm1) * (xjm2 - xj)) +
            F[prev_cell_id][i] * (left_node - xjm2) * (left_node - xj) / ((xjm1 - xjm2) * (xjm1 - xj)) +
            F[cell_id][i] * (left_node - xjm2) * (left_node - xjm1) / ((xj - xjm2) * (xj - xjm1));
          Flux_F[right_node_id][i] =
            F[prev_cell_id][i] * (right_node - xj) * (right_node - xjp1) / ((xjm1 - xj) * (xjm1 - xjp1)) +
            F[cell_id][i] * (right_node - xjm1) * (right_node - xjp1) / ((xj - xjm1) * (xj - xjp1)) +
            F[next_cell_id][i] * (right_node - xjm1) * (right_node - xj) / ((xjp1 - xjm1) * (xjp1 - xj));

        } else {
          Flux_F[right_node_id][i] = 0;
          Flux_F[left_node_id][i]  = 0;
        }
      }
    }

    return Flux_F;
  }

  NodeArray<const double>
  compute_Polynomial_reconstruction_3(const DiscreteFunctionP0Vector<const double>& F)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    NodeArray<double> Flux_F{m_mesh.connectivity(), nb_waves};
    Flux_F.fill(0);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < nb_waves; ++i) {
        const NodeId first_node_id = 0;
        const NodeId last_node_id  = m_mesh.numberOfNodes() - 1;

        const double xmin = m_xr[first_node_id][0];
        const double xmax = m_xr[last_node_id][0];

        const size_t N = m_mesh.numberOfCells();

        const auto& cell_to_node   = cell_to_node_matrix[cell_id];
        const NodeId left_node_id  = cell_to_node[0];
        const NodeId right_node_id = cell_to_node[1];

        const CellId prev_cell_id = (cell_id + N - 1) % N;
        const CellId next_cell_id = (cell_id + 1) % N;

        const double xj = m_xj[cell_id][0];

        const auto& prev_cell_to_node = cell_to_node_matrix[prev_cell_id];
        const NodeId left2_node_id    = prev_cell_to_node[0];

        const auto& next_cell_to_node = cell_to_node_matrix[next_cell_id];
        const NodeId right2_node_id   = next_cell_to_node[1];

        const double left_node  = m_xr[left_node_id][0];
        const double right_node = m_xr[right_node_id][0];

        const double left2_node  = (-1 * (cell_id < 1) * (xmax - xmin)) + m_xr[left2_node_id][0];
        const double right2_node = ((cell_id + 1 >= N) * (xmax - xmin)) + m_xr[right2_node_id][0];

        const TinyMatrix<2> Qj(1. / (3 * m_dx_table[prev_cell_id]) *
                                   (pow(left_node - xj, 3) - pow(left2_node - xj, 3)) -
                                 1. / (3 * m_dx_table[cell_id]) * (pow(right_node - xj, 3) - pow(left_node - xj, 3)),
                               1. / (2 * m_dx_table[prev_cell_id]) *
                                 ((left_node - xj) * (left_node - xj) - (left2_node - xj) * (left2_node - xj)),
                               1. / (3 * m_dx_table[next_cell_id]) *
                                   (pow(right2_node - xj, 3) - pow(right_node - xj, 3)) -
                                 1. / (3 * m_dx_table[cell_id]) * (pow(right_node - xj, 3) - pow(left_node - xj, 3)),
                               1. / (2 * m_dx_table[next_cell_id]) *
                                 ((right2_node - xj) * (right2_node - xj) - (right_node - xj) * (right_node - xj)));

        const TinyVector<2> Bj(F[prev_cell_id][i] - F[cell_id][i], F[next_cell_id][i] - F[cell_id][i]);

        const TinyMatrix<2> inv_Qj = inverse(Qj);
        const TinyVector<2> Uj     = inv_Qj * Bj;

        if (m_lambda_vector[i] < 0) {
          Flux_F[left_node_id][i] =
            Uj[0] * ((left_node - xj) * (left_node - xj) -
                     1. / (3 * m_dx_table[cell_id]) * (pow(right_node - xj, 3) - pow(left_node - xj, 3))) +
            Uj[1] * (left_node - xj) + F[cell_id][i];

          if (cell_id == 0) {
            Flux_F[last_node_id][i] = Flux_F[left_node_id][i];
          }

        } else if (m_lambda_vector[i] > 0) {
          Flux_F[right_node_id][i] =
            Uj[0] * ((right_node - xj) * (right_node - xj) -
                     1. / (3 * m_dx_table[cell_id]) * (pow(right_node - xj, 3) - pow(left_node - xj, 3))) +
            Uj[1] * (right_node - xj) + F[cell_id][i];

          if (cell_id == m_mesh.numberOfCells() - 1) {
            Flux_F[first_node_id][i] = Flux_F[right_node_id][i];
          }
        } else {
          Flux_F[right_node_id][i] = 0;
          Flux_F[left_node_id][i]  = 0;
        }
      }
    }

    return Flux_F;
  }

  DiscreteFunctionP0Vector<double>
  compute_delta2Fn_Unif(const DiscreteFunctionP0Vector<const double>& Fn)
  {
    const size_t nb_waves = m_Fn_u.size();

    DiscreteFunctionP0Vector<double> deltaFn(p_mesh, nb_waves);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < m_lambda_vector.size(); ++i) {
        if (m_lambda_vector[i] < 0) {
          deltaFn[cell_id][i] = -1. / 3 * Fn[(cell_id + m_mesh.numberOfCells() - 1) % m_mesh.numberOfCells()][i] -
                                0.5 * Fn[cell_id][i] + Fn[(cell_id + 1) % m_mesh.numberOfCells()][i] -
                                1. / 6 * Fn[(cell_id + 2) % m_mesh.numberOfCells()][i];
        } else if (m_lambda_vector[i] > 0) {
          deltaFn[cell_id][i] = 1. / 3 * Fn[(cell_id + 1) % m_mesh.numberOfCells()][i] + 0.5 * Fn[cell_id][i] -
                                Fn[(cell_id + m_mesh.numberOfCells() - 1) % m_mesh.numberOfCells()][i] +
                                1. / 6 * Fn[(cell_id + m_mesh.numberOfCells() - 2) % m_mesh.numberOfCells()][i];
        } else {
          deltaFn[cell_id][i] = 0;
        }
      }
    }
    return deltaFn;
  }

  DiscreteFunctionP0Vector<double>
  compute_delta4Fn_Unif(const DiscreteFunctionP0Vector<const double>& Fn)
  {
    const size_t nb_waves = m_lambda_vector.size();

    DiscreteFunctionP0Vector<double> deltaFn(p_mesh, nb_waves);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < nb_waves; ++i) {
        deltaFn[cell_id][i] = -(1. / 12) * (Fn[(cell_id + 2) % m_mesh.numberOfCells()][i] -
                                            Fn[(cell_id + m_mesh.numberOfCells() - 2) % m_mesh.numberOfCells()][i]) +
                              (2. / 3) * (Fn[(cell_id + 1) % m_mesh.numberOfCells()][i] -
                                          Fn[(cell_id + m_mesh.numberOfCells() - 1) % m_mesh.numberOfCells()][i]);

        if (m_lambda_vector[i] < 0) {
          deltaFn[cell_id][i] = 1. / 12 * Fn[(cell_id + 3) % m_mesh.numberOfCells()][i] -
                                0.5 * Fn[(cell_id + 2) % m_mesh.numberOfCells()][i] +
                                3. / 2 * Fn[(cell_id + 1) % m_mesh.numberOfCells()][i] - 5. / 6 * Fn[cell_id][i] -
                                0.25 * Fn[(cell_id + m_mesh.numberOfCells() - 1) % m_mesh.numberOfCells()][i];
        } else if (m_lambda_vector[i] > 0) {
          deltaFn[cell_id][i] = 0.25 * Fn[(cell_id + 1) % m_mesh.numberOfCells()][i] + 5. / 6 * Fn[cell_id][i] -
                                3. / 2 * Fn[(cell_id + m_mesh.numberOfCells() - 1) % m_mesh.numberOfCells()][i] +
                                0.5 * Fn[(cell_id + m_mesh.numberOfCells() - 2) % m_mesh.numberOfCells()][i] -
                                1. / 12 * Fn[(cell_id + m_mesh.numberOfCells() - 3) % m_mesh.numberOfCells()][i];
        } else {
          deltaFn[cell_id][i] = 0;
        }
      }
    }

    return deltaFn;
  }

  CellValue<double>
  compute_PFnp1(const DiscreteFunctionP0Vector<const double> F0,
                const DiscreteFunctionP0Vector<const double> deltaF,
                const DiscreteFunctionP0Vector<const double> deltaF0,
                const double& dt)
  {
    CellValue<double> PFnp1{m_mesh.connectivity()};

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      const double dx = m_dx_table[cell_id];
      const double c  = 0.5 * dt / dx;
      double PFn      = 0;
      for (size_t i = 0; i < m_lambda_vector.size(); ++i) {
        PFn += F0[cell_id][i] - c * m_lambda_vector[i] * (deltaF[cell_id][i] + deltaF0[cell_id][i]);
      }

      PFnp1[cell_id] = PFn;
    }
    return PFnp1;
  }

  CellValue<double>
  compute_PFn(const DiscreteFunctionP0Vector<const double> F)
  {
    CellValue<double> PFn{m_mesh.connectivity()};

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      double PF = 0;
      for (size_t i = 0; i < m_lambda_vector.size(); ++i) {
        PF += F[cell_id][i];
      }
      PFn[cell_id] = PF;
    }
    return PFn;
  }

  std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>, double>
  apply(const double& eps)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    const DiscreteFunctionP0Vector<const double>& F0_u = m_Fn_u;
    const DiscreteFunctionP0Vector<const double>& F0_p = m_Fn_p;

    DiscreteFunctionP0Vector<double> Fnp1_u = copy(F0_u);
    DiscreteFunctionP0Vector<double> Fnp1_p = copy(F0_p);

    // for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
    //   std::cout << "F_" << cell_id << "^n = " << Fnp1_p[cell_id] << "\n";
    // }
    // std::cout << "\n";

    DiscreteFunctionP0Vector<double> deltaFn_u(p_mesh, nb_waves);   // = compute_delta3Fn(F0_u); //
    DiscreteFunctionP0Vector<double> deltaFn_p(p_mesh, nb_waves);   // = compute_delta3Fn(F0_p); //

    const NodeArray<const double> Flux_Fn_u = compute_Polynomial_reconstruction_3(F0_u);
    const NodeArray<const double> Flux_Fn_p = compute_Polynomial_reconstruction_3(F0_p);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      const auto& cell_to_node   = cell_to_node_matrix[cell_id];
      const NodeId left_node_id  = cell_to_node[0];
      const NodeId right_node_id = cell_to_node[1];
      // const double dx            = m_dx_table[cell_id];

      for (size_t i = 0; i < nb_waves; ++i) {
        deltaFn_u[cell_id][i] = (Flux_Fn_u[right_node_id][i] - Flux_Fn_u[left_node_id][i]);
        deltaFn_p[cell_id][i] = (Flux_Fn_p[right_node_id][i] - Flux_Fn_p[left_node_id][i]);
      }
    }

    // double sum_delta = 0;
    // for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
    //   sum_delta += deltaFn_u[cell_id][0];
    // }
    // std::cout << "sum delta = " << sum_delta << "\n";

    DiscreteFunctionP0Vector<double> F_u = copy(F0_u);
    DiscreteFunctionP0Vector<double> F_p = copy(F0_p);

    for (size_t p = 0; p < 2; ++p) {
      DiscreteFunctionP0Vector<double> deltaFnp1_u(p_mesh, nb_waves);   // = compute_delta3Fn(Fnp1_u); //
      DiscreteFunctionP0Vector<double> deltaFnp1_p(p_mesh, nb_waves);   // = compute_delta3Fn(Fnp1_p);

      const NodeArray<const double> Flux_Fnp1_u = compute_Polynomial_reconstruction_3(Fnp1_u);
      const NodeArray<const double> Flux_Fnp1_p = compute_Polynomial_reconstruction_3(Fnp1_p);

      for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
        const auto& cell_to_node   = cell_to_node_matrix[cell_id];
        const NodeId left_node_id  = cell_to_node[0];
        const NodeId right_node_id = cell_to_node[1];

        for (size_t i = 0; i < nb_waves; ++i) {
          deltaFnp1_u[cell_id][i] = (Flux_Fnp1_u[right_node_id][i] - Flux_Fnp1_u[left_node_id][i]);
          deltaFnp1_p[cell_id][i] = (Flux_Fnp1_p[right_node_id][i] - Flux_Fnp1_p[left_node_id][i]);
        }
      }

      const CellValue<const double> PFnp1_u = compute_PFnp1(m_Fn_u, deltaFnp1_u, deltaFn_u, m_dt);
      const CellValue<const double> PFnp1_p = compute_PFnp1(m_Fn_p, deltaFnp1_p, deltaFn_p, m_dt);

      const CellValue<const double> PFn_u = compute_PFn(F0_u);
      const CellValue<const double> PFn_p = compute_PFn(F0_p);

      const CellArray<const double> APFnp1 = getA(PFnp1_u, PFnp1_p);
      const CellArray<const double> APFn   = getA(PFn_u, PFn_p);

      const CellValue<double> APFnp1_u{m_mesh.connectivity()};
      const CellValue<double> APFnp1_p{m_mesh.connectivity()};

      const CellValue<double> APFn_u{m_mesh.connectivity()};
      const CellValue<double> APFn_p{m_mesh.connectivity()};

      for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
        APFnp1_u[cell_id] = APFnp1[cell_id][0];
        APFnp1_p[cell_id] = APFnp1[cell_id][1];
        APFn_u[cell_id]   = APFn[cell_id][0];
        APFn_p[cell_id]   = APFn[cell_id][1];
      }

      const CellArray<const double> MPFnp1_u = compute_M(PFnp1_u, APFnp1_u);
      const CellArray<const double> MPFnp1_p = compute_M(PFnp1_p, APFnp1_p);
      const CellArray<const double> MPFn_u   = compute_M(PFn_u, APFn_u);
      const CellArray<const double> MPFn_p   = compute_M(PFn_p, APFn_p);

      const double c1 = 1. / (2 * eps + m_dt);

      for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
        const double dx = m_dx_table[cell_id];
        const double c2 = (eps * m_dt / dx);   //(eps * m_dt);

        for (size_t i = 0; i < nb_waves; ++i) {
          Fnp1_u[cell_id][i] = c1 * (m_dt * MPFnp1_u[cell_id][i] + 2 * eps * F0_u[cell_id][i] -
                                     c2 * m_lambda_vector[i] * (deltaFnp1_u[cell_id][i] + deltaFn_u[cell_id][i]) +
                                     m_dt * (MPFn_u[cell_id][i] - F0_u[cell_id][i]));
          Fnp1_p[cell_id][i] = c1 * (m_dt * MPFnp1_p[cell_id][i] + 2 * eps * F0_p[cell_id][i] -
                                     c2 * m_lambda_vector[i] * (deltaFnp1_p[cell_id][i] + deltaFn_p[cell_id][i]) +
                                     m_dt * (MPFn_p[cell_id][i] - F0_p[cell_id][i]));
        }
      }
    }

    return std::make_tuple(std::make_shared<DiscreteFunctionVariant>(Fnp1_u),
                           std::make_shared<DiscreteFunctionVariant>(Fnp1_p), m_dt);
  }

  WavesEquationKineticSolver(std::shared_ptr<const MeshType> mesh,
                             const double dt,
                             const TinyMatrix<2>& A,
                             const std::vector<double>& lambda_vector,
                             const DiscreteFunctionP0Vector<const double>& Fn_u,
                             const DiscreteFunctionP0Vector<const double>& Fn_p)
    : m_dt{dt}, m_A{A}, m_lambda_vector{lambda_vector}, m_Fn_u{Fn_u}, m_Fn_p{Fn_p}, p_mesh{mesh}, m_mesh{*mesh}
  {
    if ((lambda_vector.size() != Fn_u.size()) or (lambda_vector.size() != Fn_p.size())) {
      throw NormalError("Incompatible dimensions of lambda vector and Fn");
    }
  }

  ~WavesEquationKineticSolver() = default;
};

std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>, double>
wavesEquationKineticSolver(const double& dt,
                           const TinyMatrix<2>& A,
                           const std::vector<double>& lambda_vector,
                           const double& eps,
                           const std::shared_ptr<const DiscreteFunctionVariant>& F_u,
                           const std::shared_ptr<const DiscreteFunctionVariant>& F_p)
{
  const std::shared_ptr<const MeshVariant> mesh_v = getCommonMesh({F_u, F_p});
  if (mesh_v.use_count() == 0) {
    throw NormalError("F_u and F_p have to be defined on the same mesh");
  }

  return std::visit(
    [&](auto&& p_mesh) -> std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
                                     std::shared_ptr<const DiscreteFunctionVariant>, double> {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      if constexpr (is_polygonal_mesh_v<MeshType> and (MeshType::Dimension == 1)) {
        auto Fn_u = F_u->get<DiscreteFunctionP0Vector<const double>>();
        auto Fn_p = F_p->get<DiscreteFunctionP0Vector<const double>>();

        WavesEquationKineticSolver solver(p_mesh, dt, A, lambda_vector, Fn_u, Fn_p);

        return solver.apply(eps);
      } else {
        throw NotImplementedError("Invalid mesh type for Scalar Kinetic solver");
      }
    },
    mesh_v->variant());
}
