#include <scheme/WavesEquationNonUniformKineticSolver.hpp>

#include <language/utils/EvaluateAtPoints.hpp>
#include <mesh/Connectivity.hpp>
#include <mesh/Mesh.hpp>
#include <mesh/MeshVariant.hpp>
#include <scheme/DiscreteFunctionDescriptorP0Vector.hpp>
#include <scheme/DiscreteFunctionP0Vector.hpp>
#include <scheme/DiscreteFunctionUtils.hpp>
#include <scheme/IDiscreteFunctionDescriptor.hpp>

#include <tuple>

template <MeshConcept MeshType>
std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const ItemArrayVariant>>
getWavesEquationNonUniformKineticWaves(const std::shared_ptr<const MeshType> p_mesh,
                                       DiscreteFunctionP0<const double> u,
                                       DiscreteFunctionP0<const double> p,
                                       TinyMatrix<2> A,
                                       uint64_t nb_waves,
                                       double dt)
{
  if (nb_waves < 2) {
    throw NormalError("lambda vector must be of dimension greater than 2");
  }
  if (nb_waves > 3) {
    throw NotImplementedError("not implemented for more than 3 waves");
  }

  const auto& cell_to_node_matrix = p_mesh->connectivity().cellToNodeMatrix();
  const auto& node_to_cell_matrix = p_mesh->connectivity().nodeToCellMatrix();

  using MeshDataType                     = MeshData<MeshType>;
  MeshDataType& mesh_data                = MeshDataManager::instance().getMeshData(*p_mesh);
  const CellValue<const double> dx_table = mesh_data.Vj();

  DiscreteFunctionP0Vector<double> Fn_u(p_mesh, nb_waves);
  DiscreteFunctionP0Vector<double> Fn_p(p_mesh, nb_waves);
  NodeArray<double> lambda_nodes(p_mesh->connectivity(), nb_waves);
  CellArray<double> lambda_cells(p_mesh->connectivity(), nb_waves);
  CellValue<double> inv_L2norm_lambda_cells(p_mesh->connectivity());

  double inv_dt = 1. / dt;
  for (NodeId node_id = 0; node_id < p_mesh->numberOfNodes(); ++node_id) {
    const auto& node_to_cell = node_to_cell_matrix[node_id];
    lambda_nodes[node_id].fill(0);
    CellId prev_cell_id;
    CellId next_cell_id;
    if (node_id == 0) {
      prev_cell_id = p_mesh->numberOfCells() - 1;
      next_cell_id = node_to_cell[0];
    } else if (node_id == p_mesh->numberOfNodes() - 1) {
      prev_cell_id = node_to_cell[0];
      next_cell_id = CellId(0);
    } else {
      for (size_t i_cell = 0; i_cell < node_to_cell.size(); ++i_cell) {
      }
      prev_cell_id = node_to_cell[0];
      next_cell_id = node_to_cell[1];
    }
    lambda_nodes[node_id][0] = inv_dt * std::min(dx_table[prev_cell_id], dx_table[next_cell_id]);
    // const double min_Aprime  = 3;
    //  if (lambda_nodes[node_id][0] < min_Aprime) {
    //    lambda_nodes[node_id][0] = min_Aprime;
    //  }
    lambda_nodes[node_id][nb_waves - 1] = -lambda_nodes[node_id][0];
  }

  // std::cout << "lambda = " << lambda_nodes << "\n";
  lambda_cells.fill(0);
  for (CellId cell_id = 0; cell_id < p_mesh->numberOfCells(); ++cell_id) {
    const auto& cell_to_node = cell_to_node_matrix[cell_id];
    for (size_t i = 0; i < nb_waves; ++i) {
      for (size_t i_node = 0; i_node < cell_to_node.size(); ++i_node) {
        NodeId node_id = cell_to_node[i_node];
        lambda_cells[cell_id][i] += lambda_nodes[node_id][i];
      }
      lambda_cells[cell_id][i] /= 2;
      // std::cout << "lambda[" << cell_id << "][" << i << "] = " << lambda_cells[cell_id][i] << "\n \n";
    }

    inv_L2norm_lambda_cells[cell_id] = 0;
    for (size_t i = 0; i < nb_waves; ++i) {
      inv_L2norm_lambda_cells[cell_id] += std::pow(lambda_cells[cell_id][i], 2);
    }
    inv_L2norm_lambda_cells[cell_id] = 1. / inv_L2norm_lambda_cells[cell_id];

    //    std::cout << "1/|lambda_j| = " << inv_L2norm_lambda_cells[cell_id] << "\n";

    for (size_t i = 0; i < nb_waves; ++i) {
      Fn_u[cell_id][i] = (1. / nb_waves) * u[cell_id] + inv_L2norm_lambda_cells[cell_id] *
                                                          (A(0, 0) * u[cell_id] + A(0, 1) * p[cell_id]) *
                                                          lambda_cells[cell_id][i];
      Fn_p[cell_id][i] = (1. / nb_waves) * p[cell_id] + inv_L2norm_lambda_cells[cell_id] *
                                                          (A(1, 0) * u[cell_id] + A(1, 1) * p[cell_id]) *
                                                          lambda_cells[cell_id][i];
    }
  }

  return std::make_tuple(std::make_shared<DiscreteFunctionVariant>(Fn_u),
                         std::make_shared<DiscreteFunctionVariant>(Fn_p),
                         std::make_shared<ItemArrayVariant>(lambda_nodes));
}

std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const ItemArrayVariant>>
getWavesEquationNonUniformKineticWaves(const std::shared_ptr<const DiscreteFunctionVariant>& u,
                                       const std::shared_ptr<const DiscreteFunctionVariant>& p,
                                       const TinyMatrix<2>& A,
                                       const uint64_t& nb_waves,
                                       const double& dt)
{
  const std::shared_ptr<const MeshVariant> mesh_v = getCommonMesh({u, p});
  if (mesh_v.use_count() == 0) {
    throw NormalError("u and p have to be defined on the same mesh");
  }

  return std::visit(
    [&](auto&& p_mesh)
      -> std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                    std::shared_ptr<const ItemArrayVariant>> {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      if constexpr (is_polygonal_mesh_v<MeshType> and (MeshType::Dimension == 1)) {
        auto uh = u->get<DiscreteFunctionP0<const double>>();
        auto ph = p->get<DiscreteFunctionP0<const double>>();

        return getWavesEquationNonUniformKineticWaves(p_mesh, uh, ph, A, nb_waves, dt);
      } else {
        throw NotImplementedError("Invalid mesh type for Scalar Kinetic solver");
      }
    },
    mesh_v->variant());
}

template <MeshConcept MeshType>
class WavesEquationNonUniformKineticSolver
{
 private:
  constexpr static size_t Dimension = MeshType::Dimension;
  using Rd                          = TinyVector<Dimension>;

  using DiscreteWavesEquationFunction = DiscreteFunctionP0<double>;

  const double m_dt;
  const TinyMatrix<2> m_A;
  const NodeArray<const double> m_lambda_nodes;
  const DiscreteFunctionP0Vector<const double> m_Fn_u;
  const DiscreteFunctionP0Vector<const double> m_Fn_p;
  const std::shared_ptr<const MeshType> p_mesh;
  const MeshType& m_mesh;
  const size_t m_nb_components                      = m_Fn_u.size();
  const CellValue<const double> m_dx_table          = MeshDataManager::instance().getMeshData(m_mesh).Vj();
  const CellValue<const TinyVector<Dimension>> m_xj = MeshDataManager::instance().getMeshData(m_mesh).xj();

 public:
  CellArray<const double>
  getA(const CellValue<const double>& u, const CellValue<const double>& p) const
  {
    CellArray<double> vec_A{m_mesh.connectivity(), 2};
    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < 2; ++i) {
        vec_A[cell_id][i] = m_A(i, 0) * u[cell_id] + m_A(i, 1) * p[cell_id];
      }
    }
    return vec_A;
  }

  CellArray<const double>
  compute_M(const CellValue<const double>& u, const CellValue<const double> Au) const
  {
    CellArray<double> M{m_mesh.connectivity(), m_nb_components};
    const auto& cell_to_node_matrix = p_mesh->connectivity().cellToNodeMatrix();
    CellArray<double> lambda_cells{m_mesh.connectivity(), m_nb_components};
    CellValue<double> inv_L2norm_lambda_cells{m_mesh.connectivity()};

    for (CellId cell_id = 0; cell_id < p_mesh->numberOfCells(); ++cell_id) {
      const auto& cell_to_node = cell_to_node_matrix[cell_id];
      lambda_cells[cell_id].fill(0);
      for (size_t i = 0; i < m_nb_components; ++i) {
        for (size_t i_node = 0; i_node < cell_to_node.size(); ++i_node) {
          NodeId node_id = cell_to_node[i_node];
          lambda_cells[cell_id][i] += m_lambda_nodes[node_id][i];
        }
        lambda_cells[cell_id][i] /= 2;
      }

      inv_L2norm_lambda_cells[cell_id] = 0;
      for (size_t i = 0; i < m_nb_components; ++i) {
        inv_L2norm_lambda_cells[cell_id] += std::pow(lambda_cells[cell_id][i], 2);
      }
      inv_L2norm_lambda_cells[cell_id] = 1. / inv_L2norm_lambda_cells[cell_id];

      for (size_t i = 0; i < m_nb_components; ++i) {
        M[cell_id][i] = (1. / m_nb_components) * u[cell_id] +
                        inv_L2norm_lambda_cells[cell_id] * Au[cell_id] * lambda_cells[cell_id][i];
      }
    }

    return M;
  }

  DiscreteFunctionP0Vector<double>
  compute_delta1Fn(const DiscreteFunctionP0Vector<const double>& Fn)
  {
    const auto& cell_to_node_matrix = p_mesh->connectivity().cellToNodeMatrix();

    DiscreteFunctionP0Vector<double> deltaFn(p_mesh, m_nb_components);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      const auto& cell_to_node = cell_to_node_matrix[cell_id];
      NodeId prev_node_id      = cell_to_node[0];
      NodeId next_node_id      = cell_to_node[1];

      for (size_t i = 0; i < m_nb_components; ++i) {
        if (m_lambda_nodes[next_node_id][i] < 0) {
          deltaFn[cell_id][i] = Fn[(cell_id + 1) % m_mesh.numberOfCells()][i] - Fn[cell_id][i];
        } else if (m_lambda_nodes[prev_node_id][i] > 0) {
          deltaFn[cell_id][i] = Fn[cell_id][i] - Fn[(cell_id + m_mesh.numberOfCells() - 1) % m_mesh.numberOfCells()][i];
        } else {
          deltaFn[cell_id][i] = 0;
        }
      }
    }

    return deltaFn;
  }

  DiscreteFunctionP0Vector<double>
  compute_delta2Fn(const DiscreteFunctionP0Vector<const double>& Fn)
  {
    throw NotImplementedError("order 2 in space not implemented yet");

    DiscreteFunctionP0Vector<double> deltaFn(p_mesh, m_nb_components);
    double lambda_prov = 1;
    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < m_nb_components; ++i) {
        if (lambda_prov < 0) {
          deltaFn[cell_id][i] = -1. / 3 * Fn[(cell_id + m_mesh.numberOfCells() - 1) % m_mesh.numberOfCells()][i] -
                                0.5 * Fn[cell_id][i] + Fn[(cell_id + 1) % m_mesh.numberOfCells()][i] -
                                1. / 6 * Fn[(cell_id + 2) % m_mesh.numberOfCells()][i];
        } else if (lambda_prov > 0) {
          deltaFn[cell_id][i] = 1. / 3 * Fn[(cell_id + 1) % m_mesh.numberOfCells()][i] + 0.5 * Fn[cell_id][i] -
                                Fn[(cell_id + m_mesh.numberOfCells() - 1) % m_mesh.numberOfCells()][i] +
                                1. / 6 * Fn[(cell_id + m_mesh.numberOfCells() - 2) % m_mesh.numberOfCells()][i];
        } else {
          deltaFn[cell_id][i] = 0;
        }
      }
    }

    return deltaFn;
  }

  CellValue<double>
  compute_PFnp1_u(const DiscreteFunctionP0Vector<const double> F_u, const double& dt)
  {
    CellValue<double> PFnp1{m_mesh.connectivity()};
    const auto& cell_to_node_matrix = p_mesh->connectivity().cellToNodeMatrix();

    DiscreteFunctionP0Vector<double> deltaF  = compute_delta1Fn(F_u);
    DiscreteFunctionP0Vector<double> deltaF0 = compute_delta1Fn(m_Fn_u);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      const auto& cell_to_node = cell_to_node_matrix[cell_id];
      NodeId prev_node_id      = cell_to_node[0];
      NodeId next_node_id      = cell_to_node[1];

      const double dx = m_dx_table[cell_id];
      const double c  = 0.5 * dt / dx;
      double PFn      = 0;
      for (size_t i = 0; i < m_nb_components; ++i) {
        if (m_lambda_nodes[next_node_id][i] < 0) {
          PFn += m_Fn_u[cell_id][i] - c * m_lambda_nodes[next_node_id][i] * (deltaF[cell_id][i] + deltaF0[cell_id][i]);
        } else if (m_lambda_nodes[prev_node_id][i] > 0) {
          PFn += m_Fn_u[cell_id][i] - c * m_lambda_nodes[prev_node_id][i] * (deltaF[cell_id][i] + deltaF0[cell_id][i]);
        } else {
          PFn += m_Fn_u[cell_id][i];
        }
      }

      PFnp1[cell_id] = PFn;
    }
    return PFnp1;
  }

  CellValue<double>
  compute_PFnp1_p(const DiscreteFunctionP0Vector<const double> F_p, const double& dt)
  {
    CellValue<double> PFnp1{m_mesh.connectivity()};
    const auto& cell_to_node_matrix = p_mesh->connectivity().cellToNodeMatrix();

    DiscreteFunctionP0Vector<double> deltaF  = compute_delta1Fn(F_p);
    DiscreteFunctionP0Vector<double> deltaF0 = compute_delta1Fn(m_Fn_p);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      const auto& cell_to_node = cell_to_node_matrix[cell_id];
      NodeId prev_node_id      = cell_to_node[0];
      NodeId next_node_id      = cell_to_node[1];

      const double dx = m_dx_table[cell_id];
      const double c  = 0.5 * dt / dx;
      double PFn      = 0;
      for (size_t i = 0; i < m_nb_components; ++i) {
        if (m_lambda_nodes[next_node_id][i] < 0) {
          PFn += m_Fn_p[cell_id][i] - c * m_lambda_nodes[next_node_id][i] * (deltaF[cell_id][i] + deltaF0[cell_id][i]);
        } else if (m_lambda_nodes[prev_node_id][i] > 0) {
          PFn += m_Fn_p[cell_id][i] - c * m_lambda_nodes[prev_node_id][i] * (deltaF[cell_id][i] + deltaF0[cell_id][i]);
        } else {
          PFn += m_Fn_p[cell_id][i];
        }
      }

      PFnp1[cell_id] = PFn;
    }
    return PFnp1;
  }

  CellValue<double>
  compute_PFn(const DiscreteFunctionP0Vector<const double> F)
  {
    CellValue<double> PFn{m_mesh.connectivity()};

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      double PF = 0;
      for (size_t i = 0; i < m_nb_components; ++i) {
        PF += F[cell_id][i];
      }
      PFn[cell_id] = PF;
    }
    return PFn;
  }

  std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>, double>
  apply(const double& eps, const double& dt)
  {
    // std::cout << "dt = " << dt << "\n";
    /// m_lambda_vector[0]));
    const auto& cell_to_node_matrix = p_mesh->connectivity().cellToNodeMatrix();

    const DiscreteFunctionP0Vector<const double>& F0_u = m_Fn_u;
    const DiscreteFunctionP0Vector<const double>& F0_p = m_Fn_p;
    DiscreteFunctionP0Vector<double> Fnp1_u            = copy(F0_u);
    DiscreteFunctionP0Vector<double> Fnp1_p            = copy(F0_p);
    DiscreteFunctionP0Vector<double> deltaFn_u         = compute_delta1Fn(F0_u);
    DiscreteFunctionP0Vector<double> deltaFn_p         = compute_delta1Fn(F0_p);

    DiscreteFunctionP0Vector<double> Fu = copy(F0_u);
    DiscreteFunctionP0Vector<double> Fp = copy(F0_p);

    for (size_t p = 0; p < 2; ++p) {
      DiscreteFunctionP0Vector<double> deltaFnp1_u = compute_delta1Fn(Fnp1_u);
      DiscreteFunctionP0Vector<double> deltaFnp1_p = compute_delta1Fn(Fnp1_p);

      const CellValue<const double> PFnp1_u = compute_PFnp1_u(Fnp1_u, dt);
      const CellValue<const double> PFnp1_p = compute_PFnp1_p(Fnp1_p, dt);

      const CellValue<const double> PFn_u = compute_PFn(F0_u);
      const CellValue<const double> PFn_p = compute_PFn(F0_p);

      const CellArray<const double> APFnp1 = getA(PFnp1_u, PFnp1_p);
      const CellArray<const double> APFn   = getA(PFn_u, PFn_p);
      const CellValue<double> APFnp1_u{m_mesh.connectivity()};
      const CellValue<double> APFnp1_p{m_mesh.connectivity()};
      const CellValue<double> APFn_u{m_mesh.connectivity()};
      const CellValue<double> APFn_p{m_mesh.connectivity()};
      for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
        APFnp1_u[cell_id] = APFnp1[cell_id][0];
        APFnp1_p[cell_id] = APFnp1[cell_id][1];
        APFn_u[cell_id]   = APFn[cell_id][0];
        APFn_p[cell_id]   = APFn[cell_id][1];
      }

      const CellArray<const double> MPFnp1_u = compute_M(PFnp1_u, APFnp1_u);
      const CellArray<const double> MPFnp1_p = compute_M(PFnp1_p, APFnp1_p);
      const CellArray<const double> MPFn_u   = compute_M(PFn_u, APFn_u);
      const CellArray<const double> MPFn_p   = compute_M(PFn_p, APFn_p);

      const double c1 = 1. / (2 * eps + dt);

      for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
        const auto& cell_to_node = cell_to_node_matrix[cell_id];
        NodeId prev_node_id      = cell_to_node[0];
        NodeId next_node_id      = cell_to_node[1];

        const double dx = m_dx_table[cell_id];
        const double c2 = (eps * dt / dx);

        for (size_t i = 0; i < m_nb_components; ++i) {
          if (m_lambda_nodes[next_node_id][i] < 0) {
            Fnp1_u[cell_id][i] =
              c1 * (dt * MPFnp1_u[cell_id][i] + 2 * eps * F0_u[cell_id][i] -
                    c2 * m_lambda_nodes[next_node_id][i] * (deltaFnp1_u[cell_id][i] + deltaFn_u[cell_id][i]) +
                    dt * (MPFn_u[cell_id][i] - F0_u[cell_id][i]));
            Fnp1_p[cell_id][i] =
              c1 * (dt * MPFnp1_p[cell_id][i] + 2 * eps * F0_p[cell_id][i] -
                    c2 * m_lambda_nodes[next_node_id][i] * (deltaFnp1_p[cell_id][i] + deltaFn_p[cell_id][i]) +
                    dt * (MPFn_p[cell_id][i] - F0_p[cell_id][i]));

          } else if (m_lambda_nodes[prev_node_id][i] > 0) {
            Fnp1_u[cell_id][i] =
              c1 * (dt * MPFnp1_u[cell_id][i] + 2 * eps * F0_u[cell_id][i] -
                    c2 * m_lambda_nodes[prev_node_id][i] * (deltaFnp1_u[cell_id][i] + deltaFn_u[cell_id][i]) +
                    dt * (MPFn_u[cell_id][i] - F0_u[cell_id][i]));
            Fnp1_p[cell_id][i] =
              c1 * (dt * MPFnp1_p[cell_id][i] + 2 * eps * F0_p[cell_id][i] -
                    c2 * m_lambda_nodes[prev_node_id][i] * (deltaFnp1_p[cell_id][i] + deltaFn_p[cell_id][i]) +
                    dt * (MPFn_p[cell_id][i] - F0_p[cell_id][i]));

          } else {
            Fnp1_u[cell_id][i] = c1 * (dt * MPFnp1_u[cell_id][i] + 2 * eps * F0_u[cell_id][i] +
                                       dt * (MPFn_u[cell_id][i] - F0_u[cell_id][i]));
            Fnp1_p[cell_id][i] = c1 * (dt * MPFnp1_p[cell_id][i] + 2 * eps * F0_p[cell_id][i] +
                                       dt * (MPFn_p[cell_id][i] - F0_p[cell_id][i]));
          }
        }
      }
    }

    return std::make_tuple(std::make_shared<DiscreteFunctionVariant>(Fnp1_u),
                           std::make_shared<DiscreteFunctionVariant>(Fnp1_p), dt);
  }

  WavesEquationNonUniformKineticSolver(std::shared_ptr<const MeshType> mesh,
                                       const double dt,
                                       const TinyMatrix<2>& A,
                                       const NodeArray<const double>& lambda_nodes,
                                       const DiscreteFunctionP0Vector<const double>& Fn_u,
                                       const DiscreteFunctionP0Vector<const double>& Fn_p)
    : m_dt{dt}, m_A{A}, m_lambda_nodes{lambda_nodes}, m_Fn_u{Fn_u}, m_Fn_p{Fn_p}, p_mesh{mesh}, m_mesh{*mesh}
  {
    if ((m_lambda_nodes.sizeOfArrays() != Fn_u.size()) or (m_lambda_nodes.sizeOfArrays() != Fn_p.size())) {
      throw NormalError("Incompatible dimensions of lambda vector and Fn");
    }
  }

  ~WavesEquationNonUniformKineticSolver() = default;
};

std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>, double>
wavesEquationNonUniformKineticSolver(const double& dt,
                                     const TinyMatrix<2>& A,
                                     const std::shared_ptr<const ItemArrayVariant>& lambda_nodes_vector,
                                     const double& eps,
                                     const std::shared_ptr<const DiscreteFunctionVariant>& F_u,
                                     const std::shared_ptr<const DiscreteFunctionVariant>& F_p)
{
  const std::shared_ptr<const MeshVariant> mesh_v = getCommonMesh({F_u, F_p});
  if (mesh_v.use_count() == 0) {
    throw NormalError("F_u and F_p have to be defined on the same mesh");
  }

  return std::visit(
    [&](auto&& p_mesh) -> std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
                                     std::shared_ptr<const DiscreteFunctionVariant>, double> {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      if constexpr (is_polygonal_mesh_v<MeshType> and (MeshType::Dimension == 1)) {
        auto Fn_u = F_u->get<DiscreteFunctionP0Vector<const double>>();
        auto Fn_p = F_p->get<DiscreteFunctionP0Vector<const double>>();

        WavesEquationNonUniformKineticSolver solver(p_mesh, dt, A, lambda_nodes_vector->get<NodeArray<const double>>(),
                                                    Fn_u, Fn_p);

        return solver.apply(eps, dt);
      } else {
        throw NotImplementedError("Invalid mesh type for Scalar Kinetic solver");
      }
    },
    mesh_v->variant());
}
